/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_accept.c: ACCEPT command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_accept.c,v 1.10 2002/12/15 21:26:17 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_accept.so",
	"/ACCEPT command",
	4,
	"$Revision: 1.10 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("ACCEPT", m_accept, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("ACCEPT", m_accept);
	return;
}
#endif

static int add_accept(aClient *sptr, aClient *acptr)
{
	Link *lp;
	int cnt = 0;

	for (lp = sptr->user->accept; lp != NULL; lp = lp->next) {
		if (lp->flags != ACCEPT_LINK_ME) {
			continue;
		}
		if (++cnt >= MAXACCEPTS) {
			sendto_one(sptr, err_str(ERR_TOOMANYACCEPT), me.name, sptr->name, acptr->name,
				MAXACCEPTS);
			return 0;
		}
		else if (lp->value.cptr == acptr) {
			return 0;
		}
	}

	lp = make_slink();
	lp->value.cptr = acptr;
	lp->flags = ACCEPT_LINK_ME;
	lp->next = sptr->user->accept;
	sptr->user->accept = lp;

	lp = make_slink();
	lp->value.cptr = sptr;
	lp->flags = ACCEPT_LINK_REMOTE;
	lp->next = acptr->user->accept;
	acptr->user->accept = lp;

	sendto_one(sptr, rpl_str(RPL_ACCEPTSTATUS), me.name, sptr->name, acptr->name, "added to");
	return 0;
}

static int del_accept(aClient *sptr, aClient *acptr)
{
	Link *lp, **list;
	int found = 0;

	for (list = &(sptr->user->accept); *list != NULL; list = &((*list)->next)) {
		if ((*list)->flags != ACCEPT_LINK_ME) {
			continue;
		}
		if ((*list)->value.cptr == acptr) {
			lp = *list;
			*list = lp->next;
			free_slink(lp);
			found++;
			break;
		}
	}
	if (found == 0) {
		sendto_one(sptr, ":%s %d %s :%s is not in your message accept list", me.name, RPL_ACCEPTINFO,
			sptr->name, acptr->name);
		return 0;
	}

	found = 0;
	for (list = &(acptr->user->accept); *list != NULL; list = &((*list)->next)) {
		if ((*list)->flags != ACCEPT_LINK_REMOTE) {
			continue;
		}
		if ((*list)->value.cptr == sptr) {
			lp = *list;
			*list = lp->next;
			free_slink(lp);
			found++;
			break;
		}
	}
	if (found == 0) {
		ircdlog(LOG_DEFAULT, "d_acc: %s was in acceptme list for %s but not in acceptremote list",
			acptr->name, sptr->name);
	}
	sendto_one(sptr, rpl_str(RPL_ACCEPTSTATUS), me.name, sptr->name, acptr->name, "removed from");
	return 0;
}

/*
 * m_accept
 *	parv[0] = sender prefix
 *	parv[1] = parameter
 */
int m_accept(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	Link *lp;
	char *p = NULL, *s, *cn, **ptr;
	aClient *acptr, *lastcptr = NULL;
	int didlist = 0, didhelp = 0, didanything = 0;
	static char *accept_help[] = {
		"/ACCEPT [<+|->nick[,<+|->nick, ...]] [list] [help]",
		" ",
		"If you are in privacy mode (usermode +p), you can still accept private messages",
		"from users you choose using the ACCEPT command. You simply add the user(s) you wish",
		"to accept messages from to your accept list. For example, if you would like to",
		"receive messages from the user Bob, you would use the following command, and Bob",
		"will be free to send you private messages.",
		" ",
		"/quote access +Bob",
		" ",
		"If you then want to remove a user from your accept list, you do by using the",
		"following command. Again, in the example, user Bob is being removed.",
		" ",
		"/quote accept -Bob",
		" ",
		"You can also add and remove more than one user from your accept list at the same",
		"time, for example:",
		" ",
		"/quote accept +user1,-user2,+user3",
		" ",
		"Note: if either you or a client on your accept list changes nicknames, the accept",
		"list references in question will be removed. This is to prevent user tracing.",
		" ",
		"To see a list of all users on your accept list, use the 'list' parameter, and to",
		"see this help text again, use the 'help' parameter.",
		NULL
	};

	if (!MyClient(sptr)) {
		return 0;
	}
	if (parc < 2) {
		sendto_one(sptr, ":%s NOTICE %s :Type /ACCEPT HELP for more information on the message "
			"accept system.", me.name, parv[0]);
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ACCEPT");
		return 0;
	}
	for (s = strtoken(&p, parv[1], ","); s; s = strtoken(&p, NULL, ",")) {
		while (*s == ' ') {
			s++;
		}
		if (*s == '+') {
			didanything++;
			cn = s + 1;
			if (*cn == '\0') {
				continue;
			}
			if ((acptr = find_person(cn, NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], cn);
				continue;
			}
			if (acptr == sptr) {
				continue;
			}
			if (acptr == lastcptr) {
				sendto_realops_lev(SPY_LEV, "User %s (%s@%s) may be flooding ACCEPT: add %s",
					sptr->name, sptr->username, MaskedHost(sptr), cn);
			}
			lastcptr = acptr;
			add_accept(sptr, acptr);
		}
		else if (*s == '-') {
			didanything++;
			cn = s + 1;
			if (*cn == '\0') {
				continue;
			}
			if ((acptr = find_person(cn, NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], cn);
				continue;
			}
			if (acptr == sptr) {
				continue;
			}
			if (acptr == lastcptr) {
				sendto_realops_lev(SPY_LEV, "User %s (%s@%s) may be flooding ACCEPT: del %s",
					sptr->name, sptr->username, MaskedHost(sptr), cn);
			}
			lastcptr = acptr;
			del_accept(sptr, acptr);
		}
		else {
			if (didlist == 0 && !mycmp(s, "LIST")) {
				didanything++;
				didlist = 1;

				sendto_one(sptr, ":%s %d %s :The following users are on your message accept list:",
					me.name, RPL_ACCEPTINFO, sptr->name);
				for (lp = sptr->user->accept; lp != NULL; lp = lp->next) {
					if (lp->flags == ACCEPT_LINK_REMOTE) {
						continue;
					}
					acptr = lp->value.cptr;
					sendto_one(sptr, ":%s %d %s :%s (%s@%s)", me.name, RPL_ACCEPTLIST,
						sptr->name, acptr->name, acptr->username, MaskedHost(acptr));
				}
				sendto_one(sptr, rpl_str(RPL_ENDOFACCEPTLIST), me.name, parv[0], s);
			}
			else if (didhelp == 0 && !mycmp(s, "HELP")) {
				didanything++;
				didhelp = 1;
				for (ptr = accept_help; *ptr; ptr++) {
					sendto_one(sptr, ":%s %d %s :%s", me.name, RPL_ACCEPTINFO, parv[0], *ptr);
				}
				sendto_one(sptr, rpl_str(RPL_ENDOFACCEPTLIST), me.name, parv[0], s);
			}
		}
	}
	if (didanything == 0) {
		sendto_one(sptr, ":%s NOTICE %s :Type /ACCEPT HELP for more information on the message accept system.",
			me.name, parv[0]);
	}
	return 0;
}
