/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_snick.c: SNICK protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_snick.c,v 1.9 2002/12/15 21:26:19 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern int user_modes[];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_snick.so",
	"SNICK protocol",
	4,
	"$Revision: 1.9 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("SNICK", m_snick, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("SNICK", m_snick);
	return;
}
#endif

/*
 * m_snick
 *	parv[0] = sender prefix
 *	parv[1] = nickname
 *	parv[2] = client ts
 *	parv[3] = hopcount
 *	parv[4] = username
 *	parv[5] = hostname
 *	parv[6] = ip
 *	parv[7] = maskedhost
 *	parv[8] = server
 *	parv[9] = services id
 *	parv[10] = user modes
 *	parv[11] = gcos
 */
int m_snick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	int keepours = 0, samenick = 0, flag, *i;
	aClient *acptr, *uplink = NULL;
	char *target = NULL, *m;
	static char ubuf[16];
	ts_val ts = 0;

	if (!IsServer(sptr) || parc < 12) {
		return 0;
	}
	if ((uplink = find_server(parv[8], NULL)) == NULL) {
		sendto_realops_lev(SKILL_LEV, "Unknown server in SNICK for %s", parv[1]);
		ircstp->is_kill++;
		sendto_one(cptr, ":%s KILL %s :Unknown server in SNICK", me.name, parv[1]);
		return 0;
	}
	if ((acptr = find_server(parv[1], NULL)) != NULL) {
		sendto_realops_lev(SKILL_LEV, "Server/Nick collision on %s", parv[1]);
		ircstp->is_kill++;
		sendto_one(cptr, ":%s KILL %s :Server/Nick collision", me.name, parv[1]);
		return 0;
	}
	if ((acptr = find_client(parv[1], NULL)) != NULL) {
		ts = atol(parv[2]);
		if (acptr->tsinfo < ts) {
			keepours = 1;
		}
		else if (acptr->tsinfo == ts) {
			keepours = -1;
		}
		if (acptr->user != NULL && !match(acptr->host, parv[5]) && keepours != -1) {
			samenick = 1;
			keepours = (keepours == 1) ? 0 : 1;
		}
		if (keepours <= 0) {
			target = (keepours == -1) ? "all" : (samenick == 1) ? "old" : "new";
			sendto_realops_lev(SKILL_LEV, "Nick collision on %s (%s nick%s killed)", acptr->name,
				target, (keepours == -1) ? "s" : "");
			ircstp->is_kill++;
			sendto_serv_butone(cptr, ":%s KILL %s :Nick collision (%s nick%s killed)", me.name,
				acptr->name, target, (keepours == -1) ? "s" : "");
			acptr->flags |= FLAGS_KILLED;
			exit_client(NULL, acptr, &me,
				(keepours == -1) ? "Nick collision (all nicks killed)" :
				(samenick == 1) ? "Nick collision (old nick killed)" :
				"Nick collision (new nick killed)");
			if (keepours != 0) {
				return 0;
			}
		}
	}

	acptr = make_client(cptr, uplink);
	add_client_to_list(acptr);

	strcpy(acptr->name, parv[1]);
	acptr->tsinfo = (ts == 0) ? atol(parv[2]) : ts;
	acptr->hopcount = atoi(parv[3]);

	make_user(acptr);
	add_to_client_hash_table(parv[1], acptr);

	strncpyzt(acptr->username, parv[4], USERLEN + 1);
	strncpyzt(acptr->host, parv[5], HOSTLEN + 1);
	acptr->ip.s_addr = ntohl(strtol(parv[6], NULL, 0));
	acptr->user->server = find_or_add(parv[8]);
	acptr->user->servicestamp = strtoul(parv[9], NULL, 0);
	strncpyzt(acptr->info, parv[11], sizeof(acptr->info));

	SetClient(acptr);
	if (++Count.total > Count.max_tot) {
		Count.max_tot = Count.total;
	}
	for (m = &parv[10][1]; *m; m++) {
		for (i = user_modes; (flag = *i); i += 2) {
			if (*m == (char)(*(i + 1))) {
				acptr->umode |= flag & SEND_UMODES;
				break;
			}
		}
	}
	if (HasMode(acptr, UMODE_OPER)) {
		Count.oper++;
	}
	if (HasMode(acptr, UMODE_INVISIBLE)) {
		Count.invisi++;
	}
	if (HasMode(acptr, UMODE_MASKED)) {
		if (*parv[7] != '*') {
			mask_client(acptr, parv[7]);
		}
		else {
			DelMode(acptr, UMODE_MASKED);
		}
	}
	hash_check_watch(sptr, RPL_LOGON);

        send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);
        if (*ubuf == '\0') {
                ubuf[0] = '+';
                ubuf[1] = '\0';
        }
	sendto_snick_servs_butone(sptr, "SNICK %s %ld %d %s %s %lu %s %s %lu %s :%s", acptr->name,
		acptr->tsinfo, acptr->hopcount + 1, acptr->username, acptr->host,
		htonl(acptr->ip.s_addr), HasMode(acptr, UMODE_MASKED) ? acptr->user->maskedhost : "*",
		acptr->user->server, acptr->user->servicestamp, ubuf, acptr->info);
	sendto_nickip_servs_butone(0, sptr, "NICK %s %d %ld %s %s %s %s %lu :%s", acptr->name,
		acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->username, acptr->host,
		acptr->user->server, acptr->user->servicestamp, acptr->info);
	sendto_nickip_servs_butone(1, sptr, "NICK %s %d %ld %s %s %s %s %lu %lu :%s", acptr->name,
		acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->username, acptr->host,
		acptr->user->server, acptr->user->servicestamp, htonl(sptr->ip.s_addr), acptr->info);
	return 0;
}
