/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_stats.c: STATS command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_stats.c,v 1.41 2002/12/16 00:01:44 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "res.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern int max_connection_count;
extern int max_client_count;
extern void show_oper_serv_fdlists(aClient *);
extern void count_memory(aClient *, char *);
#ifdef HAVE_THROTTLE_ON
extern void throttle_stats(aClient *);
#endif

static char buf[BUFSIZE];
static int bufpos = 0;

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_stats.so",
	"/STATS command",
	4,
	"$Revision: 1.41 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("STATS", m_stats, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
        delCommand("STATS", m_stats);
	return;
}
#endif

static inline void s_addtobuf(char *string)
{
	char *ptr;
	for (ptr = string; *ptr; ptr++) {
		buf[bufpos++] = *ptr;
	}
	buf[bufpos] = '\0';
	return;
}

/*
 * m_stats
 *	parv[0] = sender prefix
 *	parv[1] = stat
 *	parv[2] = server name
 */
int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static char Sformat[] = ":%s %d %s :Name SendQ SendM SendBytes RcveM RcveBytes OpenSince Idle TS";
	static char Lformat[] = ":%s %d %s :%s %u %u %u %u %u %u %u %s";
	static char LSformat[] = ":%s %d %s :Name Clients TotalConnections SendM SendBytes RcvM RcvBytes Idle Options";
	static char LLformat[] = ":%s %d %s :%s %u %u %u %u %u %u %u (%s)";

	aClient *acptr;
	char stat = parc > 1 ? parv[1][0] : '\0', *name;
	int i, doall = 0, wilds = 0;
	time_t sincetime = 0L;   
	static time_t last_used = 0L;

	if (hunt_server(cptr, sptr, ":%s STATS %s :%s", 2, parc, parv) != HUNTED_ISME) {
		return 0;
	}
	if (!HasMode(sptr, UMODE_OPER)) {
		if (last_used + PACETIME_INTENSE > timeofday) {
			sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
			return 0;
		}
		else {
			last_used = timeofday;
		}
		if (SPYNOTICES == 1 && IsPerson(sptr)) {
			if (stat == '\0') {
				sendto_realops_lev(SPY_LEV, "STATS 'NULL' requested by %s (%s@%s) [%s]",
					sptr->name, sptr->username, MaskedHost(sptr),
					sptr->user->server);
				return 0;
			}
			else {   
				sendto_realops_lev(SPY_LEV, "STATS '%c' requested by %s (%s@%s) [%s]",
					stat, sptr->name, sptr->username, MaskedHost(sptr),
					sptr->user->server);
			}
		}
	}
	if (parc > 2) {
		name = parv[2];
		if (!mycmp(name, me.name)) {
			doall = 2;
		}
		else if (!match(name, me.name)) {
			doall = 1;
		}
		if (strchr(name, '*') || strchr(name, '?')) {
			wilds = 1;
		}
	}
	else {
		name = me.name;
	}

	switch (stat) {
		case 'A':
		case 'a':
		{
			if (!HasMode(sptr, UMODE_ADMIN)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			report_dns_nameservers(sptr);
			break;
		}
		case 'L':
		case 'l':
		{
			sendto_one(sptr, Sformat, me.name, RPL_STATSLINKINFO, parv[0]);
			if (parc > 2 && !(doall || wilds)) {
				if ((acptr = find_person(name, NULL)) == NULL || !MyConnect(sptr)) {
					break;
				}
				sincetime = acptr->since > timeofday ? 0 : timeofday - acptr->since;
				sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0],
					(HasMode(sptr, UMODE_RSTAFF)) ? get_client_name(acptr, TRUE) :
					get_client_name(acptr, HIDEME), (int)DBufLength(&acptr->sendQ),
					(int)acptr->sendM, (int)acptr->sendK, (int)acptr->receiveM,
					(int)acptr->receiveK, timeofday - acptr->firsttime,
					sincetime,
					IsServer(acptr) ? DoesTS(acptr) ? "TS" : "Non-TS" : "-");
			}
			else {
				for (i = 0; i <= highest_fd; i++) {
					if (!(acptr = local[i]) || !IsServer(acptr)) {
						continue;
					}
					if (HIDEULINES == 1 && IsULine(acptr) && !HasMode(sptr, UMODE_OPER)) {
						continue;
					}
					sincetime = acptr->since > timeofday ? 0 : timeofday - acptr->since;
					sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0],
						get_client_name(acptr, HIDEME), (int)DBufLength(&acptr->sendQ),
						(int)acptr->sendM, (int)acptr->sendK, (int)acptr->receiveM,
						(int)acptr->receiveK, timeofday - acptr->firsttime, sincetime,
						IsServer(acptr) ? DoesTS(acptr) ? "TS" : "Non-TS" : "-");
				}
			}
			break;
		}
		case 'C':
		case 'c':
		case 'H':
		case 'h':
		{
			ConfigItem_link *link = NULL;
			if (HIDEULINES == 1 && !HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			for (link = ci_link; link != NULL; link = (ConfigItem_link *)link->next) {
				sendto_one(sptr, err_str(RPL_STATSCLINE), me.name, parv[0], "C",
					HasMode(sptr, UMODE_RSTAFF) ? link->fromip : "0.0.0.0",
					link->servername, link->port, link->class->name);
			}
			break;
		}
		case 'K':
		case 'k':
		{
			ConfigItem_ban *ban = NULL;
			char flag = '\0';
			for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
				if (ban->type & BAN_USERHOST) {
					if (ban->type & BAN_USERHOST_S) {
						flag = 'A';
					}
					else {
						flag = ban->flag.temporary == 1 ? 'k' : 'K';
					}
				}
				else if (ban->type & BAN_IP) {
					flag = ban->flag.temporary == 1 ? 'z' : 'Z';
				}
				else if (ban->type & BAN_SERVER) {
					flag = 'S';
				}
				else {
					continue;
				}
				sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name, parv[0], flag,
					ban->mask, ban->reason);
			}
			break;
		}
		case 'Q':
		case 'q':
		{
			ConfigItem_ban *ban = NULL;
			for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
				if (!(ban->type & BAN_NICKNAME) && !(ban->type & BAN_CHANNEL)) {
					continue;
				}
				sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name, parv[0], 'Q',
					ban->mask, ban->reason);
			}
			break;
		}
		case 'B':
		case 'b':
		{
			ConfigItem_ban *ban = NULL;
			for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
				if (!(ban->type & BAN_FILE)) {
					continue;
				}
				sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name, parv[0], 'B',
					ban->mask, ban->reason);
			}
			break;
		}
		case 'G':
		case 'g':
		{
			ConfigItem_ban *ban = NULL;
			for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
				if (!(ban->type & BAN_GCOS)) {
					continue;
				}
				sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name, parv[0],
					ban->flag.temporary == 1 ? 'g' : 'G', ban->mask, ban->reason);
			}
			break;
		}
		case 'E':
		{
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			show_events(sptr);
			break;
		}
		case 'e':
		{
			ConfigItem_except *except = NULL;
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			for (except = ci_except; except != NULL; except = (ConfigItem_except *)except->next) {
				sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name, parv[0],
					except->flag.temporary == 1 ? 'e' : 'E', except->mask, "*");
			}
			break;
		}
		case 'U':
		{
			ConfigItem_uline *uline = NULL;
			if (HIDEULINES == 1 && !HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			for (uline = ci_uline; uline != NULL; uline = (ConfigItem_uline *)uline->next) {
				sendto_one(sptr, rpl_str(RPL_STATSULINE), me.name, parv[0], "U",
					uline->servername);
			}
			break;
		}
		case 'u':
		{
			time_t now = timeofday - me.since;
			sendto_one(sptr, rpl_str(RPL_STATSUPTIME), me.name, parv[0], now / 86400,
				(now / 3600) % 24, (now / 60) % 60, now % 60);
			sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0], max_connection_count,
				max_client_count, Count.cli_restart);
			break;
		}
		case 'I':
		case 'i':
		{
			ConfigItem_allow *allow = NULL;
			for (allow = ci_allow; allow !=  NULL; allow = (ConfigItem_allow *)allow->next) {
				sendto_one(sptr, rpl_str(RPL_STATSILINE), me.name, parv[0], allow->ip,
					allow->hostname, allow->maxperip, allow->class->name);
			}
			break;
		}
		case 'M':
		case 'm':
		{
			int i = 0;
			aCommand *command = NULL;
			for (i = 0; i < 256; i++) {
				for (command = CommandHash[i]; command; command = command->next) {
					sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS), me.name, parv[0],
						command->cmd, command->count, command->bytes);
				}
			}
			break;
		}
		case 'O':
		case 'o':
		{
			ConfigItem_oper *oper = NULL;
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			for (oper = ci_oper; oper != NULL; oper = (ConfigItem_oper *)oper->next) {
				sendto_one(sptr, rpl_str(RPL_STATSOLINE), me.name, parv[0], "O",
					oper->name, oper->class->name);
			}
			break;
		}
		case 'p':
		case 'P':
		{
			Listener *l;
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			sendto_one(sptr, LSformat, me.name, RPL_STATSLINKINFO, sptr->name);
			for (l = listener_list; l != NULL; l = l->next) {
				buf[(bufpos = 0)] = '\0';
				if (l->options & BIND_CLIENTONLY) {
					s_addtobuf("client-only");
				}
				if (l->options & BIND_SERVERONLY) {
					if (bufpos > 0) {
						s_addtobuf(", ");
					}
					s_addtobuf("server-only");
				}
				if (l->options & BIND_SECURE) {
					if (bufpos > 0) {
						s_addtobuf(", ");
					}
					s_addtobuf("secure");
				}
				if (l->options & BIND_CRJAVA) {
					if (bufpos > 0) {
						s_addtobuf(", ");
					}
					s_addtobuf("crjava");
				}
				if (l->bind->flag.temporary == 1) {
					if (bufpos > 0) {
						s_addtobuf(", ");
					}
					s_addtobuf("temporary");
				}
				sendto_one(sptr, LLformat, me.name, RPL_STATSLINKINFO, sptr->name,
					get_listener_name(l), l->bind->clients, l->count, (int)l->sendM, (int)l->sendK,
					(int)l->receiveM, (int)l->receiveK, timeofday - l->lasttime,
					(bufpos > 0) ? buf : "none");
			}
			break;
		}
#ifdef HAVE_THROTTLE_ON
		case 'T':
		{
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			throttle_stats(sptr);
			break;
		}
#endif
		case 't':
		{
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			tstats(sptr, parv[0]);
			break;
		}
#ifdef DEBUGMODE
		case 'R':
		case 'r':
		{
			send_usage(sptr, parv[0]);
			break;
		}
#endif
		case 'S':
		case 's':
		{
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			list_scache(sptr, sptr, parc, parv);
			break;
		}
		case 'N':
		case 'n':
		{
			sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
				"User Connects Today: ", Count.today);
			sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
				"User Connects past week: ", Count.weekly);
			sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
				"User Connects past month: ", Count.monthly);
			sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
				"User Conects past year: ", Count.yearly);
			break;
		}
		case 'V':
		{
			ConfigItem_vhost *vhost = NULL;
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			for (vhost = ci_vhost; vhost != NULL; vhost = (ConfigItem_vhost *)vhost->next) {
				sendto_one(sptr, ":%s %d %s :V %s %s", me.name, RPL_TEXT, parv[0],
					vhost->mask, vhost->login);
			}
			break;
		}
		case 'v':
		{
			show_oper_serv_fdlists(sptr);
			break;
		}
		case 'Y':
		case 'y':
		{
			ConfigItem_class *class = NULL;
			for (class = ci_class; class != NULL; class = (ConfigItem_class *)class->next) {
				sendto_one(sptr, rpl_str(RPL_STATSYLINE), me.name, parv[0], "Y",
					class->name, class->pingtime, class->connectfreq,
					class->maxclients, class->sendq, class->clients);
			}
			break;
		}
		case 'Z':
		case 'z':
		{
			if (!HasMode(sptr, UMODE_OPER)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				break;
			}
			count_memory(sptr, parv[0]);
			break;
		}
		case '?':
		{
			serv_info(sptr, parv[0]);
			break;
		}
		default:
		{
			if (stat == '\0') {
				sendto_one(sptr, ":%s NOTICE %s :*** Please specify a staticstic flag.",
					me.name, parv[0]);
			}
			else {
				sendto_one(sptr, ":%s NOTICE %s :*** '%c' is an unknown statistic.", me.name,
					parv[0], stat);
			}
			break;
		}
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
	return 0;
}
