/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_svinfo.c: SVINFO protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_svinfo.c,v 1.19 2002/12/15 21:26:19 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "patchlevel.h"
#include "zlink.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_svinfo.so",
	"SVINFO protocol",
	4,
	"$Revision: 1.19 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("SVINFO", m_svinfo, MAXPARA, MSG_UNREGUSE);
	return;
}
void _module_unload(void) {
	delCommand("SVINFO", m_svinfo);
	return;
}
#endif

/*
 * m_svinfo
 *	parv[0] = sender prefix
 *
 *	if (parc == 2)
 *		parv[1] = ZIP (compression initialisation)
 *
 *	if (parc > 2)
 *		parv[1] = TS_CURRENT
 *		parv[2] = TS_MIN
 *		parv[3] = standalone or connected to non-TS (unused)
 *		parv[4] = UTC time
 *
 *		if (parc > 5) (link is rage2<->rage2)
 *			parv[5] = ircd codename
 *			parv[6] = ircnetwork
 *			parv[7] = masking keys (key1.key2.key3)
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *inpath = get_client_name(cptr, HIDEME);
	time_t delta, tmptime = timeofday, theirtime;
	char keys[50];

	if (!IsServer(sptr) || !MyConnect(sptr)) {
		return 0;
	}
	if (parc == 2 && !mycmp(parv[1], "ZIP")) {
		sptr->serv->sflags |= SFLAGS_ZIP_IN;
		sptr->serv->zip_in = zip_create_input_session();
		sendto_gnotice("Input from %s is now compressed", inpath);
#ifdef HAVE_HUB_ON
		sendto_serv_butone(sptr, ":%s GNOTICE :Input from %s is now compressed", inpath);
#endif
		return ZIP_NEXT_BUFFER;
	}
	if (parc < 5) {
		sendto_gnotice("Link %s dropped, incorrect number of SVINFO parameters", inpath);
		sendto_one(sptr, "ERROR :Incorrect number of SVINFO parameters");
		return exit_client(sptr, sptr, sptr, "Incorrect number of SVINFO parameters");
	}
	if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) {
		sendto_gnotice("Link %s dropped, wrong TS protocol version (%s, %s)",
			inpath, parv[1], parv[2]);
		sendto_one(sptr, "ERROR :Wrong TS protocol version");
		return exit_client(sptr, sptr, sptr, "Wrong TS protocl version");
	}

	theirtime = atol(parv[4]);
	delta = abs(theirtime - tmptime);

	if (delta > TS_MAX_DELTA) {
		sendto_gnotice("Link %s dropped, excessive TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
		sendto_one(sptr, "ERROR :Excessive TS delta");
		return exit_client(sptr, sptr, sptr, "Excessive TS delta");
	}
	if (delta > TS_WARN_DELTA) {
		sendto_gnotice("Link %s has a notable TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
	}
	if (parc > 5) {
		if (irccmp(CODENAME, parv[5])) {
			sendto_gnotice("Link %s dropped, incompatible server version (ours=%s, theirs=%s)",
				inpath, CODENAME, parv[5]);
			sendto_one(sptr, "ERROR :Incompatible server version (please upgrade)");
			return exit_client(sptr, sptr, sptr, "Incompatible server version (please upgrade)");
		}
		if (irccmp(IRCNETWORK, parv[6])) {
			sendto_gnotice("Link %s dropped, irc network mismatch (ours=%s, theirs=%s)",
				inpath, IRCNETWORK, parv[6]);
			sendto_one(sptr, "ERROR :IRC network mismatch");
			return exit_client(sptr, sptr, sptr, "IRC network mismatch");
		}

		ircsprintf(keys, "%ld.%ld.%ld", MASK_KEY1, MASK_KEY2, MASK_KEY3);
		if (irccmp(keys, parv[7])) {
			sendto_gnotice("Link %s dropped, masking key mismatch", inpath);
			sendto_one(sptr, "ERROR :Masking key mismatch");
			return exit_client(sptr, sptr, sptr, "Masking key mismatch");
		}
	}
	return 0;
}
