/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_vhost.c: VHOST command/protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_vhost.c,v 1.22 2002/12/15 21:26:20 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_vhost.so",
	"/VHOST command/protocol",
	4,
	"$Revision: 1.22 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("VHOST", m_vhost, MAXPARA, MSG_NONE);
	return;
}
void _module_unload(void) {
	delCommand("VHOST", m_vhost);
	return;
}
#endif

static void handle_vhost_server(aClient *cptr, char *client, char *mask)
{
	aClient *acptr;
	if (cptr == NULL || BadPtr(client) || BadPtr(mask)) {
		ircdlog(LOG_ERROR, "handle_vhost_server() encountered a NULL pointer! cptr %lX "
			"client %s mask %s", cptr, BadPtr(client) ? "NULL!" : client,
			BadPtr(mask) ? "NULL!" : mask);
		sendto_realops_lev(DEBUG_LEV, "Encountered NULL pointer in Server-VHOST function (logged)");
		return;
	}
	if ((acptr = find_client(client, NULL)) == NULL) {
		ircdlog(LOG_ERROR, "Received Server-VHOST for unknown client %s", client);
		return;
	}
	AddMode(acptr, UMODE_MASKED);
	mask_client(acptr, mask);
	sendto_serv_butone(cptr, "VHOST %s :%s", acptr->name, acptr->user->maskedhost);
        return;
}

static void handle_vhost_client(aClient *cptr, char *command, char *login, char *password)
{
	if (!MyClient(cptr)) {
		return;
	}
	if (!mycmp(command, "ON")) {
		ConfigItem_vhost *vhost = NULL;
		char *from;

		if (BadPtr(login) || BadPtr(password)) {
			sendto_one(cptr, err_str(ERR_NEEDMOREPARAMS), me.name, cptr->name, "VHOST");
			return;
		}

		from = make_user_host(cptr->username, cptr->host);
		for (vhost = ci_vhost; vhost != NULL; vhost = (ConfigItem_vhost *)vhost->next) {
			if (!irccmp(vhost->login, login) && source_check((Source **)&vhost->source, from) &&
			  auth_check(vhost->auth, password)) {
				break;
			}
		}
		if (vhost == NULL) {
			sendto_realops_lev(SPY_LEV, "%s (%s@%s) failed to log into a vhost", cptr->name,
				cptr->username, MaskedHost(cptr));
			sendto_one(cptr, ":%s NOTICE %s :*** Authentication failure", me.name, cptr->name);
			return;
		}

		sendto_realops_lev(SPY_LEV, "%s (%s@%s) has successfully logged into vhost %s", cptr->name,
			cptr->username, MaskedHost(cptr), vhost->mask);
		if (!HasMode(cptr, UMODE_MASKED)) {
			AddMode(cptr, UMODE_MASKED);
			sendto_one(cptr, ":%s MODE %s :+z", cptr->name, cptr->name);
		}
		mask_client(cptr, vhost->mask);
		cptr->user->uflags |= UFLAGS_VHOST;
		sendto_one(cptr, ":%s NOTICE %s :*** Your hostname is now %s", me.name, cptr->name,
			cptr->user->maskedhost);
		sendto_serv_butone(cptr, "VHOST %s :%s", cptr->name, cptr->user->maskedhost);
	}
	else if (!mycmp(command, "OFF")) {
		if (!(cptr->user->uflags & UFLAGS_VHOST)) {
			return;
		}
		if (!HasMode(cptr, UMODE_MASKED)) {
			AddMode(cptr, UMODE_MASKED);
			sendto_one(cptr, ":%s MODE %s :+z", cptr->name, cptr->name);
		}
		cptr->user->uflags &= ~UFLAGS_VHOST;
		mask_client(cptr, maskme(cptr));
		sendto_one(cptr, ":%s NOTICE %s :*** Your hostname is now %s", me.name, cptr->name,
			cptr->user->maskedhost);
		sendto_serv_butone(cptr, "VHOST %s :%s", cptr->name, cptr->user->maskedhost);
	}
	else {
		sendto_one(cptr, ":%s NOTICE %s :*** VHOST %s: Unknown command", me.name, cptr->name,
			command);
	}
	return;
}

/*
 * m_vhost
 *	parv[0] = sender prefix
 *
 *	if (parc == 3) - server VHOST
 *		parv[1] = client
 *		parv[2] = vhost
 *
 *	if (parc == 2) - client VHOST OFF
 *		parv[1] = OFF
 *
 *	if (parc > 3) - client VHOST ON
 *		parv[1] = ON
 *		parv[2] = login
 *		parv[3] = password
 */
int m_vhost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (IsServer(sptr)) {
		if (!HasCapab(sptr, CAPAB_VHOST)) {
			sendto_realops_lev(DEBUG_LEV, "Received VHOST command from non-VHOST server %s",
				get_client_name(sptr, HIDEME));
			return 0;
		}
		if (parc < 3 || BadPtr(parv[2])) {
			sendto_realops_lev(DEBUG_LEV, "Invalid parameter count in VHOST received from "
				"%s", get_client_name(sptr, HIDEME));
			return 0;
		}
		handle_vhost_server(sptr, parv[1], parv[2]);
		return 0;
	}
	if (parc < 2 || (parc == 2 && BadPtr(parv[1])) || (parc > 3 && BadPtr(parv[3]))) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "VHOST");
		return 0;
	}
	handle_vhost_client(sptr, parv[1], (parc > 3) ? parv[2] : NULL, (parc > 3) ? parv[3] : NULL);
	return 0;
}
