/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/connauth.c: dns/identd connection authentication
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: connauth.c,v 1.18 2002/11/16 21:12:27 rageircd Exp $
 */

#include "memory.h"
#include "struct.h"
#include "common.h"
#include "config.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(__hpux)
#include "inet.h"
#endif
#include <fcntl.h>
#include "sock.h"
#include "setup.h"
#include "ssl.h"
#include "h.h"
#include "fdlist.h"

extern char aup_notice[256];
extern int aup_notice_len;

ConnectAuth *connauth_list = NULL;

static ConnectAuth *connauth_incomplete = NULL;

static HeaderMessage HeaderMessages[] = {
	{ "NOTICE AUTH :*** Looking up your hostname...\r\n", 46 },
	{ "NOTICE AUTH :*** Found your hostname\r\n", 38 },
	{ "NOTICE AUTH :*** Couldn't look you your hostname\r\n", 50 },
	{ "NOTICE AUTH :*** Checking identd...\r\n", 37 },
	{ "NOTICE AUTH :*** Received identd response\r\n", 43 },
	{ "NOTICE AUTH :*** No identd response\r\n", 37 },
	{ "NOTICE AUTH :*** Your hostname is too long; using ip address instead\r\n", 70 },
	{ "NOTICE AUTH :*** Your forward and reverse DNS do not match; using ip address instead\r\n", 86 }
};

static inline void connauth_send_header(aClient *cptr, int index)
{
	if (SHOWHEADERS == 1
#ifdef HAVE_OPENSSL_ON
	  && !(cptr->listener->options & BIND_SECURE)
#endif
	  ) {
		ircsend(cptr, HeaderMessages[index].string, HeaderMessages[index].length);
	}
	return;
}

static void connauth_link(ConnectAuth *auth, ConnectAuth **list)
{
	auth->prev = NULL;
	if ((auth->next = *list) != NULL) {
		auth->next->prev = auth;
	}
	*list = auth;
	return;
}

static void connauth_unlink(ConnectAuth *auth, ConnectAuth **list)
{
	if (auth->prev != NULL) {
		auth->prev->next = auth->next;
	}
	else {
		*list = auth->next;
	}
	if (auth->next != NULL) {
		auth->next->prev = auth->prev;
	}
	return;
}

static void connauth_release_client(aClient *cptr)
{
	if (cptr->fd > highest_fd) {
		highest_fd = cptr->fd;
	}
	local[cptr->fd] = cptr;
	add_client_to_list(cptr);
	return;
}

static void connauth_handle_id_error(ConnectAuth *auth)
{
	ircstp->is_abad++;

	close(auth->fd);
	auth->fd = -1;
	auth->flags &= ~(CAF_ID_CONNECTING|CAF_ID_PENDING);

	connauth_send_header(auth->cptr, CONNAUTH_FAIL_ID);
	connauth_unlink(auth, &connauth_list);

	if (auth->flags & CAF_DNS_PENDING) {
		connauth_link(auth, &connauth_incomplete);
	}
	else {
		connauth_release_client(auth->cptr);
		MyFree(auth);
	}
	return;
}

static int connauth_init_id(ConnectAuth *auth)
{
	struct sockaddr_in sock, localaddr;
	int addrlen = sizeof(struct sockaddr_in);
	int fd = -1;

	if (CHECKIDENTD == 0) {
		return 0;
	}
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		char *tmp = get_client_name(auth->cptr, TRUE);
		Debug((DEBUG_ERROR, "unable to create connauth id socket for %s", tmp));
		ircdlog(LOG_ERROR, "unable to create connauth id socket for %s", tmp);
		ircstp->is_abad++;
		return 0;
	}
	if (fd > (HARD_FDLIMIT - 10)) {
		sendto_realops_lev(DEBUG_LEV, "Cannot allocate fd for auth on %s",
			get_client_name(auth->cptr, TRUE));
		close(fd);
		return 0;
	}

	set_non_blocking(fd);
	connauth_send_header(auth->cptr, CONNAUTH_DO_ID);
	memset(&localaddr, '\0', addrlen);
	getsockname(auth->cptr->fd, (struct sockaddr *)&localaddr, &addrlen);
	localaddr.sin_port = 0;

	if (bind(fd, (struct sockaddr *)&localaddr, addrlen) == -1) {
		report_error("binding auth stream socket %s:%s", auth->cptr);
		close(fd);
		return 0;
	}

	memcpy(&sock.sin_addr, &auth->cptr->ip, sizeof(struct in_addr));
	sock.sin_port = htons(113);
	sock.sin_family = AF_INET;

	if (connect(fd, (struct sockaddr *)&sock, addrlen) == -1 && errno != EINPROGRESS) {
		ircstp->is_abad++;
		close(fd);
		connauth_send_header(auth->cptr, CONNAUTH_FAIL_ID);
		return 0;
	}
	auth->fd = fd;
	auth->flags |= CAF_ID_CONNECTING;
	return 1;
}

static char *connauth_parse_id(char *id)
{
	int remp = 0, locp = 0;
	char *colon1, *colon2, *colon3, *comma, *remstring = id;

	if ((colon1 = strchr(remstring, ':')) == NULL) {
		return NULL;
	}
	*colon1 = '\0';
	colon1++;

	if ((colon2 = strchr(colon1, ':')) == NULL) {
		return NULL;
	}
	*colon2 = '\0';
	colon2++;

	if ((comma = strchr(remstring, ',')) == NULL) {
		return NULL;
	}
	*comma = '\0';
	comma++;

	if ((remp = atoi(remstring)) == 0) {
		return NULL;
	}
	if ((locp = atoi(comma)) == 0) {
		return NULL;
	}
	if (!strstr(colon1, "USERID")) {
		return NULL;
	}
	if ((colon3 = strchr(colon2, ':')) == NULL) {
		return NULL;
	}
	*colon3++ = '\0';
	return colon3;
}

void connauth_send_id(ConnectAuth *auth)
{
	struct sockaddr_in us, them;
	char authbuf[32];
	int ulen, tlen, alen;

	tlen = ulen = sizeof(us);
	if (getsockname(auth->cptr->fd, (struct sockaddr *)&us, &ulen) ||
	  getpeername(auth->cptr->fd, (struct sockaddr *)&them, &tlen)) {
		ircdlog(LOG_ERROR, "auth get{sock,peer}name error for %s", get_client_name(auth->cptr, TRUE));
		connauth_handle_id_error(auth);
		return;
	}

	alen = ircsprintf(authbuf, "%u , %u\r\n", (unsigned int)ntohs(them.sin_port), (unsigned int)ntohs(us.sin_port));
	if (send(auth->fd, authbuf, alen, 0) == -1) {
		if (errno == EAGAIN) {
			return;
		}
		connauth_handle_id_error(auth);
		return;
	}
	auth->flags &= ~CAF_ID_CONNECTING;
	auth->flags |= CAF_ID_PENDING;
	return;
}

void connauth_read_id(ConnectAuth *auth)
{
	char buf[129], *s = NULL, *t = NULL;
	int len, usercnt;

	if ((len = recv(auth->fd, buf, 128, 0)) > 0) {
		buf[len] = '\0';
		if ((s = connauth_parse_id(buf)) != NULL) {
			t = auth->cptr->username;
			for (usercnt = USERLEN; *s && *s != '@' && usercnt > 0; s++) {
				if (!isspace(*s) && *s != ':') {
					*t++ = *s;
					usercnt--;
				}
			}
			*t = '\0';
		}
	}
	if (len < 0 && errno == EAGAIN) {
		return;
	}
	close(auth->fd);
	auth->fd = -1;
	auth->flags &= ~(CAF_ID_CONNECTING|CAF_ID_PENDING);

	if (s == NULL) {
		ircstp->is_abad++;
		strcpy(auth->cptr->username, "unknown");
	}
	else {
		connauth_send_header(auth->cptr, CONNAUTH_FIN_ID);
		ircstp->is_asuc++;
		auth->cptr->flags |= FLAGS_GOTID;
	}

	connauth_unlink(auth, &connauth_list);
	if (auth->flags & CAF_DNS_PENDING) {
		connauth_link(auth, &connauth_incomplete);
	}
	else {
		connauth_release_client(auth->cptr);
		MyFree(auth);
	}
	return;
}

void connauth_dns_callback(void *vauth, adns_answer *answer)
{
	ConnectAuth *auth = (ConnectAuth *)vauth;

	auth->flags &= ~CAF_DNS_PENDING;
	if (answer != NULL && answer->status == adns_s_ok) {
		if (strlen(*answer->rrs.str) <= HOSTLEN) {
			strncpyzt(auth->cptr->host, *answer->rrs.str, HOSTLEN + 1);
			connauth_send_header(auth->cptr, CONNAUTH_FIN_DNS);
		}
		else {
			strncpyzt(auth->cptr->host, auth->cptr->sockhost, HOSTLEN + 1);
			connauth_send_header(auth->cptr, CONNAUTH_HOST_TOOLONG);
		}
	}
	else {
		strncpyzt(auth->cptr->host, auth->cptr->sockhost, HOSTLEN + 1);
		connauth_send_header(auth->cptr, CONNAUTH_FAIL_DNS);
	}
	if (answer != NULL) {
		MyFree(answer);
	}

	auth->cptr->dnsquery = NULL;
	if (!(auth->flags & CAF_ID_CONNECTING) && !(auth->flags & CAF_ID_PENDING)) {
		connauth_release_client(auth->cptr);
		connauth_unlink(auth, &connauth_incomplete);
		MyFree(auth);
	}
	return;
}

void connauth_init(aClient *cptr)
{
	ConnectAuth *auth = NULL;

	if (cptr == NULL) {
		return;
	}
	if (aup_notice_len > 0) {
		ircsend(cptr, aup_notice, aup_notice_len);
	}

	auth = (ConnectAuth *)MyMallocX(sizeof(ConnectAuth));
	auth->fd = -1;
	auth->cptr = cptr;
	auth->timeout = timeofday + CONNAUTH_TIMEOUT;

	cptr->dnsquery = (DNSQuery *)MyMalloc(sizeof(DNSQuery));
	cptr->dnsquery->ptr = auth;
	cptr->dnsquery->callback = connauth_dns_callback;

	connauth_send_header(cptr, CONNAUTH_DO_DNS);
	dns_getaddr(&cptr->ip, cptr->dnsquery);
	auth->flags |= CAF_DNS_PENDING;

	if (connauth_init_id(auth) == 1) {
		connauth_link(auth, &connauth_list);
	}
	else if (auth->flags & CAF_DNS_PENDING) {
		connauth_link(auth, &connauth_incomplete);
	}
	else {
		MyFree(auth);
		connauth_release_client(cptr);
	}
	return;
}

void connauth_timeout_queries(time_t now)
{
	ConnectAuth *auth, *next = NULL;
	for (auth = connauth_list; auth != NULL; auth = next) {
		next = auth->next;
		if (now > auth->timeout) {
			if (auth->fd > -1) {
				close(auth->fd);
			}
			connauth_send_header(auth->cptr, CONNAUTH_FAIL_ID);

			if (auth->flags & CAF_DNS_PENDING) {
				del_dns_queries(auth->cptr->dnsquery);
				auth->cptr->dnsquery = NULL;
				connauth_send_header(auth->cptr, CONNAUTH_FAIL_DNS);
			}
			Debug((DEBUG_NOTICE, "AUTH/DNS timeout for %s", get_client_name(auth->cptr, TRUE)));
			auth->cptr->since = now;
			connauth_release_client(auth->cptr);
			connauth_unlink(auth, &connauth_list);
			MyFree(auth);
		}
	}
	for (auth = connauth_incomplete; auth != NULL; auth = next) {
		next = auth->next;
		if (now > auth->timeout) {
			if (auth->fd > -1) {
				close(auth->fd);
			}
			del_dns_queries(auth->cptr->dnsquery);
			auth->cptr->dnsquery = NULL;
			connauth_send_header(auth->cptr, CONNAUTH_FAIL_DNS);
			Debug((DEBUG_NOTICE, "DNS timeout for %s", get_client_name(auth->cptr, TRUE)));
			auth->cptr->since = now;
			connauth_release_client(auth->cptr);
			connauth_unlink(auth, &connauth_incomplete);
			MyFree(auth);
		}
	}
	return;
}
