/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/dbuf.c: data buffer management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: dbuf.c,v 1.7 2002/10/25 22:42:23 rageircd Exp $
 */

#include "dbuf.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include <stdlib.h>
#include "memory.h"

int DBufCountC = 0;
int DBufCountS = 0;
int DBufUsedCountC = 0;
int DBufUsedCountS = 0;

static DBufBuffer *dbufFreeListC = NULL;
static DBufBuffer *dbufFreeListS = NULL;
static DBufBlock *dbufBlocks = NULL;
static int dbuf_buffer_header_size = 0;
static unsigned int block_seq = 0;

void count_dbuf_memory(size_t *allocatedC, size_t *usedC, size_t *allocatedS, size_t *usedS, int *allocnC,
  int *usednC, int *allocnS, int *usednS)
{
	*allocatedC = DBufCountC * (dbuf_buffer_header_size + DBUF_SIZE_CLIENT);
	*usedC = DBufUsedCountC * (dbuf_buffer_header_size + DBUF_SIZE_CLIENT);
	*allocatedS = DBufCountS * (dbuf_buffer_header_size + DBUF_SIZE_SERVER);
	*usedS = DBufUsedCountS * (dbuf_buffer_header_size + DBUF_SIZE_SERVER);
	*allocnC = DBufCountC;
	*usednC = DBufUsedCountC;
	*allocnS = DBufCountS;
	*usednS = DBufUsedCountC;
	return;
}

void dbuf_allocblock(int num, int server)
{
	DBufBlock *dblock;
	DBufBuffer *dfree, *dbp;
	int i;
	int bufsize = (server == 1) ? DBUF_SIZE_SERVER : DBUF_SIZE_CLIENT;

	dblock = (DBufBlock *)MyMallocX(sizeof(DBufBlock));
	dblock->num = num;
	dblock->block_seq = block_seq++;
	dblock->next = dbufBlocks;
	dblock->blocksize = bufsize;
	dbufBlocks = dblock;

	dbp = (DBufBuffer *)MyMalloc((dbuf_buffer_header_size + bufsize) * num);
	dblock->buf = dbp;
	dfree = dbp;

	for (i = 0; i < num - 1; i++) {
		DBufBuffer *dbnext;
		dfree->parent = dblock;
		dbnext = (DBufBuffer *)(((char *)dfree) + bufsize + dbuf_buffer_header_size);
		dfree->next = dbnext;
		dfree = dbnext;
	}

	dfree->parent = dblock;
	if (server == 1) {
		dfree->next = dbufFreeListS;
		dbufFreeListS = dbp;
		DBufCountS += num;
	}
	else {
		dfree->next = dbufFreeListC;
		dbufFreeListC = dbp;
		DBufCountC += num;
	}
	return;
}

void init_dbuf()
{
	dbuf_buffer_header_size = sizeof(DBufBuffer) - sizeof(char[1]);
	dbuf_allocblock(INITIAL_CLIENT_DBUFS, 0);
	dbuf_allocblock(INITIAL_SERVER_DBUFS, 1);
	return;
}

static DBufBuffer *dbuf_alloc(int server)
{
	DBufBuffer *db = (server == 1) ? dbufFreeListS : dbufFreeListC;

	if (db == NULL) {
		dbuf_allocblock((server == 1) ? INITIAL_SERVER_DBUFS : INITIAL_CLIENT_DBUFS, server);
		if ((db = (server == 1) ? dbufFreeListS : dbufFreeListC) == NULL) {
			return NULL;
		}
	}
	if (server == 1) {
		dbufFreeListS = dbufFreeListS->next;
		DBufUsedCountS++;
	}
	else {
		dbufFreeListC = dbufFreeListC->next;
		DBufUsedCountC++;
	}
	db->next = NULL;
	db->start = db->end = db->data;
	return db;
}

void dbuf_block_free(DBufBlock *dbb)
{
	DBufBlock *db, *dbprev = NULL;
	int (*callback)(int, int);
	int mem, memsize;

	for (db = dbufBlocks; db != NULL; db = db->next) {
		if (db == dbb) {
			if (dbprev != NULL) {
				dbprev->next = db->next;
			}
			else {
				dbufBlocks = db->next;
			}
			mem = db->num;
			memsize = db->num + db->blocksize;
			callback = db->callback;
			MyFree(db->buf);
			MyFree(db);
			if (callback != NULL) {
				callback(mem, memsize);
			}
			return;
		}
		dbprev = db;
	}
	return;
}

static void dbuf_free(DBufBuffer *ptr)
{
	if (ptr->parent->is_freeing == 1) {
		int server;
		if (ptr->parent->blocksize == DBUF_SIZE_CLIENT) {
			DBufUsedCountC--;
			server = 0;
		}
		else if (ptr->parent->blocksize == DBUF_SIZE_SERVER) {
			DBufUsedCountS--;
			server = 1;
		}
		else {
			ircdlog(LOG_ERROR, "FATAL: dbuf_free() intercepted unknown buffer, not client, not server");
			abort();
		}

		ptr->parent->given_back++;
		if (ptr->parent->given_back == ptr->parent->num) {
			if (server == 1) {
				DBufCountS -= ptr->parent->num;
			}
			else {
				DBufCountC -= ptr->parent->num;
			}
			dbuf_block_free(ptr->parent);
		}
		return;
	}
	if (ptr->parent->blocksize == DBUF_SIZE_CLIENT) {
		DBufUsedCountC--;
		ptr->next = dbufFreeListC;
		dbufFreeListC = ptr;
	}
	else if (ptr->parent->blocksize == DBUF_SIZE_SERVER) {
		DBufUsedCountS--;
		ptr->next = dbufFreeListS;
		dbufFreeListS = ptr;
	}
	else {
		ircdlog(LOG_ERROR, "FATAL: dbuf_free() intercepted unknown block, not client, not server");
		abort();
	}
	return;
}

static int dbuf_malloc_error(DBuf *dyn)
{
	DBufBuffer *db;

	dyn->length = 0;
	while ((db = dyn->head) != NULL) {
		dyn->head = db->next;
		dbuf_free(db);
	}
	dyn->tail = NULL;
	return 0;
}

int dbuf_put(DBuf *dyn, const char *buf, size_t length, int server)
{
	DBufBuffer **h, *d;
	int chunk;

	if (dyn->length == 0) {
		h = &(dyn->head);
	}
	else {
		h = &(dyn->tail);
	}
	dyn->length += length;

	for (; length > 0; h = &(d->next)) {
		if ((d = *h) == NULL) {
			if ((d = dbuf_alloc(server)) == NULL) {
				return dbuf_malloc_error(dyn);
			}
			dyn->tail = d;
			*h = d;
		}
		if ((chunk = (d->data + d->parent->blocksize) - d->end) != 0) {
			if (chunk > length) {
				chunk = length;
			}
			memcpy(d->end, buf, chunk);
			length -= chunk;
			buf += chunk;
			d->end += chunk;
		}
	}
	return 1;
}

char *dbuf_map(DBuf *dyn, size_t *length)
{
	if (dyn->length == 0) {
		*length = 0;
		return NULL;
	}
	*length = dyn->head->end - dyn->head->start;
	return dyn->head->start;
}

void dbuf_delete(DBuf *dyn, size_t length)
{
	DBufBuffer *db;
	size_t chunk;

	if (length > dyn->length) {
		length = dyn->length;
	}
	while (length > 0) {
		if ((db = dyn->head) == NULL) {
			break;
		}
		if ((chunk = db->end - db->start) > length) {
			chunk = length;
		}
		length -= chunk;
		dyn->length -= chunk;
		db->start += chunk;
		if (db->start == db->end) {
			dyn->head = db->next;
			dbuf_free(db);
		}
	}
	if (dyn->head == NULL) {
		dyn->tail = 0;
		dyn->length = 0;
	}
	return;
}

size_t dbuf_get(DBuf *dyn, char *buf, size_t length)
{
	size_t moved = 0, chunk;
	const char *b;

	while (length > 0 && (b = dbuf_map(dyn, &chunk)) != NULL) {
		if (chunk > length) {
			chunk = length;
		}
		memcpy(buf, b, chunk);
		dbuf_delete(dyn, chunk);
		buf += chunk;
		length -= chunk;
		moved += chunk;
	}
	return moved;
}

static size_t dbuf_flush(DBuf *dyn)
{
	DBufBuffer *db = dyn->head;

	if (db == NULL) {
		return 0;
	}
	while (IsEol(*db->start)) {
		if (++db->start == db->end) {
			dyn->head = db->next;
			dbuf_free(db);
			if ((db = dyn->head) == NULL) {
				dyn->tail = 0;
				dyn->length = 0;
				break;
			}
		}
		dyn->length--;
	}
	return dyn->length;
}

int dbuf_getmsg(DBuf *dyn, char *buf, size_t length)
{
	DBufBuffer *db;
	char *start, *end;
	size_t count, copied = 0;

	if (dbuf_flush(dyn) == 0) {
		return 0;
	}

	db = dyn->head;
	start = db->start;

	if (length > dyn->length) {
		length = dyn->length;
	}
	while (length > 0) {
		end = IRCD_MIN(db->end, start + length);
		while (start < end && !IsEol(*start)) {
			*buf++ = *start++;
		}

		count = start - db->start;
		if (start < end) {
			*buf = '\0';
			copied += count;
			dbuf_delete(dyn, copied);
			dbuf_flush(dyn);
			return copied;
		}
		if ((db = db->next) == NULL) {
			break;
		}
		copied += count;
		length -= count;
		start = db->start;
	}
	return 0;
}

int dbuf_startfree_block(int server, int doall, int (*callback)(int, int))
{
	DBufBlock *recentdb, *dbb;
	int block_count, numblocks = 0;

	while (1) {
		block_count = 0;
		recentdb = NULL;

		for (dbb = dbufBlocks; dbb != NULL; dbb = dbb->next) {
			if ((server == 1 && dbb->blocksize != DBUF_SIZE_SERVER) || (server == 0 &&
			  dbb->blocksize != DBUF_SIZE_CLIENT) || dbb->is_freeing == 1) {
				dbb = dbb->next;
				continue;
			}
			block_count++;

			if (recentdb == NULL) {
				recentdb = dbb;
			}
			else if (recentdb->block_seq < dbb->block_seq) {
				recentdb = dbb;
			}
		}
		if (block_count <= 1) {
			return numblocks;
		}
		if (recentdb != NULL) {
			recentdb->callback = callback;
			recentdb->is_freeing = 1;
			numblocks++;
		}
		if (doall == 0) {
			return numblocks;
		}
	}
	return 9;
}
