/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/event.c: internal events management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: event.c,v 1.9 2002/11/10 15:25:55 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "h.h"
#include "memory.h"

static Event EventList[MAXEVENTS];
static int event_count = 0;
static const char *last_event_run = NULL;
static time_t event_time_min = -1;

static int findEvent(void (*func)(), void *arg)
{
	int i;
	for (i = 0; event_count > 0 && i < MAXEVENTS; i++) {
		if (EventList[i].active == 1 && EventList[i].func == func && EventList[i].arg == arg) {
			return i;
		}
	}
	return -1;
}

void addEvent(const char *name, void (*func)(), void *arg, time_t when, unsigned loop)
{
	int i;

	if (event_count >= MAXEVENTS) {
		sendto_realops_lev(DEBUG_LEV, "Internal event list is full, event %s not added!", name);
		ircdlog(LOG_DEFAULT, "Internal event list is full, event %s not added!", name);
		return;
	}
	for (i = 0; i < MAXEVENTS; i++) {
		if (EventList[i].active == 0) {
			break;
		}
	}

	EventList[i].name = name;
	EventList[i].func = func;
	EventList[i].arg = arg;
	EventList[i].when = timeofday + when;
	EventList[i].frequency = when;
	EventList[i].loop = loop;
	EventList[i].active = 1;
	event_count++;

	if (EventList[i].when < event_time_min || event_time_min == -1) {
		event_time_min = EventList[i].when;
	}
}

void delEvent(void (*func)(), void *arg)
{
	int i = -1;

	if ((i = findEvent(func, arg)) == -1) {
		return;
	}
	EventList[i].name = NULL;
	EventList[i].func = NULL;
	EventList[i].arg = NULL;
	EventList[i].loop = 0;
	EventList[i].active = 0;
	event_count--;
	return;
}

void checkEvent(void)
{
	int i;
	for (i = 0; event_count > 0 && i < MAXEVENTS; i++) {
		if (EventList[i].active == 1 && EventList[i].when <= timeofday) {
			last_event_run = EventList[i].name;
			EventList[i].func(EventList[i].arg);
			if (EventList[i].loop == 1) {
				EventList[i].when = timeofday + EventList[i].frequency;
			}
			else {
				delEvent(EventList[i].func, EventList[i].arg);
			}
			event_time_min = -1;
		}
	}
	return;
}

time_t next_checkEvent(void)
{
	int i;

	if (event_count == 0) {
		return timeofday + 1;
	}
	else if (event_time_min == -1 && event_count > 0) {
		for (i = 0; i < MAXEVENTS; i++) {
			if (EventList[i].active == 1 && (EventList[i].when < event_time_min ||
			  event_time_min == -1)) {
				event_time_min = EventList[i].when;
			}
		}
	}
	return event_time_min;
}

void show_events(aClient *cptr)
{
	int i;

	if (last_event_run != NULL) {
		sendto_one(cptr, ":%s NOTICE %s :*** The last event to be executed was: %s", me.name,
			cptr->name, last_event_run);
	}
	if (event_count == 0) {
		sendto_one(cptr, ":%s NOTICE %s :*** There are currently no pending events.", me.name,
			cptr->name);
	}
	else {
		int pending_events = 0;
		for (i = 0; i < MAXEVENTS; i++) {
			if (EventList[i].active == 1) {
				sendto_one(cptr, ":%s NOTICE %s :*** %s (type: %s, next execution: %s)", me.name,
					cptr->name, EventList[i].name, EventList[i].loop == 1 ? "continuous" : "once",
					secstodhms(EventList[i].when - timeofday));
				pending_events++;
			}
		}
		sendto_one(cptr, ":%s NOTICE %s :*** Displayed a total of %d pending event%s.", me.name,
			cptr->name, pending_events, pending_events == 1 ? "" : "s");
	}
	return;
}
