/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircd.c: core ircd functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircd.c,v 1.116 2002/12/22 18:11:44 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "memory.h"
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HPUX
#define _KERNEL
#endif
#include <sys/resource.h>
#ifdef HPUX
#undef _KERNEL
#endif
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <fcntl.h>
#include "inet.h"
#include "h.h"
#include "patchlevel.h"
#include "fdlist.h"
#include "setup.h"
#include "isupport.h"
#include "channel.h"
#include "res.h"
#include "ssl.h"
#include "dbuf.h"

extern int conf_fail;

static size_t initial_vmem = 0;
static char *usage[] = {
	" ",
	"RageIRCd v2",
	"Copyright (C) 2002 the past and present IRCd developers, and others.",
	"Refer to doc/authors/ for full credits and copyright details.",
	" ",
	"Usage: ircd [flags]",
	" ",
	"Where [flags] can be any of the following:",
	" ",
#ifdef CMDLINE_CONFIG
	"    -c X   Load X instead of default config file (ircd.conf)",
#endif
	"    -h     Displays usage information (this)",
	"    -t     Run in terminal (doesn't fork to background)",
	"    -v     Print the version to stderr and exit",
	"    -x Y   Run in debug level Y (needs -t)",
	" ",
	NULL
};

char server_options[] = {
#ifdef DEBUGMODE
	'D',
#endif
#ifdef LOG_LIBADNS_DEBUG
	'A',
#endif
#ifdef CLIENT_SERVER
	'C',
#endif
#ifdef IRCII_KLUDGE
	'K',
#endif
#ifdef STATIC_MODULES
	'S',
#endif
#ifdef HAVE_OPENSSL_ON
	'E',
#endif
#ifdef HAVE_POLL_ON
	'P',
#endif
#ifdef HAVE_BLALLOC_ON
	'B',
#endif
#ifdef HAVE_THROTTLE_ON
	'T',
#endif
#ifdef HAVE_HUB_ON
	'H',
#endif
#ifdef INCREASE_SOCK_BUFS
	'I',
#endif
#ifdef USE_NOQUIT
	'N',
#endif
#ifdef FLUD
	'F',
#endif
#ifdef OS_CYGWIN
	'W',
#endif
#ifdef CMDLINE_CONFIG
	'c',
#endif
	'\0'
};

char isupport_one[512];
char isupport_two[512];
char **myargv;
char *ircd_config_file = CONFIG_FILE;
char *debug_mode = "";
int debug_level = -1;
int run_in_terminal = 0;
int do_rehash = 0;
int do_bancheck = 0;
time_t timeofday = 0;
time_t nextconnect = 1;
time_t nextping = 1;
time_t nextevent = -1;
int noisy_htm = NOISY_HTM_DEFAULT;
int lifesux = LIFESUX_DEFAULT;
int LRV = LOADRECV;
int LCF = LOADCFREQ;
int currlife = 0;
int HTMLOCK = 0;

aClient me, *client = &me;
struct Counter Count;

extern void init_signal();
extern void init_signal_profiling();
extern void check_profiling();
#ifndef STATIC_MODULES
extern void init_Modules();
extern void check_Modules();
#endif
extern void build_chmode_string();
extern void build_usermode_string();
extern void init_clients();
extern void init_channels();
extern void init_link();
#ifdef FLUD
extern void init_fludbot();
#endif
#ifdef HAVE_BLALLOC_ON
extern void init_blockheap();
#endif
extern void init_MessageFiles();
extern void init_conf();
extern void run_listeners();
extern void init_fdlist(fdlist *);
#ifdef HAVE_THROTTLE_ON
extern void init_throttle();
extern void throttle_timer(time_t);
#endif

#ifdef RLIMIT_FDMAX
#define RLIMIT_FD_MAX RLIMIT_FDMAX
#else
#ifdef RLIMIT_NOFLE
#define RLIMIT_FD_MAX RLIMIT_NOFILE
#else
#ifdef RLIMIT_OPEN_MAX
#define RLIMIT_FD_MAX RLIMIT_OPEN_MAX
#else
#undef RLIMIT_FD_MAX
#endif
#endif
#endif

static size_t get_vmem()
{
	void *ret = sbrk(0);
	return (size_t)ret;
}

static void display_usage()
{
	char **text;

	fprintf(stderr, "\nircd %s\n", ircd_version);
	for (text = usage; *text != NULL; text++) {
		fprintf(stderr, "%s\n", *text);
	}
	return;
}

static void open_debugfile()
{
#ifdef DEBUGMODE
	int fd = -1;
	aClient *cptr;
	char *terminal = NULL;

	if (debug_level >= 0) {
		cptr = make_client(NULL, NULL);
		cptr->fd = -2;
		SetLog(cptr);
		cptr->port = debug_level;
		cptr->flags = 0;
		cptr->listener = NULL;
		local[2] = cptr;
		strcpy(cptr->sockhost, me.sockhost);
		fprintf(stderr, "isatty = %d ttyname = %#x\n", isatty(2), (u_int)ttyname(2));

		if (run_in_terminal == 0) {
			truncate(LOG_DEBUG, 0);
			if ((fd = open(LOG_DEBUG, O_CREAT|O_WRONLY, 0600)) < 0) {
				if ((fd = open("/dev/null", O_WRONLY)) < 0) {
					fprintf(stderr, "ERROR: failed to open %s or /dev/null.", LOG_DEBUG);
					exit(-1);
				}
			}
			if (fd != 2) {
				dup2(fd, 2);
				close(fd);
			}
			strncpyzt(cptr->name, LOG_DEBUG, sizeof(cptr->name));
		}
		else if (isatty(2) == 1 && (terminal = ttyname(2)) != NULL) {
			strncpyzt(cptr->name, terminal, sizeof(cptr->name));
		}
		else {
			strcpy(cptr->name, "FD2-Pipe");
		}
		Debug((DEBUG_FATAL, "Debug: file %s level %d at %s", cptr->name, cptr->port, myctime(time(NULL))));
	}
	else {
		local[2] = NULL;
	}
#endif
	return;
}

static void save_statsfile()
{
	FILE *fd = NULL;
	if ((fd = fopen(STATS_FILE, "w+")) != NULL) {
		fprintf(fd, "%d %d %ld %ld %ld %ld %ld %ld %ld", Count.max_loc, Count.max_tot,
			Count.weekly, Count.monthly, Count.yearly, Count.start, Count.week,
			Count.month, Count.year);
		fclose(fd);
	}
	return;
}

static void init_statsfile()
{
	FILE *fd = NULL;
	char buf[1024], *item, *p;
	int cnt = 0;

	if (STATSFILE_SAVE_FREQ > 0) {
		addEvent("save_statsfile", save_statsfile, NULL, STATSFILE_SAVE_FREQ, 1);
	}
	if ((fd = fopen(STATS_FILE, "r")) != NULL) {
		fgets(buf, 1024, fd);
		if (*buf != '\0') {
			for (item = strtoken(&p, buf, " "); item != NULL;
			  item = strtoken(&p, NULL, " "), cnt++) {
				switch (cnt) {
					case 0:
						Count.max_loc = atoi(item);
						break;
					case 1:
						Count.max_tot = atoi(item);
						break;
					case 2:
						Count.weekly = atol(item);
						break;
					case 3:
						Count.monthly = atol(item);
						break;
					case 4:
						Count.yearly = atol(item);
						break;
					case 5:
						Count.start = atol(item);
						break;
					case 6:
						Count.week = atol(item);
						break;
					case 7:
						Count.month = atol(item);
						break;
					case 8:
						Count.year = atol(item);
						break;
					default:
						ircdlog(LOG_ERROR, "init_statsfile(): ignoring extra "
							"parameter (cnt %d) in %s", cnt, STATS_FILE);
				}
			}
			fprintf(stderr, "Loaded IRC statistics file.\n");
		}
		fclose(fd);
	}
	return;
}

void set_time()
{
	if ((timeofday = time(NULL)) == -1) {
		ircdlog(LOG_ERROR, "Clock failure (%d), TS may become corrupt!", errno);
		sendto_ops("Clock failure (%d), TS may become corrupt!", errno);
	}
	return;
}

void write_pidfile()
{
	int fd = -1;
	char buf[20];

	if ((fd = open(PID_FILE, O_CREAT|O_WRONLY, 0600)) == -1) {
		Debug((DEBUG_NOTICE, "Error opening pid file %s", PID_FILE));
		ircdlog(LOG_ERROR, "Error opening pid file %s", PID_FILE);
	}
	else {
		int len = ircsprintf(buf, "%5d\n", (int)getpid());
		if (write(fd, buf, len) == -1) {
			Debug((DEBUG_NOTICE, "Error writing to pid file %s", PID_FILE));
			ircdlog(LOG_ERROR, "Error opening pid file %s", PID_FILE);
		}
		close(fd);
	}
	return;
}

static int check_pid()
{
	char buf[20];
	pid_t pid = 0;
	int fd, retval = 0;

	if ((fd = open(PID_FILE, O_RDONLY)) >= 0) {
		if (read(fd, buf, sizeof(buf)) > 0) {
			if ((pid = atoi(buf)) != (int)getpid() && !kill(pid, 0)) {
				retval = (int)pid;
			}
		}
		close(fd);
	}
	else if (errno != ENOENT) {
		fprintf(stderr, "WARNING: problem opening %s: %s\n", PID_FILE, strerror(errno));
	}
	return retval;
}

static void init_daemon()
{
#if !defined(SOL20) && !defined(OS_CYGWIN)
	setlinebuf(stderr);
#endif
	if (debug_level == -1) {
		close(2);
	}
	close(1);

	if (isatty(0)) {
		int pid;
#ifdef TIOCNOTTY
		int fd;
#endif

		if ((pid = fork()) < 0) {
			fprintf(stderr, "ERROR: couldn't fork (%s)\n", strerror(errno));
			exit(0);
		}
		else if (pid > 0) {
			exit(0);
		}
#ifdef TIOCNOTTY
		if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
			ioctl(fd, TICONOTTY, NULL);
			close(fd);
		}
#endif
		setsid();
		close(0);
		local[0] = NULL;
	}
	return;
}

void restart(char *message)
{
	static int been_here = 0;
	if (been_here == 0) {
		ircdlog(LOG_DEFAULT, "Server restarting: %s (%d)", message, get_vmem() - initial_vmem);
		been_here = 1;
		server_reboot();
	}
	return;
}

void server_reboot()
{
	char *errmsg;
	int i;

	sendto_ops("Aieeeeee! Restarting server... (%d)", get_vmem() - initial_vmem);
	Debug((DEBUG_DEBUG, "Restarting server..."));
	dump_connections(me.fd);

	for (i = 0; i < MAXCONNECTIONS; i++) {
		close(i);
	}

	execv(IRCD_EXE, myargv);

	errmsg = strerror(errno);
	ircdlog(LOG_ERROR, "FATAL: execv(%s,%s) failed (%s)", IRCD_EXE, myargv[0], errmsg);
	Debug((DEBUG_FATAL, "Couldn't restart server: %s", errmsg));
#ifndef STATIC_MODULES
	unload_all_modules();
#endif
	exit(-1);
}

static time_t try_connections(time_t currenttime)
{
	ConfigItem_link *link;
	ConfigItem_ban *ban;
	aClient *cptr;
	int connectfreq;
	time_t next = 0;

	Debug((DEBUG_NOTICE, "Connection check at: %s", myctime(currenttime)));
	for (link = ci_link; link != NULL; link = (ConfigItem_link *)link->next) {
		if (!(link->options & LINK_AUTOCONNECT)) {
			continue;
		}
		if (link->hold > currenttime) {
			if (next > link->hold || next == 0) {
				next = link->hold;
			}
			continue;
		}

		connectfreq = link->class->connectfreq;
		link->hold = currenttime + connectfreq;

		if ((cptr = find_name(link->servername, (aClient *)NULL)) == NULL &&
		  link->class->clients < link->class->maxclients) {
			if ((ban = find_ban(BAN_SERVER, link->servername)) != NULL) {
				sendto_gnotice("Passing autoconnect for banned server %s: %s.",
					link->servername, ban->reason);
				continue;
			}
			if (connect_server(link, NULL, NULL)) {
				sendto_gnotice("Connection to %s[%s] activated.", link->servername,
					link->fromip);
			}
		}
		if (next > link->hold || next == 0) {
			next = link->hold;
		}
	}
	Debug((DEBUG_NOTICE, "Next connection check at: %s", myctime(next)));
	return next;
}

static time_t check_pings(time_t currenttime)
{
	aClient *cptr;
	time_t oldest = 0;
	int ping = 0, i, break_ban_loop;
	ConfigItem_ban *ban;
	char *userhost, ping_time_out_buffer[64];

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsMe(cptr) || IsLog(cptr)) {
			continue;
		}
		if (cptr->flags & FLAGS_DEADSOCKET) {
			exit_client(cptr, cptr, &me, (cptr->flags & FLAGS_SENDQEX) ? "SendQ exceeded" : "Dead Socket");
			i--;
			continue;
		}
		if (do_bancheck == 1 && IsPerson(cptr)) {
			break_ban_loop = 0;
			userhost = make_user_host(cptr->username, cptr->host);
			for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)ban->next) {
				if (ban->type & BAN_USERHOST) {
					if (!match(ban->mask, userhost)) {
						if (find_except(userhost) == NULL) {
							conf_doban(cptr, cptr, ban, NULL, -1);
							break_ban_loop = 1;
							break;
						}
					}
				}
				if (ban->type & BAN_IP) {
					if (!match(ban->mask, cptr->hostip)) {
						if (find_except(cptr->hostip) == NULL) {
							conf_doban(cptr, cptr, ban, NULL, -1);
							break_ban_loop = 1;
							break;
						}
					}
				}
				if (ban->type & BAN_GCOS) {
					if (!match(ban->mask, cptr->info)) {
						if (find_except(cptr->info) == NULL) {
							conf_doban(cptr, cptr, ban, NULL, -1);
							break_ban_loop = 1;
							break;
						}
					}
				}
			}
			if (break_ban_loop == 1) {
				i--;
				continue;
			}
		}

		ping = IsRegistered(cptr) ? cptr->pingval : CONNECTTIMEOUT;
		if (ping < currenttime - cptr->lasttime) {
			if ((cptr->flags & FLAGS_PINGSENT && currenttime - cptr->lasttime >= ping * 2) ||
			  (!IsRegistered(cptr) && currenttime - cptr->since >= ping)) {
				if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr)) {
					char *path = get_client_name(cptr, HIDEME);
					sendto_gnotice("No responce from %s, closing link.", path);
					sendto_serv_butone(cptr, ":%s GNOTICE :No responce from %s, closing link.",
						me.name, path);
				}
				ircsprintf(ping_time_out_buffer, "Ping timeout: %d seconds",
					currenttime - cptr->lasttime);
				exit_client(cptr, cptr, &me, ping_time_out_buffer);
				i--;
				continue;
			}
			else if (!(cptr->flags & FLAGS_PINGSENT)) {
				if (!(IsServer(cptr) && (cptr->serv->sflags & SFLAGS_BURST))) {
					cptr->flags |= FLAGS_PINGSENT;
					cptr->lasttime = currenttime - ping;
					sendto_one(cptr, "PING :%s", me.name);
				}
			}
		}
		if (IsUnknown(cptr)) {
			if (cptr->firsttime ? (timeofday - cptr->firsttime) > 100 : 0) {
				exit_client(cptr, cptr, &me, "Connection timed out");
			}
		}
	}
	do_bancheck = 0;
	oldest = currenttime + 10;
	Debug((DEBUG_NOTICE, "Next check_ping() call at: %s, %d %d %d", myctime(oldest), ping,
		oldest, currenttime));
	return oldest;
}

static void parse_args(int argc, char *argv[])
{
	extern char *optarg;
	int i = -1;

	while ((i = getopt(argc, argv, "c:tvx:")) != -1) {
		switch (i) {
			case 'c':
#ifdef CMDLINE_CONFIG
				ircd_config_file = optarg;
#else
				fprintf(stderr, "ERROR: the -c option is not enabled.\n");
				exit(0);
#endif
				break;
			case 't':
				run_in_terminal = 1;
				break;
			case 'v':
				fprintf(stderr, "ircd %s\n", ircd_version);
				exit(0);
				break;
			case 'x':
#ifdef DEBUGMODE
				if (optarg != NULL) {
					debug_level = atoi(optarg);
					debug_mode = optarg;
				}
#else
				fprintf(stderr, "WARNING: DEBUGMODE must be defined for -x (ignoring)\n");
#endif
				break;
			default:
				display_usage();
				exit(0);
		}
	}
	return;
}

static void build_version()
{
	char *s = PATCHES;
	ircsprintf(ircd_version, "%s(%s)-%d.%d.%d%s%s%s", BASENAME, CODENAME, MAJOR, MINOR, REVISION,
		*s != '\0' ? PATCHES : "",
#ifdef INCLUDE_BUILD_REVISION
	".build:", build);
#else
	"", "");
#endif
	return;
}

void build_isupport()
{
	ircsprintf(isupport_one, ISUPPORT_ONE_FEATURES, ISUPPORT_ONE_PARAMS);
	ircsprintf(isupport_two, ISUPPORT_TWO_FEATURES, ISUPPORT_TWO_PARAMS);
	return;
}

static void init_core()
{
#ifdef RLIMIT_CORE
	struct rlimit ircd_lim;
	if (!getrlimit(RLIMIT_CORE, &ircd_lim)) {
		ircd_lim.rlim_cur = ircd_lim.rlim_max;
		if (setrlimit(RLIMIT_CORE, &ircd_lim)) {
			fprintf(stderr, "WARNING: unlimit core size failed (%s)\n", strerror(errno));
		}
	}
#endif
	return;
}

static void init_vmem()
{
	initial_vmem = get_vmem();
	return;
}

static void init_counter()
{
	Count.server = 1;
	Count.oper = 0;
	Count.chan = 0;
	Count.local = 0;
	Count.total = 0;
	Count.invisi = 0;
	Count.unknown = 0;
	Count.max_loc = 0;
	Count.max_tot = 0;
	Count.cli_restart = 0;
	Count.today = 0;
	Count.weekly = 0;
	Count.monthly = 0;
	Count.yearly = 0;
	Count.start = timeofday;
	Count.day = timeofday;
	Count.week = timeofday;
	Count.month = timeofday;
	Count.year = timeofday;
	return;
}

static void init_hash_tables()
{
	clear_client_hash_table();
	clear_channel_hash_table();
	clear_scache_hash_table();
	clear_ip_hash_table();
	return;
}

static void init_memory_heaps()
{
	init_hash_tables();
	init_clients();
	init_channels();
	init_link();
#ifdef HAVE_THROTTLE_ON
	init_throttle();
#endif
#ifdef FLUD
	init_fludbot();
#endif
#ifdef HAVE_BLALLOC_ON
	init_blockheap();
#endif
#ifndef STATIC_MODULES
	init_whowas();
#endif
	return;
}

static void init_fdlists()
{
	int i;

	init_fdlist(&serv_fdlist);
#ifndef NO_PRIORITY
	init_fdlist(&busycli_fdlist);
#endif
	init_fdlist(&default_fdlist);
	init_fdlist(&oper_fdlist);
	init_fdlist(&listen_fdlist);

	for (i = MAXCONNECTIONS + 1; i > 0; i--) {
		default_fdlist.entry[i] = i - 1;
	}
	return;
}

static void init_maximumfds()
{
#ifdef RLIMIT_FD_MAX
	struct rlimit limit;

	if (!getrlimit(RLIMIT_FD_MAX, &limit)) {
		if (limit.rlim_max < MAXCONNECTIONS) {
			fprintf(stderr, "ERROR: ircd fd table is too big, please fix MAXCONNECTIONS!\n");
			fprintf(stderr, "       system fd limit: %ld current MAXCONNECTIONS: %d\n",
				(long)limit.rlim_max, MAXCONNECTIONS);
			exit(-1);
		}

		limit.rlim_cur = limit.rlim_max;
		if (setrlimit(RLIMIT_FD_MAX, &limit) == -1) {
			fprintf(stderr, "ERROR: failed to set max fd limit to %ld\n",
				(long)limit.rlim_max);
			exit(-1);
		}

#ifndef HAVE_POLL_ON
		if (MAXCONNECTIONS > FD_SETSIZE) {
			fprintf(stderr, "ERROR: FD_SETSIZE: %d MAXCONNECTIONS: %d\n", FD_SETSIZE,
				MAXCONNECTIONS);
			fprintf(stderr, "       Make sure your system supports a large FD_SETSIZE,\n");
			fprintf(stderr, "       then reconfigure with a max fd limit of %ld.\n",
				MAXCONNECTIONS);
			exit(-1);
		}
#endif
#ifdef HAVE_FD_ALLOC
		read_set = FD_ALLOC(MAXCONNECTIONS);
		write_set = FD_ALLOC(MAXCONNECTIONS);
#endif
	}
#endif
	return;
}

static time_t io_loop(time_t delay)
{
	static char to_send[200];
#ifndef HAVE_HUB_ON
	static time_t lasttime = 0;
	static long lastrecvK = 0;
	static int lrv = 0;
#endif
	time_t lasttimeofday = timeofday;

	set_time();
	if (timeofday < lasttimeofday) {
		ircsprintf(to_send, "System clock is running backwards: %d < %d", timeofday,
			lasttimeofday);
		report_error(to_send, &me);
	}
	if (nextevent == -1) {
		nextevent = next_checkEvent();
	}
	if (nextevent <= timeofday) {
		checkEvent();
		nextevent = next_checkEvent();
	}

#ifndef HAVE_HUB_ON
	if (timeofday - lasttime >= LCF) {
		lrv = LRV * LCF;
		lasttime = timeofday;
		currlife = (me.receiveK - lastrecvK) / LCF;

		if (me.receiveK - lrv > lastrecvK || HTMLOCK == 1) {
			if (lifesux == 0) {
				lifesux = 1;
				if (noisy_htm == 1) {
					sendto_ops("Entering high-traffic mode (%dk/s > %dk/s)", currlife, LRV);
				}
			}
			else {
				lifesux++;
				LCF += 2;
				if (noisy_htm == 1) {
					sendto_ops("Still in high-traffic mode %d%s (%d delay): %dk/s",
						lifesux, (lifesux > 9) ? " (TURBO)" : "", LCF, currlife);
				}
				if (lifesux > 15) {
					if (noisy_htm == 1) {
						sendto_ops("Resetting HTM and raisin glimit to: %dk/s", LRV + 5);
					}
					LCF = LOADCFREQ;
					lifesux = 0;
					LRV += 5;
				}
			}
		}
		else {
			LCF = LOADCFREQ;
			if (lifesux > 0) {
				lifesux = 0;
				if (noisy_htm == 1) {
					sendto_ops("Resuming standard operation...");
				}
			}
		}
		lastrecvK = me.receiveK;
	}
#endif

	if (nextconnect != 0 && timeofday >= nextconnect) {
		nextconnect = try_connections(timeofday);
	}
	if (nextconnect != 0) {
		delay = IRCD_MIN(nextping, nextconnect);
	}
	else {
		delay = nextping;
	}
	delay -= timeofday;

	if (delay < 1) {
		delay = 1;
	}
	else {
		delay = IRCD_MIN(delay, TIMESEC);
	}

#ifndef NO_PRIORITY
	read_message(0, &serv_fdlist);
	read_message(0, &busycli_fdlist);

	if (lifesux > 0) {
		read_message(1, &serv_fdlist);
		if (lifesux > 9) {
			read_message(1, &busycli_fdlist);
			read_message(1, &serv_fdlist);
		}
		flush_fdlist_connections(&serv_fdlist);
	}

	set_time();
	{
		static time_t lslasttime = 0;

#ifdef CLIENT_SERVER
		if (lifesux == 0 || (lslasttime + lifesux) < timeofday) {
#else
		if (lslasttime + lifesux + 1 < timeofday) {
#endif
			read_message(delay > 0 ? delay : 1, NULL);
			lslasttime = timeofday;
		}
	}
#else
	read_message(delay, NULL);
#endif

	if (timeofday >= nextping) {
		nextping = check_pings(timeofday);
		connauth_timeout_queries(timeofday);
	}

	check_profiling();

	if (do_rehash == 1 && lifesux == 0) {
		rehash(1);
		do_rehash = 0;
	}

	flush_connections(me.fd);

#ifndef NO_PRIORITY
	check_fdlists();
#endif
#ifdef HAVE_THROTTLE_ON
	throttle_timer(timeofday);
#endif
	return delay;
}

int main(int argc, char *argv[])
{
	char *line_break = "-----------------------------------------------------------------------------\n";
#ifdef HAVE_OPENSSL_ON
	char *ssl_error = NULL;
#endif
	int i = 0;
	time_t delay = 0;

	if (!geteuid()) {
		fprintf(stderr, "ERROR: never run ircd as root.\n");
		return -1;
	}
	if ((timeofday = time(NULL)) == -1) {
		fprintf(stderr, "ERROR: clock failure (%s)\n", strerror(errno));
		exit(errno);
	}

	init_core();
	init_vmem();
	init_counter();
	init_dbuf();
	init_signal_profiling();
	init_signal();
	build_version();

	memset((char *)&me, '\0', sizeof(me));
	me.fd = -1;
	myargv = argv;
	umask(077);
	parse_args(argc, argv);

	if (chdir(IRCD_PREFIX) < 0) {
		perror("chdir");
		fprintf(stderr, "ERROR: please ensure that your installation directory (IRCD_PREFIX)\n");
		fprintf(stderr, "       exists and that it is accessible.\n\n");
		fprintf(stderr, "       Current IRCD_PREFIX: %s\n", IRCD_PREFIX);
		exit(-1);
	}
	if ((i = check_pid()) > 0) {
		fprintf(stderr, "ERROR: ircd already running under process id %d.\n", i);
		exit(-1);
	}
#ifdef HAVE_OPENSSL_ON
	if (init_dh() == -1) {
		fprintf(stderr, "ERROR: failed to load SSL encryption layer.\n");
		exit(-1);
	}
#endif

	init_HookTable();
	init_CommandHash();
#ifndef STATIC_MODULES
	init_Modules();
#endif
	init_MessageFiles();
	init_memory_heaps();
	init_stats();
	init_send();
	open_debugfile();
	init_fdlists();

	fprintf(stderr, line_break);
	init_conf();
#ifndef STATIC_MODULES
	check_Modules();
#endif

	if (conf_fail == 0) {
#ifdef HAVE_OPENSSL_ON
		if ((ssl_error = init_ssl()) != NULL) {
			report(1, "%s", ssl_error);
		}
#endif
		if (conf_fail == 0) {
			init_statsfile();
		}
	}
	if (conf_fail > 0) {
		fprintf(stderr, "\nReceived %d fatal errors during server initialisation, terminating program.\n%s",
			conf_fail, line_break);
		exit(-1);
	}

	set_time();
	init_maximumfds();
	if (!init_resolver()) {
		fprintf(stderr, "ERROR: failed to load nameservers from %s: %s.\n",
			RESOLV_FILE, strerror(errno));
		exit(-1);
	}

	build_chmode_string();
	build_usermode_string();
	build_isupport();

	fprintf(stderr, "\nServer initialization successfully completed, attempting to boot.\n%s", line_break);

	strncpyzt(me.name, ci_servinfo->name, sizeof(me.name));
	strncpyzt(me.info, ci_servinfo->desc, sizeof(me.info));
	me.ip.s_addr = (!BadPtr(ci_servinfo->bindip)) ? inet_addr(ci_servinfo->bindip) : INADDR_ANY;

	set_non_blocking(me.fd);
	me.hopcount = 0;
	me.confs = NULL;
	me.next = NULL;
	me.user = NULL;
	me.from = NULL;
	SetMe(&me);
	make_server(&me);
	me.serv->up = me.name;
	me.lasttime = me.since = me.firsttime = timeofday;
	add_to_client_hash_table(me.name, &me);

	if (run_in_terminal == 0) {
		init_daemon();
	}
	write_pidfile();
	run_listeners();

	Debug((DEBUG_NOTICE, "Server ready..."));

#ifndef NO_PRIORITY
	check_fdlists();
#endif
	set_time();
	
	while (1) {
		usleep(100000);
		process_dns_queries();
		delay = io_loop(delay);
		process_dns_queries();
	}
	return 0;
}

void ircdlog(char *filename, char *pattern, ...)
{
	FILE *fd;
	char buffer[1024];
	va_list vl;

	if (LOGGING == 1) {
		if ((fd = fopen(filename, "a")) == NULL) {
			Debug((DEBUG_ERROR, "Failed to open/create %s for log", filename));
			return;
		}
		va_start(vl, pattern);
		ircvsprintf(buffer, pattern, vl);
		va_end(vl);
		fprintf(fd, "[%s] %s\n", myctime(timeofday), buffer);
		fclose(fd);
	}
	return;
}
