/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/list.c: linked list management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: list.c,v 1.34 2002/11/23 20:45:18 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "blalloc.h"
#include "zlink.h"
#include "setup.h"
#include "memory.h"

int numclients = 0;
BlockHeap *slink_heap;

void init_link()
{
	slink_heap = BlockHeapCreate((size_t)sizeof(Link), LINKS_PREALLOCATE);
	return;
}

aClient *make_client(aClient *from, aClient *uplink)
{
	aClient *cptr = NULL;

	if (from == NULL) {
		if ((cptr = (aClient *)BlockHeapAlloc(local_client_heap)) == NULL) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_client where from is NULL");
			outofmemory();
		}
		memset((char *)cptr, '\0', CLIENT_LOCAL_SIZE);
		cptr->from = cptr;
		cptr->status = STAT_UNKNOWN;
		cptr->fd = -1;
		cptr->sockerr = -1;
		cptr->uplink = uplink;
		cptr->class = (ConfigItem_class *)NULL;
		strcpy(cptr->username, "unknown");
		cptr->since = cptr->lasttime = cptr->firsttime = timeofday;
		cptr->sendqlen = MAXSENDQLENGTH;
	}
	else {
		if ((cptr = (aClient *)BlockHeapAlloc(remote_client_heap)) == NULL) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_client where from is non-NULL");
			outofmemory();
		}
		memset((char *)cptr, '\0', CLIENT_REMOTE_SIZE);
		cptr->from = from;
		cptr->status = STAT_UNKNOWN;
		cptr->fd = -1;
		cptr->uplink = uplink;
	}
	return cptr;
}

void free_client(aClient *cptr)
{
	if (BlockHeapFree(cptr->fd == -2 ? local_client_heap : remote_client_heap, cptr)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_client (cptr=%lX)", cptr);
	}
	return;
}

aChannel *make_channel()
{
	aChannel *chan = NULL;
	if ((chan = (aChannel *)BlockHeapAlloc(channel_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_channel");
		outofmemory();
	}
	memset((char *)chan, '\0', sizeof(aChannel));
	return chan;
}

void free_channel(aChannel *chptr)
{
	if (BlockHeapFree(channel_heap, chptr)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_channel [chptr=%lX]", chptr);
	}
	return;
}

aBan *make_channelban()
{
	aBan *ban = NULL;
	if ((ban = (aBan *)BlockHeapAlloc(channelban_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_channelban");
		outofmemory();
	}
	memset((char *)ban, '\0', sizeof(aBan));
	return ban;
}

void free_channelban(aBan *ban)
{
	if (BlockHeapFree(channelban_heap, ban)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_channelban [ban=%lX]", ban);
	}
	return;
}

anUser *make_user(aClient *cptr)
{
	anUser *user = NULL;
	int local = (cptr->fd == -2 || cptr->fd >= 0) ? 1 : 0;

	if ((user = cptr->user) == NULL) {
		if ((user = (anUser *)BlockHeapAlloc((local == 1) ? local_user_heap : remote_user_heap)) == NULL) {
				ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_user for %s "
					"client [cptr=%lX]", (local == 1) ? "local" : "remote", cptr);
				outofmemory();
		}
		memset((char *)user, '\0', (local == 1) ? USER_LOCAL_SIZE : USER_REMOTE_SIZE);
		cptr->user = user;
	}
	return user;
}

aServer *make_server(aClient *cptr)
{
	aServer *serv = NULL;
	if ((serv = cptr->serv) == NULL) {
		serv = (aServer *)MyMallocX(sizeof(aServer));
		cptr->serv = serv;
	}
	return serv;
}

void free_user(anUser *user, aClient *cptr)
{
	int local = (cptr->fd == -2 || cptr->fd >= 0) ? 1 : 0;

	if (user->away != NULL) {
		MyFree(user->away);
	}
	if (user->maskedhost != NULL) {
		MyFree(user->maskedhost);
	}
	if (user->joined != 0 || user->invited != NULL || user->channel != NULL) {
		ircdlog(LOG_ERROR, "free_user() encountered unparted/unjoined/uninvited client %s!%s@%s "
			"%#x %#x %#x %d", cptr ? cptr->name : "<noname>", cptr->username, cptr->host, user,
			user->invited, user->channel, user->joined);
	}
	if (BlockHeapFree((local == 1) ? local_user_heap : remote_user_heap, user)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_user for %s client [user=%lX]",
			(local == 1) ? "local" : "remote", user);
	}
	return;
}

void remove_client_from_list(aClient *cptr)
{
	if (IsServer(cptr)) {
		Count.server--;
	}
	else if (IsClient(cptr)) {
		Count.total--;
		if (HasMode(cptr, UMODE_OPER)) {
			Count.oper--;
		}
		if (HasMode(cptr, UMODE_INVISIBLE)) {
			Count.invisi--;
		}
	}
	if (cptr->prev != NULL) {
		cptr->prev->next = cptr->next;
	}
	else {
		client = cptr->next;
		client->prev = NULL;
	}
	if (cptr->next != NULL) {
		cptr->next->prev = cptr->prev;
	}
	if (IsPerson(cptr) && cptr->user != NULL) {
		add_history(cptr, 0);
		off_history(cptr);
	}
#ifdef FLUD
	if (MyFludConnect(cptr)) {
		free_fluders(cptr, NULL);
	}
	free_fludees(cptr);
#endif
	if (cptr->user != NULL) {
		free_user(cptr->user, cptr);
	}
	if (cptr->serv != NULL) {
#ifdef HAVE_OPENSSL_ON
		if (cptr->serv->sessioninfo_in != NULL) {
			dh_end_session(cptr->serv->sessioninfo_in);
		}
		if (cptr->serv->sessioninfo_out != NULL) {
			dh_end_session(cptr->serv->sessioninfo_out);
		}
		if (cptr->serv->rc4_in != NULL) {
			rc4_destroystate(cptr->serv->rc4_in);
		}
		if (cptr->serv->rc4_out != NULL) {
			rc4_destroystate(cptr->serv->rc4_out);
		}
#endif
		if (cptr->serv->zip_in != NULL) {
			zip_destroy_input_session(cptr->serv->zip_in);
		}
		if (cptr->serv->zip_out != NULL) {
			zip_destroy_output_session(cptr->serv->zip_out);
		}
		MyFree(cptr->serv);
	}
	free_client(cptr);
	return;
}

void add_client_to_list(aClient *cptr)
{
	cptr->next = client;
	client = cptr;
	if (cptr->next != NULL) {
		cptr->next->prev = cptr;
	}
	return;
}

chanMember *find_user_member(chanMember *cm, aChannel *chptr)
{
	if (chptr != NULL) {
		while (cm != NULL) {
			if (cm->chptr == chptr) {
				return cm;
			}
			cm = cm->nextchan;
		}
	}
	return NULL;
}

Link *find_str_link(Link *lp, char *charptr)
{
	if (charptr != NULL) {
		for (; lp != NULL; lp = lp->next) {
			if (!match(lp->value.cp, charptr)) {
				return lp;
			}
		}
	}
	return NULL;
}

Link *make_slink()
{
	Link *lp = NULL;
	if ((lp = (Link *)BlockHeapAlloc(slink_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_slink");
		outofmemory();
	}
	lp->next = NULL;
	return lp;
}

void free_slink(Link *lp)
{
	if (BlockHeapFree(slink_heap, lp)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_slink [lp=%lX]", lp);
	}
	return;
}

chanMember *make_chanmember()
{
	chanMember *cm = NULL;
	if ((cm = (chanMember *)BlockHeapAlloc(chanmember_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_chanmember");
		outofmemory();
	}
	return cm;
}

void free_chanmember(chanMember *cm)
{
	if (BlockHeapFree(chanmember_heap, cm)) {
		ircdlog(LOG_ERROR, "BlockHEapFree failed in free_chanmember [cm=%lX]", cm);
	}
	return;
}

#ifdef HAVE_THROTTLE_ON
throttle *make_throttle()
{
	throttle *tp = NULL;
	if ((tp = (throttle *)BlockHeapAlloc(throttle_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_throttle");
		outofmemory();
	}
	return tp;
}

void free_throttle(throttle *tp)
{
	if (BlockHeapFree(throttle_heap, tp)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_throttle [tp=%lX]", tp);
	}
	return;
}

hashent *make_hashent()
{
	hashent *h = NULL;
	if ((h = (hashent *)BlockHeapAlloc(hashent_heap)) == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_hashent");
		outofmemory();
	}
	return h;
}

void free_hashent(hashent *h)
{
	if (BlockHeapFree(hashent_heap, h)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_hashent [h=%lX]", h);
	}
	return;
}
#endif
