/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/listener.c: listener management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: listener.c,v 1.11 2002/11/16 21:12:28 rageircd Exp $
 */

#include "memory.h"
#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "ssl.h"
#include "sock.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include "inet.h"
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <utmp.h>
#include <sys/resource.h>
#include <time.h>
#include "h.h"
#include "fdlist.h"

#ifdef SOMAXCONN
#undef HYBRID_SOMAXCONN
#define HYBRID_SOMAXCONN SOMAXCONN
#endif

Listener *listener_list;

extern int run_in_terminal;

char *get_listener_name(Listener *listener)
{
	static char buf[HOSTLEN + HOSTIPLEN + 10];
	ircsprintf(buf, "%s[%s/%d]", me.name, listener->ip, listener->port);
	return buf;
}

void report_listener_error(char *text, Listener *listener)
{
	int errtmp = errno, err, len = sizeof(err);
	char *host = get_listener_name(listener);

	Debug((DEBUG_ERROR, text, host, strerror(errtmp)));
#ifdef SO_ERROR
	if (listener->fd >= 0) {
		if (getsockopt(listener->fd, SOL_SOCKET, SO_ERROR, (char *)&err, &len) == 0) {
			if (err != 0) {
				errtmp = err;
			}
		}
	}
#endif
	sendto_realops_lev(DEBUG_LEV, text, host, strerror(errtmp));
	if (run_in_terminal == 1) {
		fprintf(stderr, text, host, strerror(errtmp));
		fprintf(stderr, "\n");
	}
	return;
}

int add_listener(ConfigItem_bind *b)
{
	Listener ltmp, *listener;
	struct sockaddr_in server;
	int len = sizeof(server), opt = 1;

	memset(&ltmp, '\0', sizeof(Listener));
	ltmp.ipaddr.s_addr = inet_addr(b->ip);
	DupString(ltmp.ip, b->ip);
	ltmp.port = b->port;
	if ((ltmp.fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		if (errno == EAGAIN) {
			sendto_realops("Error opening stream socket for listener %d: no more sockets", b->port);
		}
		else {
			report_listener_error("opening stream socket %s:%s", &ltmp);
		}
		MyFree(ltmp.ip);
		return -1;
	}
	if (ltmp.fd >= HARD_FDLIMIT - 10) {
		sendto_realops("No more listeners allowed (%d)", ltmp.bind->port);
		close(ltmp.fd);
		MyFree(ltmp.ip);
		return -1;
	}
	if (setsockopt(ltmp.fd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt))) {
		report_listener_error("setting SO_REUSEADDR socket opt on %s:%s", &ltmp);
		close(ltmp.fd);
		MyFree(ltmp.ip);
		return -1;
	}

	memset(&server, '\0', len);
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = (ltmp.ipaddr.s_addr != 0) ? ltmp.ipaddr.s_addr : INADDR_ANY;
	server.sin_port = htons(b->port);

	if (bind(ltmp.fd, (struct sockaddr *)&server, sizeof(struct sockaddr_in))) {
		report_listener_error("binding stream socket %s:%s", &ltmp);
		close(ltmp.fd);
		MyFree(ltmp.ip);
		return -1;
	}
	if (getsockname(ltmp.fd, (struct sockaddr *)&server, &len)) {
		report_listener_error("getsockname failed for %s:%s", &ltmp);
		close(ltmp.fd);
		MyFree(ltmp.ip);
		return -1;
	}
	if (ltmp.fd > highest_fd) {
		highest_fd = ltmp.fd;
	}

	listen(ltmp.fd, HYBRID_SOMAXCONN);
	listener = (Listener *)MyMalloc(sizeof(Listener));
	memcpy(listener, &ltmp, sizeof(Listener));

	if (local[listener->fd] != NULL) {
		report_listener_error("listener fd held by client in local[] array %s:%s", &ltmp);
		ircdlog(LOG_ERROR, "FATAL: listener fd held by client in local[] array");
		exit(-1);
	}

	set_non_blocking(listener->fd);
	listener->options = b->options;
	listener->bind = b;
	listener->next = listener_list;
	listener_list = listener;
	listener->lasttime = timeofday;
	return 0;
}

void close_one_listener(Listener *listener, Listener *prev)
{
	close(listener->fd);
	if (prev != NULL) {
		prev->next = listener->next;
	}
	else {
		listener_list = listener->next;
	}
	MyFree(listener);
	return;
}

void close_listeners()
{
	Listener *l, *next = NULL, *prev = NULL;
	for (l = listener_list; l != NULL; l = next) {
		next = l->next;
		if (l->bind == NULL) {
			sendto_realops_lev(DEBUG_LEV, "close_listeners() bind pointer is NULL - fix me!");
		}
		else if (l->bind->flag.temporary == 1 && l->bind->clients == 0) {
			close_one_listener(l, prev);
		}
		prev = l;
	}
	return;
}

Listener *find_listener(char *ip, int port)
{
	Listener *l;
	for (l = listener_list; l != NULL; l = l->next) {
		if (!mycmp(l->ip, ip) && l->port == port) {
			return l;
		}
	}
	return NULL;
}
