/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/modules.c: dynamic module routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: modules.c,v 1.64 2002/12/22 18:11:45 rageircd Exp $
 */

#include "struct.h"
#include "setup.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "send.h"
#include "memory.h"
#include <sys/types.h>
#include <sys/stat.h>   
#include <time.h>
#include <stdlib.h>
#include <string.h>
#ifndef STATIC_MODULES
#include <dlfcn.h>
#include <dirent.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include "patchlevel.h"

#ifndef STATIC_MODULES
static ModuleT *Modules = NULL;
static int loaded_modules = 0;
static void addModule(ModuleT *);
static void delModule(ModuleT *);

static char *module_error_table[] =
{
	"object not loaded",
	"object already loaded",
	"missing _module_header",
	"invalid _module_header",
	"module version too old",
	"missing _module_load()",
	"missing _module_unload()",
	"cannot unload core modules",
	NULL
};

#define MOD_SUCCESS		-2
#define MOD_DLOPEN_FAILURE	-1
#define MOD_NOT_LOADED		0
#define MOD_ALREADY_LOADED	1
#define MOD_MISSING_HEADER	2
#define MOD_INVALID_HEADER	3
#define MOD_OLD_VERSION		4
#define MOD_MISSING_LOADFUNC	5
#define MOD_MISSING_UNLOADFUNC	6
#define MOD_CORE_MODULE		7

extern short rehashing;
#define module_error(x)		if (rehashing == 0) report(1, "ERROR: module %s has not been loaded.", x)

#endif

ModuleR RequiredModules;

#ifndef RTLD_NOW
#define RTLD_NOW RTLD_LAZY
#endif

#ifndef STATIC_MODULES
void init_Modules()
{
	addCommand("MODULE", m_module, MAXPARA, MSG_NONE);
	memset(&RequiredModules, '\0', sizeof(RequiredModules));
	return;
}

void check_Modules()
{
	if (RequiredModules.commands == 0) {
		if (RequiredModules.m_join == 0) {
			module_error("m_join.so");
		}
		if (RequiredModules.m_kick == 0) {
			module_error("m_kick.so");
		}
		if (RequiredModules.m_message == 0) {
			module_error("m_message.so");
		}
		if (RequiredModules.m_mode == 0) {
			module_error("m_mode.so");
		}
		if (RequiredModules.m_nick == 0) {
			module_error("m_nick.so");
		}
		if (RequiredModules.m_part == 0) {
			module_error("m_part.so");
		}
		if (RequiredModules.m_ping == 0) {
			module_error("m_ping.so");
		}
		if (RequiredModules.m_pong == 0) {
			module_error("m_pong.so");
		}
		if (RequiredModules.m_quit == 0) {
			module_error("m_quit.so");
		}
		if (RequiredModules.m_user == 0) {
			module_error("m_user.so");
		}
	}
	return;
}

static void addModule(ModuleT *mod)
{
	mod->prev = NULL;
	if ((mod->next = Modules) != NULL) {
		mod->next->prev = mod;
	}
	Modules = mod;
	return;
}

static void delModule(ModuleT *mod_target)
{
	ModuleT *mod;
	for (mod = Modules; mod != NULL; mod = mod->next) {
		if (mod == mod_target) {
			if (mod->prev != NULL) {
				mod->prev->next = mod->next;
			}
			else {
				Modules = mod->next;
			}
			if (mod->next != NULL) {
				mod->next->prev = mod->prev;
			}
			break;
		}
	}
	return;
}

static unsigned int load_one_module(char *name)
{
	void *ptr, (*loadfunc)(), (*unloadfunc)();
	Module *header;
	ModuleT *mod;

	for (mod = Modules; mod != NULL; mod = mod->next) {
		if (!irccmp(mod->header->name, name)) {
			return MOD_ALREADY_LOADED;
		}
	}
	if ((ptr = dlopen(name, RTLD_NOW)) == NULL) {
		return MOD_DLOPEN_FAILURE;
	}
	if ((header = dlsym(ptr, "_module_header")) == NULL) {
		if ((header = dlsym(ptr, "__module_header")) == NULL) {
			dlclose(ptr);
			return MOD_MISSING_HEADER;
		}
	}
	if (BadPtr(header->name)) {
		dlclose(ptr);
		return MOD_INVALID_HEADER;
	}
	if (header->version < MODULE_VERSION) {
		dlclose(ptr);
		return MOD_OLD_VERSION;
	}
	if ((loadfunc = dlsym(ptr, "_module_load")) == NULL) {
		if ((loadfunc = dlsym(ptr, "__module_load")) == NULL) {
			dlclose(ptr);
			return MOD_MISSING_LOADFUNC;
		}
	}
	if ((unloadfunc = dlsym(ptr, "_module_unload")) == NULL) {
		if ((unloadfunc = dlsym(ptr, "__module_unload")) == NULL) {
			dlclose(ptr);
			return MOD_MISSING_UNLOADFUNC;
		}
	}
	mod = (ModuleT *)MyMallocX(sizeof(ModuleT));
	mod->header = header;
	mod->header->ptr = ptr;
	mod->header->unloadfunc = unloadfunc;
	addModule(mod);
	loadfunc();
	loaded_modules++;
	return MOD_SUCCESS;
}

static unsigned int unload_one_module(char *name, int reloading)
{
	ModuleT *mod;

	for (mod = Modules; mod != NULL; mod = mod->next) {
		if (!irccmp(mod->header->name, name)){
			break;
		}
	}
	if (mod == NULL) {
		return MOD_NOT_LOADED;
	}
	if (reloading == 0 && mod->header->core_module == 1 && RequiredModules.commands == 0) {
		return MOD_CORE_MODULE;
	}

	(*mod->header->unloadfunc)();
	dlclose(mod->header->ptr);
	delModule(mod);
	MyFree(mod);
	loaded_modules--;
	return MOD_SUCCESS;
}

static unsigned int reload_one_module(char *name)
{
	unsigned int i;
	if ((i = unload_one_module(name, 1)) != MOD_SUCCESS) {
		return i;
	}
	return load_one_module(name);
}

void load_modules_from_path(char *mod_path)
{
	DIR *mod_dir;
	struct dirent *file_ptr;
	char mod_name[PATH_MAX + 1];
	int len;

	if ((mod_dir = opendir(mod_path)) == NULL) {
		char *error_msg = strerror(errno);
		ircdlog(LOG_ERROR, "Couldn't load modules from path %s: %s", mod_path,
			error_msg);
		report(1, "ERROR: couldn't load modules from path %s: %s", mod_path,
			error_msg);
		return;
	}
	while ((file_ptr = readdir(mod_dir)) != NULL) {
		if ((len = strlen(file_ptr->d_name)) > 3) {
			if (file_ptr->d_name[len - 3] == '.' && file_ptr->d_name[len - 2] == 's'
			  && file_ptr->d_name[len - 1] == 'o') {
				ircsprintf(mod_name, "%s/%s", mod_path, file_ptr->d_name);
				handle_module_call(mod_name, NULL, MOD_CALLTYPE_LOAD);
			}
		}
	}
	closedir(mod_dir);
	return;
}

void handle_module_call(char *name, aClient *cptr, int calltype)
{
	unsigned int handle_value = 0;
	char *calltype_str = "[unknown]";

	if (BadPtr(name)) {
		return;
	}
	switch (calltype) {
		case MOD_CALLTYPE_LOAD:
			calltype_str = "load";
			if ((handle_value = load_one_module(name)) == MOD_SUCCESS) {
				if (cptr != NULL) {
					sendto_realops("Loaded module %s [%s] by request of %s",
						Modules->header->name, Modules->header->revision, cptr->name);
				}
				return;
			}
			break;
		case MOD_CALLTYPE_RELOAD:
			calltype_str = "reload";
			if ((handle_value = reload_one_module(name)) == MOD_SUCCESS) {
				if (cptr != NULL) {
					sendto_realops("Reloaded module %s by request of %s", name, cptr->name);
				}
				return;
			}
			break;
		case MOD_CALLTYPE_UNLOAD:
			calltype_str = "unload";
			if ((handle_value = unload_one_module(name, 0)) == MOD_SUCCESS) {
				if (cptr != NULL) {
					sendto_realops("Unloaded module %s by request of %s", name, cptr->name);
				}
				return;
			}
			break;
		default:
			sendto_realops_lev(DEBUG_LEV, "Ignoring unknown handle_module_call() type %d",
				calltype);
			return;
	}
	switch (handle_value) {
		case MOD_SUCCESS:
			break;
		case MOD_DLOPEN_FAILURE:
			if (cptr != NULL) {
				sendto_one(cptr, ":%s NOTICE %s :Couldn't %s module %s",
					me.name, cptr->name, calltype_str, (char *)dlerror());
			}
			else {
				report(0, "servinfo::modules: failed to %s %s", calltype_str,
					(char *)dlerror());
			}
			break;
		default:
			if (cptr != NULL) {
				sendto_one(cptr, ":%s NOTICE %s :Couldn't %s module %s: %s",
					me.name, cptr->name, calltype_str, name,
					module_error_table[handle_value]);
			}
			else if (handle_value != MOD_ALREADY_LOADED) {
				report(0, "servinfo::modules: error %sing module %s: %s", calltype_str,
					name, module_error_table[handle_value]);
			}
			break;
	}
	return;
}

void unload_all_modules()
{
	ModuleT *mod, *next = NULL;
	for (mod = Modules; mod != NULL; mod = next) {
		next = mod->next;
		(*mod->header->unloadfunc)();
		dlclose(mod->header->ptr);
		delModule(mod);
		MyFree(mod);
		loaded_modules--;
	}
	return;
}

/*
 * m_module
 *	parv[0] = sender prefix
 *	parv[1] = function
 *	parv[2] = module
 */
int m_module(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (!HasMode(sptr, UMODE_NETADMIN|UMODE_ADMIN|UMODE_COADMIN)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODULE");
		return 0;
	}
	if (!mycmp(parv[1], "LIST")) {
		sendto_one(sptr, ":%s NOTICE %s :There %s currently %d module%s loaded", me.name, parv[0],
			loaded_modules == 1 ? "is" : "are", loaded_modules, loaded_modules == 1 ? "" : "s");
		sendto_one(sptr, ":%s NOTICE %s :The current module version is %d.", me.name,
			cptr->name, MODULE_VERSION);
		if (loaded_modules > 0) {
			ModuleT *mod;
			sendto_one(sptr, ":%s NOTICE %s :-----------------------------------------", me.name,
				parv[0]);
			for (mod = Modules; mod != NULL; mod = mod->next) {
				sendto_one(sptr, ":%s NOTICE %s :%s (%s) [%s]", me.name, parv[0],
					mod->header->name,
					!BadPtr(mod->header->info) ? mod->header->info : "unknown",
					!BadPtr(mod->header->revision) ? mod->header->revision : "unknown");
			}
			sendto_one(sptr, ":%s NOTICE %s :-----------------------------------------", me.name,
				parv[0]);
			sendto_one(sptr, ":%s NOTICE %s :End of module list", me.name, parv[0]);
		}
		return 0;
	}
	else {
		if (parc < 3 || *parv[2] == '\0') {
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODULE");
			return 0;
		}
		if (!mycmp(parv[1], "LOAD")) {
			handle_module_call(parv[2], cptr, MOD_CALLTYPE_LOAD);
		}
		else if (!mycmp(parv[1], "RELOAD")) {
			handle_module_call(parv[2], cptr, MOD_CALLTYPE_RELOAD);
		}
		else if (!mycmp(parv[1], "UNLOAD")) {
			handle_module_call(parv[2], cptr, MOD_CALLTYPE_UNLOAD);
		}
		return 0;
	}
	sendto_one(sptr, ":%s NOTICE %s :Syntax: /MODULE <LOAD|RELOAD|UNLOAD|LIST> [path.to.module.so]",
		me.name, parv[0]);
	return 0;
}
#endif
