/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_bsd.c: network functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_bsd.c,v 1.91 2002/11/23 21:51:06 rageircd Exp $
 */

#include "memory.h"
#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include "zlink.h"
#include "ssl.h"
#include "sock.h"
#include "h.h"
#include "fdlist.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(SOL20)
#include <sys/filio.h>
#include <sys/select.h>
#include <unistd.h>
#endif
#include "inet.h"
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <utmp.h>
#include <sys/resource.h>
#ifdef AIX
#include <time.h>
#include <arpa/nameser.h>
#else
#include "nameser.h"
#endif

#ifdef HAVE_POLL_ON
#include <poll.h>
#if defined(SOL20) || defined(AIX)
#define CONNECTFAST
#endif
#endif

#ifdef HAVE_THROTTLE_ON
extern int throttle_check(char *, int, time_t);
extern int throttle_force(char *);
#endif
#ifdef HAVE_OPENSSL_ON
extern int ssl_do_handshake(aClient *);
extern int safe_SSL_accept(aClient *, int);
#endif

extern int AllowClient(aClient *, char *);
extern void report_listener_error(char *, Listener *);
extern void set_time();

#ifndef IN_LOOPBACKNET
#define IN_LOOPBACKNET 0x7f
#endif

aClient *local[MAXCONNECTIONS];
int highest_fd = 0;

#ifdef HAVE_FD_ALLOC
fd_set *write_set, *read_set;
#endif

static struct sockaddr_in mysk;

#ifdef INCREASE_SOCK_BUFS
static char *readbuf;
int rcvbufmax = 0;
int sndbufmax = 0;
#else
static char readbuf[8192];
#endif

#ifdef RLIMIT_FDMAX
#define RLIMIT_FD_MAX RLIMIT_FDMAX
#else
#ifdef RLIMIT_NOFILE
#define RLIMIT_FD_MAX RLIMIT_NOFILE
#else
#ifdef RLIMIT_OPEN_MAX
#define RLIMIT_FD_MAX RLIMIT_OPEN_MAX
#else
#undef RLIMIT_FD_MAX
#endif
#endif
#endif

void report_error(char *text, aClient * cptr)
{
	int errtmp = errno, err, len = sizeof(err);
	char *host = (cptr != NULL) ? get_client_name(cptr, (IsServer(cptr) ? HIDEME : FALSE)) : "";

	Debug((DEBUG_ERROR, text, host, strerror(errtmp)));

#ifdef	SO_ERROR
	if (!IsMe(cptr) && cptr->fd >= 0) {
		if (!getsockopt(cptr->fd, SOL_SOCKET, SO_ERROR, (char *) &err, &len)) {
			if (err != 0) {
				errtmp = err;
			}
		}
	}
#endif
	sendto_realops_lev(DEBUG_LEV, text, host, strerror(errtmp));
	ircdlog(LOG_ERROR, text, host, strerror(errtmp));
	return;
}

static int check_init(aClient *cptr, char *sockn)
{
	struct sockaddr_in sk;
	int len = sizeof(struct sockaddr_in);

	if (getpeername(cptr->fd, (struct sockaddr *) &sk, &len) == -1) {
		return -1;
	}
	strcpy(sockn, inetntoa((char *)&sk.sin_addr));
	if (inet_netof(sk.sin_addr) == IN_LOOPBACKNET) {
		strncpyzt(sockn, me.sockhost, HOSTLEN);
	}
	memcpy((char *)&cptr->ip, (char *)&sk.sin_addr, sizeof(struct in_addr));
	cptr->port = ntohs(sk.sin_port);
	return 0;
}

int check_client(aClient *cptr)
{
	static char sockname[HOSTLEN + 1];
	int i;

	Debug((DEBUG_DNS, "ch_cl: check access for %s[%s]", cptr->name, inetntoa((char *) &cptr->ip)));
	if (check_init(cptr, sockname) != 0) {
		return CLIENTAUTH_SOCKERR;
	}
	if ((i = AllowClient(cptr, sockname)) != 0) {
		Debug((DEBUG_DNS, "ch_cl: access denied: %s[%s]", cptr->name, sockname));
		return i;
	}

	Debug((DEBUG_DNS, "ch_cl: access ok: %s[%s]", cptr->name, sockname));
	if (inet_netof(cptr->ip) == IN_LOOPBACKNET || inet_netof(cptr->ip) == inet_netof(mysk.sin_addr)) {
		ircstp->is_loc++;
		cptr->flags |= FLAGS_LOCAL;
	}
	return CLIENTAUTH_MATCHED;
}

static int completed_connection(aClient *cptr)
{
	ConfigItem_link *link = cptr->serv ? cptr->serv->conf : NULL;

	SetHandshake(cptr);

	if (link == NULL) {
		sendto_gnotice("Lost link configuration for %s", get_client_name(cptr, HIDEME));
		return -1;
	}
	if (link->auth != NULL) {
		sendto_one(cptr, "PASS %s :TS", link->auth->string);
	}

	sendto_one(cptr, "CAPAB NOQUIT BURST UNCONNECT ZIP SSJv2 NICKIP VHOST SNICK");
#ifdef HAVE_OPENSSL_ON
	if (link->options & LINK_SECURE) {
		sendto_one(cptr, "CAPAB DKEY");
	}
#endif
	sendto_one(cptr, "SERVER %s 1 :%s", me.name, me.info);
	return (IsDead(cptr)) ? -1 : 0;
}

void close_connection(aClient *cptr)
{
	int i, j;
	int empty = cptr->fd;

	if (IsServer(cptr)) {
		ircstp->is_sv++;
		ircstp->is_sbs += cptr->sendB;
		ircstp->is_sbr += cptr->receiveB;
		ircstp->is_sks += cptr->sendK;
		ircstp->is_skr += cptr->receiveK;
		ircstp->is_sti += timeofday - cptr->firsttime;
		if (ircstp->is_sbs > 2047) {
			ircstp->is_sks += (ircstp->is_sbs >> 10);
			ircstp->is_sbs &= 0x3ff;
		}
		if (ircstp->is_sbr > 2047) {
			ircstp->is_skr += (ircstp->is_sbr >> 10);
			ircstp->is_sbr &= 0x3ff;
		}
	} else if (IsClient(cptr)) {
		ircstp->is_cl++;
		ircstp->is_cbs += cptr->sendB;
		ircstp->is_cbr += cptr->receiveB;
		ircstp->is_cks += cptr->sendK;
		ircstp->is_ckr += cptr->receiveK;
		ircstp->is_cti += timeofday - cptr->firsttime;
		if (ircstp->is_cbs > 2047) {
			ircstp->is_cks += (ircstp->is_cbs >> 10);
			ircstp->is_cbs &= 0x3ff;
		}
		if (ircstp->is_cbr > 2047) {
			ircstp->is_ckr += (ircstp->is_cbr >> 10);
			ircstp->is_cbr &= 0x3ff;
		}
	}
	else {
		ircstp->is_ni++;
	}

	if (IsServer(cptr) && cptr->serv->conf->flag.temporary == 0 &&
	  (cptr->serv->conf->options & LINK_AUTOCONNECT)) {
		ConfigItem_link *link = cptr->serv->conf;
		link->hold = time(NULL);
		link->hold += (link->hold - cptr->since > HANGONGOODLINK) ? HANGONRETRYDELAY : link->class->connectfreq;
		if (nextconnect > link->hold) {
			nextconnect = link->hold;
		}
	}
	if (cptr->fd >= 0) {
#ifdef HAVE_OPENSSL_ON
		if (!IsDead(cptr)) {
#endif
			dump_connections(cptr->fd);
#ifdef HAVE_OPENSSL_ON
		}
#endif
		local[cptr->fd] = NULL;
#ifdef HAVE_OPENSSL_ON
		if (cptr->flags & FLAGS_SECURE && cptr->ssl != NULL) {
			SSL_set_shutdown(cptr->ssl, SSL_RECEIVED_SHUTDOWN);
			SSL_smart_shutdown(cptr->ssl);
			SSL_free(cptr->ssl);
			cptr->ssl = NULL;
		}
#endif
		close(cptr->fd);
		cptr->fd = -2;
		DBufClear(&cptr->sendQ);
		DBufClear(&cptr->recvQ);
		memset(cptr->passwd, '\0', sizeof(cptr->passwd));
	}
	for (; highest_fd > 0; highest_fd--) {
		if (local[highest_fd] != NULL) {
			break;
		}
	}

	cptr->from = NULL;
	if (empty > 0) {
		if ((j = highest_fd) > (i = empty) && !IsLog(local[j]) && !IsMe(local[j])) {
			if (dup2(j, i) == -1) {
				return;
			}
			local[i] = local[j];
			local[i]->fd = i;
#ifdef HAVE_OPENSSL_ON
			if (local[i]->ssl != NULL && (local[i]->flags & FLAGS_SECURE)) {
				BIO_set_fd(SSL_get_rbio(local[i]->ssl), i, BIO_NOCLOSE);
				BIO_set_fd(SSL_get_wbio(local[i]->ssl), i, BIO_NOCLOSE);
			}
#endif
			local[j] = NULL;

			if (IsServer(local[i])) {
#ifndef NO_PRIORITY
				delfrom_fdlist(j, &busycli_fdlist);
#endif
				delfrom_fdlist(j, &serv_fdlist);
#ifndef NO_PRIORITY
				addto_fdlist(i, &busycli_fdlist);
#endif
				addto_fdlist(i, &serv_fdlist);
			}

			if (HasMode(local[i], UMODE_OPER)) {
#ifndef NO_PRIORITY
				delfrom_fdlist(j, &busycli_fdlist);
#endif
				delfrom_fdlist(j, &oper_fdlist);
#ifndef NO_PRIORITY
				addto_fdlist(i, &busycli_fdlist);
#endif
				addto_fdlist(i, &oper_fdlist);
			}

			close(j);
			while (local[highest_fd] == NULL) {
				highest_fd--;
			}
		}
	}
	return;
}

#ifdef INCREASE_SOCK_BUFS
void increase_sock_bufs(int fd, int server)
{
	int opt = (server == 1) ? rcvbufmax : CLIENT_BUFFER_SIZE;
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&opt, sizeof(opt)) < 0) {
		ircdlog(LOG_ERROR, "increase_sock_bufs(SO_RCVBUF) for fd %d (%s) failed", fd,
			(server == 1) ? "server" : "client");
	}
	opt = (server == 1) ? (SEND_BUF_SIZE * 4) : SEND_BUF_SIZE;
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&opt, sizeof(opt)) < 0) {
		ircdlog(LOG_ERROR, "increase_sock_bufs(SO_SNDBUF) for fd %d (%s) failed", fd,
			(server == 1) ? "server" : "client");
	}
	return;
}
#endif

void set_sock_opts(int fd, aClient *cptr, Listener *listener)
{
	int opt;

#ifdef SO_REUSEADDR
	opt = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt));
#endif

#if defined(SO_DEBUG) && defined(DEBUGMODE) && 0 && !defined(SOL20)
	opt = 1;
	setsockopt(fd, SOL_SOCKET, SO_DEBUG, (char *)&opt, sizeof(opt));
#endif

#ifdef SO_USELOOPBACK
	opt = 1;
	setsockopt(fd, SOL_SOCKET, SO_USELOOPBACK, (char *)&opt, sizeof(opt));
#endif

#ifdef SO_RCVBUF
#ifdef INCREASE_SOCK_BUFS
	if (rcvbufmax == 0) {
		int optlen = sizeof(rcvbufmax);
		getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&rcvbufmax, &optlen);
		while ((rcvbufmax < 16385) && (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&rcvbufmax, optlen) >= 0)) {
			rcvbufmax += 1024;
		}
		getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&rcvbufmax, &optlen);
		readbuf = (char *)malloc(rcvbufmax * sizeof(char));
	}
	opt = (cptr != NULL && IsServer(cptr)) ? rcvbufmax : 4096;
#else
	opt = 8192;
#endif
	setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&opt, sizeof(opt));
#endif

#ifdef SO_SNDBUF
#ifdef INCREASE_SOCK_BUFS
	if (sndbufmax == 0) {
		int optlen = sizeof(sndbufmax);
		getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&sndbufmax, &optlen);
		while ((sndbufmax < 16385) && (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&sndbufmax, optlen) >= 0)) {
			sndbufmax += 1024;
		}
		getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&sndbufmax, &optlen);
	}
	opt = (cptr != NULL && IsServer(cptr)) ? sndbufmax : 4096;
#else
	opt = 8192;
#endif
	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&opt, sizeof(opt));
#endif

#if defined(IP_OPTIONS) && defined(IPPROTO_IP)
	if (cptr != NULL) {
		char *s = readbuf;
#ifdef INCREASE_SOCK_BUFS
		char *t = readbuf + (rcvbufmax * sizeof(char)) / 2;
		int opt = (rcvbufmax * sizeof(char)) / 8;
#else
		char *t = readbuf + sizeof(readbuf) / 2;
		int opt = sizeof(readbuf) / 8;
#endif

		if (getsockopt(fd, IPPROTO_IP, IP_OPTIONS, t, &opt) >= 0 && opt > 0) {
			for (*readbuf = '\0'; opt > 0; opt--, s += 3) {
				ircsprintf(s, "%02.2x:", *t++);
			}
			*s = '\0';
			sendto_realops("Connection %s using IP opts: %s", get_client_name(cptr, HIDEME), readbuf);
		}
		setsockopt(fd, IPPROTO_IP, IP_OPTIONS, (char *)NULL, 0);
	}
#endif
	return;
}

int get_sockerr(aClient *cptr)
{
	int errtmp = errno, err = 0, len = sizeof(err);

#ifdef SO_ERROR
	if (cptr->fd >= 0) {
		if (!getsockopt(cptr->fd, SOL_SOCKET, SO_ERROR, (char *)&err, &len)) {
			if (err != 0) {
				errtmp = err;
			}
		}
	}
#endif
	return errtmp;
}

char *irc_get_sockerr(aClient *cptr)
{
	if (cptr->sockerr == 0) {
		return "No error";
	}
	if (cptr->sockerr > 0) {
		return strerror(cptr->sockerr);
	}
	switch (cptr->sockerr) {
		case -1:
			return "unset error message (this is a bug!)";
		case IRCERR_BUFALLOC:
			return "dbuf allocation error";
		case IRCERR_ZIP:
			return "compression failure";
		case IRCERR_SSL:
			return "SSL error";
		default:
			break;
	}
	return "unknown error";
}

int set_blocking(int fd)
{
	int flags;
	if ((flags = fcntl(fd, F_GETFL, 0)) < 0 || fcntl(fd, F_SETFL, flags & ~O_NONBLOCK) < 0) {
		return 0;
	}
	else {
		return 1;
	}
}

void set_non_blocking(int fd)
{
	int res, nonb = 0;

#ifdef NBLOCK_POSIX
	nonb |= O_NONBLOCK;
#endif
#ifdef NBLOCK_BSD
	nonb |= O_NDELAY;
#endif
#ifdef NBLOCK_SYSV
	res = 1;

	ioctl(fd, FIONBIO, &res);
#else
	if ((res = fcntl(fd, F_GETFL, 0)) != -1) {
		fcntl(fd, F_SETFL, res | nonb);
	}
#endif
	return;
}

static void add_connection(Listener *listener, int fd)
{
	aClient *acptr = NULL;
	struct sockaddr_in addr;
	int len = sizeof(struct sockaddr_in);

	if (getpeername(fd, (struct sockaddr *)&addr, &len) == -1) {
		ircstp->is_ref++;
		close(fd);
		acptr->fd = -2;
		return;
	}

	acptr = make_client(NULL, &me);
	strncpy(acptr->sockhost, inetntoa((char *)&addr.sin_addr), HOSTLEN);
	strcpy(acptr->host, acptr->sockhost);
	acptr->ip.s_addr = addr.sin_addr.s_addr;
	acptr->port = ntohs(addr.sin_port);
	acptr->fd = fd;
	acptr->listener = listener;
	acptr->listener->bind->clients++;
	acptr->listener->count++;

	set_non_blocking(acptr->fd);
	set_sock_opts(acptr->fd, acptr, NULL);

#ifdef HAVE_OPENSSL_ON
	if (acptr->listener->options & BIND_SECURE) {
		if (!ssl_do_handshake(acptr)) {
			ircstp->is_ref++;
			close(acptr->fd);
			acptr->fd = -2;
			free_client(acptr);
			return;
		}
	}
#endif
	connauth_init(acptr);
	return;
}

static int do_client_queue(aClient *cptr)
{
	int dolen = 0, done;

	while (DBufLength(&cptr->recvQ) && !NoNewLine(cptr) && ((cptr->status < STAT_UNKNOWN) ||
	  (cptr->since - timeofday < 10) || (cptr->serv != NULL && (cptr->serv->sflags & SFLAGS_RC4_NEG)))) {
		if (IsServer(cptr) || (cptr->serv != NULL && (cptr->serv->sflags & SFLAGS_RC4_NEG))) {
#ifdef INCREASE_SOCK_BUFS
			if ((dolen = dbuf_get(&cptr->recvQ, readbuf, rcvbufmax * sizeof(char))) <= 0) {
#else
			if ((dolen = dbuf_get(&cptr->recvQ, readbuf, sizeof(readbuf))) <= 0) {
#endif
				break;
			}
			if ((done = dopacket(cptr, readbuf, dolen))) {
				return done;
			}
		}

#ifdef INCREASE_SOCK_BUFS
		if ((dolen = dbuf_getmsg(&cptr->recvQ, readbuf, rcvbufmax * sizeof(char))) <= 0) {
#else
		if ((dolen = dbuf_getmsg(&cptr->recvQ, readbuf, sizeof(readbuf))) <= 0) {
#endif
			if (dolen < 0) {
				return exit_client(cptr, cptr, cptr, "dbuf_getmsg fail");
			}
			if (DBufLength(&cptr->recvQ) < 510) {
				cptr->flags |= FLAGS_NONL;
				break;
			}
			DBufClear(&cptr->recvQ);
			break;
		}
		else if (client_dopacket(cptr, readbuf, dolen) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}
	}
	return 1;
}

#define MAX_CLIENT_RECVQ 8192

static int read_packet(aClient *cptr)
{
	int length = 0, done;
#ifdef INCREASE_SOCK_BUFS
	int readbufsize = (IsPerson(cptr)) ? (MAX_CLIENT_RECVQ * sizeof(char)) : (rcvbufmax * sizeof(char));
#else
	int readbufsize = sizeof(readbuf);
#endif

	if (!(IsPerson(cptr) && DBufLength(&cptr->recvQ) > MAX_CLIENT_RECVQ)) {
		errno = 0;
		length = ircrecv(cptr, readbuf, readbufsize);
		cptr->lasttime = timeofday;
		if (cptr->lasttime > cptr->since) {
			cptr->since = cptr->lasttime;
		}
		cptr->flags &= ~(FLAGS_PINGSENT|FLAGS_NONL);

		if (length == -1 && (errno == EWOULDBLOCK || errno == EAGAIN)) {
			return 1;
		}
		if (length <= 0) {
			cptr->sockerr = length ? errno : 0;
			return length;
		}
	}
	if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr) || (cptr->serv != NULL &&
	  (cptr->serv->sflags & SFLAGS_RC4_NEG))) {
		if (length > 0) {
			if ((done = dopacket(cptr, readbuf, length))) {
				return done;
			}
		}
	}
	else {
		if (dbuf_put(&cptr->recvQ, readbuf, length, 0) < 0) {
			return exit_client(cptr, cptr, cptr, "dbuf_put fail");
		}
		if (IsPerson(cptr) && ((NOOPERFLOOD == 1 && !HasMode(cptr, UMODE_OPER)) || NOOPERFLOOD == 0)
		  && DBufLength(&cptr->recvQ) > CLIENT_FLOOD) {
			sendto_realops_lev(FLOOD_LEV, "Flood -- %s!%s@%s (%d) Exceeds %d RecvQ", cptr->name,
				cptr->username, MaskedHost(cptr), DBufLength(&cptr->recvQ), CLIENT_FLOOD);
			return exit_client(cptr, cptr, cptr, "Excess Flood");
		}
		return do_client_queue(cptr);
	}
	return 1;
}

void accept_connection(Listener *listener)
{
	static struct sockaddr_in addr;
	char host[HOSTLEN + 2];
	int newfd, addrlen = sizeof(struct sockaddr_in);
	ConfigItem_ban *ban;

	listener->lasttime = timeofday;
	if ((newfd = accept(listener->fd, (struct sockaddr *)&addr, &addrlen)) < 0) {
		switch (errno) {
#ifdef EMFILE
			case EMFILE:
				report_listener_error("Cannot accept connections %s:%s", listener);
				break;
#endif
			default:
				break;
		}
		return;
	}

	strncpyzt(host, inetntoa((char *)&addr.sin_addr), sizeof(host));
	if ((ban = find_ban(BAN_IP, host)) != NULL) {
#ifdef HAVE_THROTTLE_ON
		if (THROTTLE_REJECTED_CLIENTS == 1) {
			throttle_force(host);
		}
#endif
		conf_doban(NULL, NULL, ban, host, newfd);
		close(newfd);
		return;
	}

#ifdef HAVE_THROTTLE_ON
	if (!throttle_check(host, newfd, timeofday)) {
		ircstp->is_ref++;
		close(newfd);
		return;
	}
#endif

	if (newfd >= HARD_FDLIMIT - 10) {
		ircstp->is_ref++;
		sendto_ops("All connections in use. (%s)", get_listener_name(listener));
		send(newfd, "ERROR :All connections in use\r\n", 32, 0);
		close(newfd);
		return;
	}

	ircstp->is_ac++;
	add_connection(listener, newfd);
	nextping = timeofday;
	return;
}

static void read_error_exit(aClient *cptr, int length, int err)
{
	char errmsg[512], *clipath = get_client_name(cptr, HIDEME);

	if (IsServer(cptr) || IsHandshake(cptr) || IsConnecting(cptr)) {
		if (length == 0) {
			sendto_gnotice("Server %s closed the connection", clipath);
			sendto_serv_butone(cptr, ":%s GNOTICE :Server %s closed the connection",
				me.name, clipath);
		}
		else {
			sendto_gnotice("Read error from %s, closing link (%s)", clipath, strerror(err));
			sendto_serv_butone(cptr, ":%s GNOTICE :Read error from %s, closing link (%s)",
				me.name, clipath, strerror(err));
		}
	}
	if (err != 0) {
		ircsprintf(errmsg, "Read error: %s", strerror(err));
	}
	else {
		ircsprintf(errmsg, "Client closed connection");
	}
	exit_client(cptr, cptr, &me, errmsg);
	return;
}


#ifdef FAST_FD_ISSET
#define MYFD_ISSET_READ		(read_set->fds_bits[fd_read_offset] & fd_read_mask)
#define MYFD_ISSET_WRITE	(write_set->fds_bits[fd_write_offset] & fd_write_mask)
#define MYFD_CLR_READ		read_set->fds_bits[fd_read_offset] &= ~fd_read_mask;
#define FAST_FD_INC		fd_read_mask <<= 1; \
				if (fd_read_mask == 0) { \
					fd_read_offset++; \
					fd_read_mask = 1; \
				} \
				fd_write_mask <<= 1; \
				if (fd_write_mask == 0) { \
					fd_write_offset++; \
					fd_write_mask = 1; \
				}
#else
#define MYFD_ISSET_READ		(FD_ISSET(i, read_set))
#define MYFD_ISSET_WRITE	(FD_ISSET(i, write_set))
#define MYFD_CLR_READ		FD_CLR(i, read_set);
#define FAST_FD_INC
#endif

#ifndef HAVE_POLL_ON

int read_message(time_t delay, fdlist *listp)
{
	aClient *cptr;
	int nfds, res, length, i, j;
	struct timeval wait;
#ifndef HAVE_FD_ALLOC
	fd_set readset, writeset;
	fd_set *read_set = &readset, *write_set = &writeset;
#endif
	time_t delay2 = delay, now;
	char errmsg[512];
#ifdef USE_FAST_FD_ISSET
	int fd_read_mask, fd_read_offset, fd_write_mask, fd_write_offset;
#endif
	Listener *l;
#ifndef NO_PRIORITY
	int check_listeners = (listp == &busycli_fdlist || listp == &default_fdlist || listp == NULL) ? 1 : 0;
#else
	int check_listeners = (listp == &default_fdlist || listp == NULL) ? 1 : 0;
#endif
	ConnectAuth *auth, *next = NULL;

	if (listp == NULL) {
		listp = &default_fdlist;
		listp->last_entry = highest_fd + 1;
	}

	now = timeofday;
	for (res = 0;;) {
		FD_ZERO(read_set);
		FD_ZERO(write_set);
		for (auth = connauth_list; auth != NULL; auth = auth->next) {
			if (auth->fd < 0) {
				ircdlog(LOG_ERROR, "read_message() connauth encountered fd -1, auth %lX", auth);
				continue;
			}
			if (auth->flags & CAF_ID_CONNECTING) {
				FD_SET(auth->fd, write_set);
			}
			else {
				FD_SET(auth->fd, read_set);
			}
		}
		if (check_listeners == 1) {
			for (l = listener_list; l != NULL; l = l->next) {
				FD_SET(l->fd, read_set);
			}
		}
		for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j]) {
			if (!(cptr = local[i]) || IsLog(cptr) || IsMe(cptr)) {
				continue;
			}
#ifdef HAVE_OPENSSL_ON
			if (cptr->flags & FLAGS_SECURE && cptr->ssl != NULL &&
			  !SSL_is_init_finished(cptr->ssl)) {
				if (IsDead(cptr) || !safe_SSL_accept(cptr, cptr->fd)) {
					close_connection(cptr);
				}
				continue;
			}
#endif
			if (IsConnecting(cptr)) {
				FD_SET(i, read_set);
			}
			else {
				if (DBufLength(&cptr->recvQ) && delay2 > 2) {
					delay2 = 1;
				}
				if (DBufLength(&cptr->recvQ) < 4088) {
					FD_SET(i, read_set);
				}
			}

			length = DBufLength(&cptr->sendQ);
			if (DoList(cptr) && IsSendable(cptr)) {
				send_list(cptr, 64);
				length = DBufLength(&cptr->sendQ);
			}
			if (length || IsConnecting(cptr) || (IsServer(cptr) && (cptr->serv->sflags & SFLAGS_ZIP_OUT)
			  && zip_is_data_out(cptr->serv->zip_out))) {
				FD_SET(i, write_set);
			}
		}

		wait.tv_sec = MIN(delay2, delay);
		wait.tv_usec = 0;
		nfds = select(MAXCONNECTIONS, read_set, write_set, 0, &wait);
		set_time();
		if (nfds == -1 && errno == EINTR) {
			return -1;
		}
		else if (nfds >= 0) {
			break;
		}

		report_error("select %s:%s", &me);
		res++;
		if (res > 5) {
			restart("too many select errors");
		}
		sleep(10);
	}
	for (auth = connauth_list; auth != NULL; auth = next) {
		next = auth->next;
		if (auth->fd < 0) {
			ircdlog(LOG_ERROR, "read_message() stage 2 connauth encountered fd -1, auth %lX", auth);
			nfds--;
			continue;
		}
		if (auth->flags & CAF_ID_CONNECTING && FD_ISSET(auth->fd, write_set)) {
			connauth_send_id(auth);
			nfds--;
		}
		else if (FD_ISSET(auth->fd, read_set)) {
			connauth_read_id(auth);
			nfds--;
		}
	}
	if (check_listeners == 1) {
		for (l = listener_list; l != NULL; l = l->next) {
			if (FD_ISSET(l->fd, read_set)) {
				nfds--;
				accept_connection(l);
			}
		}
	}

#ifdef USE_FAST_FD_ISSET
	fd_read_mask = fd_write_mask = 1;
	fd_read_offset = fd_write_offset = 0;

	for (i = 0; i <= highest_fd; i++)
#else
	for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j])
#endif
	{
		if (!(cptr = local[i])) {
			FAST_FD_INC
			continue;
		}
		if (MYFD_ISSET_WRITE) {
			int write_err = 0;

			nfds--;
			if (IsConnecting(cptr)) {
				write_err = completed_connection(cptr);
			}
			if (write_err == 0) {
				send_queued(cptr);
			}
			if (IsDead(cptr) || write_err) {
				if (MYFD_ISSET_READ) {
					MYFD_CLR_READ
				}
				ircsprintf(errmsg, "Write Error: %s",
					cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
				exit_client(cptr, cptr, &me, errmsg);
				FAST_FD_INC
				continue;
			}
		}

		length = 1;
		if (MYFD_ISSET_READ) {
			length = read_packet(cptr);
		}
		else if (IsPerson(cptr) && !NoNewLine(cptr)) {
			length = do_client_queue(cptr);
		}

#ifdef DEBUGMODE
		readcalls++;
#endif

		if (length != FLUSH_BUFFER && IsDead(cptr)) {
			if (MYFD_ISSET_READ) {
				MYFD_CLR_READ
			}
			ircsprintf(errmsg, "Read/Dead Error: %s",
				cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
			exit_client(cptr, cptr, &me, errmsg);
			FAST_FD_INC
			continue;
		}
		if (length > 0) {
			if (MYFD_ISSET_READ) {
				nfds--;
			}
			FAST_FD_INC
			continue;
		}
		if (length != FLUSH_BUFFER) {
			Debug((DEBUG_ERROR, "READ ERROR: fd = %d %d %d", i, errno, length));
			read_error_exit(cptr, length, cptr->sockerr);
		}
		FAST_FD_INC
	}
	return 0;
}

#else

#ifdef AIX
#define POLLREADFLAGS (POLLIN|POLLMSG)
#endif
#if defined(POLLMSG) && defined(POLLIN) && defined(POLLRDNORM)
#define POLLREADFLAGS (POLLMSG|POLLIN|POLLRDNORM)
#endif
#if defined(POLLIN) && defined(POLLRDNORM) && !defined(POLLMSG)
#define POLLREADFLAGS (POLLIN|POLLRDNORM)
#endif
#if defined(POLLIN) && !defined(POLLRDNORM) && !defined(POLLMSG)
#define POLLREADFLAGS POLLIN
#endif
#if defined(POLLRDNORM) && !defined(POLLIN) && !defined(POLLMSG)
#define POLLREADFLAGS POLLRDNORM
#endif

#if defined(POLLOUT) && defined(POLLWRNORM)
#define POLLWRITEFLAGS (POLLOUT|POLLWRNORM)
#else
#if defined(POLLOUT)
#define POLLWRITEFLAGS POLLOUT
#else
#if defined(POLLWRNORM)
#define POLLWRITEFLAGS POLLWRNORM
#endif
#endif
#endif

#if defined(POLLERR) && defined(POLLHUP)
#define POLLERRORS (POLLERR|POLLHUP)
#else
#define POLLERRORS POLLERR
#endif

#define PFD_SETR(thisfd)	CHECK_PFD(thisfd); pfd->events |= POLLREADFLAGS;
#define PFD_SETW(thisfd)	CHECK_PFD(thisfd); pfd->events |= POLLWRITEFLAGS;
#define CHECK_PFD(thisfd)	if (pfd->fd != thisfd) { \
					pfd = &poll_fdarray[nbr_pfds++]; \
					pfd->fd = thisfd; \
					pfd->events = 0; \
				}

int read_message(time_t delay, fdlist *listp)
{
	aClient *cptr;
	int nfds, nbr_pfds = 0, res, length, fd, rr, rw, i, j;
	static struct pollfd poll_fdarray[MAXCONNECTIONS];
	struct pollfd *pfd = poll_fdarray;
	u_long waittime;
	time_t delay2 = delay;
	static char errmsg[512];
	Listener *l;
#ifndef NO_PRIORITY
	int check_listeners = (listp == &busycli_fdlist || listp == &default_fdlist || listp == NULL) ? 1 : 0;
#else
	int check_listeners = (listp == &default_fdlist || listp == NULL) ? 1 : 0;
#endif
	ConnectAuth *auth, *next = NULL;

	if (listp == NULL) {
		listp = &default_fdlist;
		listp->last_entry = highest_fd + 1;
	}
	for (res = 0;;) {
		nbr_pfds = 0;
		pfd = poll_fdarray;
		pfd->fd = -1;

		for (auth = connauth_list; auth != NULL; auth = auth->next) {
			if (auth->fd < 0) {
				ircdlog(LOG_ERROR, "read_message() connauth encountered fd -1, auth %lX", auth);
				continue;
			}
			if (auth->flags & CAF_ID_CONNECTING) {
				PFD_SETW(auth->fd);
			}
			else {
				PFD_SETR(auth->fd);
			}
		}
		if (check_listeners == 1) {
			for (l = listener_list; l != NULL; l = l->next) {
				PFD_SETR(l->fd);
			}
		}
		for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j]) {
			if (!(cptr = local[i]) || IsLog(cptr) || IsMe(cptr)) {
				continue;
			}
#ifdef HAVE_OPENSSL_ON
			if (cptr->flags & FLAGS_SECURE && cptr->ssl != NULL &&
			  !SSL_is_init_finished(cptr->ssl)) {
				if (IsDead(cptr) || !safe_SSL_accept(cptr, cptr->fd)) {
					close_connection(cptr);
				}
				continue;
			}
#endif
			if (DBufLength(&cptr->recvQ) && delay2 > 2) {
				delay2 = 1;
			}
			if (DBufLength(&cptr->recvQ) < 4088) {
				PFD_SETR(i);
			}

			length = DBufLength(&cptr->sendQ);
			if (DoList(cptr) && IsSendable(cptr)) {
				send_list(cptr, 64);
				length = DBufLength(&cptr->sendQ);
			}
			if (length || IsConnecting(cptr) || (IsServer(cptr) && (cptr->serv->sflags & SFLAGS_ZIP_OUT)
			  && zip_is_data_out(cptr->serv->zip_out))) {
				PFD_SETW(i);
			}
		}

		waittime = MIN(delay2, delay) * 1000;
		nfds = poll(poll_fdarray, nbr_pfds, waittime);
		if (nfds == -1 && (errno == EINTR || errno == EAGAIN)) {
			return -1;
		}
		else if (nfds >= 0) {
			break;
		}
		report_error("poll %s:%s", &me);
		res++;
		if (res > 5) {
			restart("too many poll errors");
		}
		sleep(10);
	}

	i = 0;
	for (auth = connauth_list; auth != NULL; auth = next) {
		next = auth->next;
		if (poll_fdarray[i].revents) {
			if (auth->flags & CAF_ID_CONNECTING) {
				connauth_send_id(auth);
			}
			else {
				connauth_read_id(auth);
			}
			if (--nfds == 0) {
				break;
			}
		}
		i++;
	}
	if (check_listeners == 1) {
		for (l = listener_list; l != NULL; l = l->next) {
			if (poll_fdarray[i].revents & (POLLREADFLAGS|POLLERRORS)) {
				accept_connection(l);
				nfds--;
			}
			i++;
		}
	}
	for (pfd = &poll_fdarray[i]; i < nbr_pfds; i++, pfd++) {
		fd = pfd->fd;
		if (nfds > 0 && pfd->revents) {
			nfds--;
			rr = pfd->revents & (POLLREADFLAGS|POLLERRORS);
			rw = pfd->revents & POLLWRITEFLAGS;
			if (!(cptr = local[fd])) {
				continue;
			}
			if (rw) {
				int write_err = 0;

				if (IsConnecting(cptr)) {
					write_err = completed_connection(cptr);
				}
				if (write_err == 0) {
					send_queued(cptr);
				}
				if (IsDead(cptr) || write_err) {
					ircsprintf(errmsg, "Write Error: %s",
						cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
					exit_client(cptr, cptr, &me, errmsg);
					continue;
				}
			}

			length = 1;
			if (rr) {
				length = read_packet(cptr);
			}
			else if (IsPerson(cptr) && !NoNewLine(cptr)) {
				length = do_client_queue(cptr);
			}
		}
		else {
			if (!(cptr = local[i])) {
				continue;
			}
			if (IsPerson(cptr) && !NoNewLine(cptr)) {
				length = do_client_queue(cptr);
			}
			else {
				continue;
			}
		}

#ifdef DEBUGMODE
		readcalls++;
#endif

		if (length == FLUSH_BUFFER) {
			continue;
		}
		if (IsDead(cptr)) {
			ircsprintf(errmsg, "Read/Dead Error: %s",
				cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
			exit_client(cptr, cptr, &me, errmsg);
			continue;
		}
		if (length > 0) {
			continue;
		}
		read_error_exit(cptr, length, cptr->sockerr);
	}
	return 0;
}

#endif

void connect_dns_callback(void *vlink, adns_answer *answer)
{
	ConfigItem_link *link = (ConfigItem_link *)vlink;

	link->dns_pending = 0;
	if (answer != NULL && answer->status == adns_s_ok) {
		link->ipnum.s_addr = answer->rrs.addr->addr.inet.sin_addr.s_addr;
		MyFree(answer);
		connect_server(link, NULL, NULL);
	}
	else {
		sendto_realops("Connect to %s failed: host lookup failed.", link->servername);
	}
	return;
}

static int connect_inet(ConfigItem_link *link, aClient *cptr)
{
	static struct sockaddr_in server;

	if (link == NULL || cptr == NULL) {
		ircdlog(LOG_ERROR, "connect_inet() encountered NULL structure link %lX cptr %lX",
			link, cptr);
		return 0;
	}
	if ((cptr->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		report_error("opening stream socket to server %s:%s", cptr);
		return 0;
	}
	if (cptr->fd >= (HARD_FDLIMIT - 10)) {
		sendto_realops("No more connections allowed (%s)", cptr->name);
		return 0;
	}

	memset((char *)&server, '\0', sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = 0;
	server.sin_addr = me.ip;

	if (!BadPtr(link->bindip)) {
		server.sin_addr.s_addr = inet_addr(link->bindip);
	}
	if (bind(cptr->fd, (struct sockaddr *)&server, sizeof(server)) == -1) {
		report_error("error binding to local port for %s:%s", cptr);
		return 0;
	}

	memset((char *)&server, '\0', sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(link->port);
	server.sin_addr.s_addr = link->ipnum.s_addr;

	cptr->port = link->port;
	cptr->ip.s_addr = link->ipnum.s_addr;
	strncpyzt(cptr->sockhost, inetntoa((char *)&cptr->ip), HOSTLEN + 1);

	set_non_blocking(cptr->fd);
	set_sock_opts(cptr->fd, cptr, NULL);

	if (connect(cptr->fd, (struct sockaddr *)&server, sizeof(server)) != 0 &&
	  errno != EINPROGRESS) {
		int errtmp = errno;
		report_error("Connect to host %s failed: %s", cptr);
		errno = errtmp;
		return 0;
	}

	return 1;
}

int connect_server(ConfigItem_link *link, aClient *by, DNSQuery *answer)
{
	aClient *cptr;

	if (link == NULL) {
		ircdlog(LOG_ERROR, "connect_server() with NULL link structure");
		return 0;
	}
	if (link->dns_pending == 1) {
		return 0;
	}

	Debug((DEBUG_NOTICE, "Connect to %s[%s] @%s", link->servername, link->fromip,
		inetntoa((char *)&link->ipnum)));

	if ((cptr = find_server(link->servername, NULL)) != NULL) {
		char *inpath = get_client_name(cptr, HIDEME);
		sendto_realops("Server %s already present from %s.", link->servername, inpath);
		if (by != NULL && IsPerson(by) && !MyClient(by)) {
			sendto_one(by, ":%s NOTICE %s :Server %s already present from %s",
				link->servername, inpath);
		}
		return 0;
	}

	if (link->ipnum.s_addr == 0) {
		if ((link->ipnum.s_addr = inet_addr(link->fromip)) == -1) {
			DNSQuery *q = (DNSQuery *)MyMalloc(sizeof(DNSQuery));
			q->ptr = link;
			q->callback = connect_dns_callback;
			dns_gethost(link->fromip, q);
			link->dns_pending = 1;
			return 1;
		}
	}

	cptr = make_client(NULL, &me);
	cptr->dnsquery = answer;

	strncpyzt(cptr->name, link->servername, sizeof(cptr->name));
	strncpyzt(cptr->sockhost, link->fromip, HOSTLEN + 1);

	if (!connect_inet(link, cptr)) {
		if (by != NULL && IsPerson(by) && !MyClient(by)) {
			sendto_one(by, ":%s NOTICE %s :Connect to server %s failed.", me.name, by->name,
				cptr->name);
		}
		close(cptr->fd);
		cptr->fd = -2;
		free_client(cptr);
		return 0;
	}

	make_server(cptr);
	cptr->serv->conf = link;
	cptr->class = cptr->serv->conf->class;
	cptr->serv->connected_by = (by != NULL && IsPerson(by)) ?
		make_nick_user_host(by->name, by->username, by->host) : make_nick_user_host("AutoConnect", "*", "*");
	cptr->serv->up = me.name;

	if (cptr->fd > highest_fd) {
		highest_fd = cptr->fd;
	}
	local[cptr->fd] = cptr;
	SetConnecting(cptr);

	cptr->sendqlen = cptr->serv->conf->class->sendq;
	add_client_to_list(cptr);
	nextping = timeofday;
	return 1;
}
