/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_conf2.c: server configuration system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_conf2.c,v 1.148 2002/12/22 18:11:45 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "h.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "ssl.h"
#include "conf2.h"
#include "config.h"
#include "memory.h"
#include <time.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <utmp.h>
#include <fcntl.h>

Conf dConf;
char aup_notice[256];
int aup_notice_len = 0;
short rehashing = 0;
int conf_fail = 0;

ConfigCommandEntry *conf = NULL;
ConfigCommandEntry *conf_tail = NULL;
ConfigItem_class *default_class = NULL;

ConfigItem_servinfo *ci_servinfo = NULL;
ConfigItem_admin *ci_admin = NULL;
ConfigItem_class *ci_class = NULL;
ConfigItem_allow *ci_allow = NULL;
ConfigItem_oper *ci_oper = NULL;
ConfigItem_uline *ci_uline = NULL;
ConfigItem_link *ci_link = NULL;
ConfigItem_bind *ci_bind = NULL;
ConfigItem_ban *ci_ban = NULL;
ConfigItem_except *ci_except = NULL;
ConfigItem_vhost *ci_vhost = NULL;
ConfigItem_include *ci_include = NULL;

ConfigItem *parse_servinfo(ConfigEntry *);
ConfigItem *parse_admin(ConfigEntry *);
ConfigItem *parse_class(ConfigEntry *);
ConfigItem *parse_allow(ConfigEntry *);
ConfigItem *parse_oper(ConfigEntry *);
ConfigItem *parse_uline(ConfigEntry *);
ConfigItem *parse_link(ConfigEntry *);
ConfigItem *parse_bind(ConfigEntry *);
ConfigItem *parse_ban(ConfigEntry *);
ConfigItem *parse_except(ConfigEntry *);
ConfigItem *parse_vhost(ConfigEntry *);
ConfigItem *parse_include(ConfigEntry *);

int test_servinfo(ConfigEntry *, ConfigItem *);
int test_class(ConfigEntry *, ConfigItem *);
int test_allow(ConfigEntry *, ConfigItem *);
int test_oper(ConfigEntry *, ConfigItem *);
int test_link(ConfigEntry *, ConfigItem *);
int test_ban(ConfigEntry *, ConfigItem *);
int test_vhost(ConfigEntry *, ConfigItem *);

ConfigItem *free_servinfo(ConfigItem *);
ConfigItem *free_class(ConfigItem *);
ConfigItem *free_allow(ConfigItem *);
ConfigItem *free_oper(ConfigItem *);
ConfigItem *free_link(ConfigItem *);
ConfigItem *free_bind(ConfigItem *);
ConfigItem *free_ban(ConfigItem *);
ConfigItem *free_vhost(ConfigItem *);

extern Auth *die_auth;
extern Auth *restart_auth;
extern char *ircd_config_file;

extern IP_ENTRY *find_or_add_ip(unsigned long);
#ifdef HAVE_THROTTLE_ON
extern int throttle_force(char *);
extern void throttle_rehash();
#endif
extern void build_isupport();

static ConfigCommand ConfigCommands[] = {
	{ "servinfo",	parse_servinfo,	test_servinfo,	free_servinfo },
	{ "admin",	parse_admin,	NULL,		NULL },
	{ "class",	parse_class,	test_class,	free_class },
	{ "allow",	parse_allow,	test_allow,	free_allow },
	{ "oper",	parse_oper,	test_oper,	free_oper },
	{ "uline",	parse_uline,	NULL,		NULL },
	{ "link",	parse_link,	test_link,	free_link },
	{ "bind",	parse_bind,	NULL,		free_bind },
	{ "ban",	parse_ban,	test_ban,	free_ban },
	{ "except",	parse_except,	NULL,		NULL },
	{ "vhost",	parse_vhost,	test_vhost,	free_vhost },
	{ "include",	parse_include,	NULL,		NULL },
	{ NULL, 	NULL,		NULL,		NULL }
};

static ConfigOption servinfo_modesonconnect[] = {
	{ "invisible", UMODE_INVISIBLE },
	{ "wallops", UMODE_WALLOPS },
	{ "masked", UMODE_MASKED },
	{ NULL, 0L }
};

static ConfigOption class_options[] = {
	{ "require-identd", CLASS_REQUIREIDENTD },
	{ "ignore-identd", CLASS_IGNOREIDENTD },
	{ "notilde", CLASS_NOTILDE },
	{ NULL, 0L }
};

static ConfigOption oper_options[] = {
	{ "die", OFLAG_DIE },
	{ "restart", OFLAG_RESTART },
	{ "rehash", OFLAG_REHASH },
	{ "kline", OFLAG_KLINE },
	{ "unkline", OFLAG_UNKLINE },
	{ "zap", OFLAG_ZAP },
	{ "localroute", OFLAG_LROUTE },
	{ "globalroute", OFLAG_GROUTE },
	{ "localkill", OFLAG_LKILL },
	{ "globalkill", OFLAG_GKILL },
	{ "globops", OFLAG_GLOBOPS },
	{ "locops", OFLAG_LOCOPS },
	{ "wallops", OFLAG_WALLOPS },
	{ "localnotice", OFLAG_LNOTICE },
	{ "globalnotice", OFLAG_GNOTICE },
	{ "floodnotices", OFLAG_FNOTICE },
	{ "local-cliconnects", OFLAG_LCLICONN },
	{ "global-cliconnects", OFLAG_GCLICONN },
	{ "netadmin", OFLAG_NETADMIN },
	{ "admin", OFLAG_ADMIN },
	{ "coadmin", OFLAG_COADMIN },
	{ "services-admin", OFLAG_SADMIN },
	{ "routing-staff", OFLAG_RSTAFF },
	{ "showopermotd", OFLAG_OPERMOTD },
	{ "automask", OFLAG_AUTOMASK },
	{ NULL, 0L }
};

static ConfigOption link_options[] = {
	{ "autoconnect", LINK_AUTOCONNECT },
	{ "compress", LINK_COMPRESS },
	{ "secure", LINK_SECURE },
	{ NULL, 0L }
};

static ConfigOption bind_options[] = {
	{ "client-only", BIND_CLIENTONLY },
	{ "server-only", BIND_SERVERONLY },
	{ "secure", BIND_SECURE },
	{ "crjava", BIND_CRJAVA },
	{ NULL, 0L }
};

static ConfigOption ban_types[] = {
	{ "userhost", BAN_USERHOST },
	{ "ip", BAN_IP },
	{ "nickname", BAN_NICKNAME },
	{ "gcos", BAN_GCOS },
	{ "server", BAN_SERVER },
	{ "file", BAN_FILE },
	{ "channel", BAN_CHANNEL },
	{ NULL, 0L }
};

void report(short error, char *string, ...)
{
	va_list vl;
	char buffer[1024];

	va_start(vl, string);
	ircvsprintf(buffer, string, vl);
	va_end(vl);

	if (rehashing == 1) {
		sendto_ops("%s", buffer);
	}
	else {
		fprintf(stderr, "%s\n", buffer);
	}
	if (error == 1) {
		conf_fail++;
		ircdlog(LOG_ERROR, "%s", buffer);
	}
	return;
}

void add_ConfigItem(ConfigItem *item, ConfigItem **list)
{
	item->prev = NULL;
	if ((item->next = *list) != NULL) {
		item->next->prev = item;
	}
	*list = item;
	return;
}

ConfigItem *del_ConfigItem(ConfigItem *item, ConfigItem **list)
{
	ConfigItem *ci, *next;

	for (ci = *list; ci != NULL; ci = ci->next) {
		if (ci == item) {
			next = ci->next;
			if (ci->prev != NULL) {
				ci->prev->next = ci->next;
			}
			else {
				*list = ci->next;
			}
			if (ci->next != NULL) {
				ci->next->prev = ci->prev;
			}
			return next;
		}
	}
	return NULL;
}

static void free_configentry(ConfigEntry *ce)
{
	ConfigEntry *next;
	while (ce != NULL) {
		next = ce->ce_next;
		if (ce->ce_entries != NULL) {
			free_configentry(ce->ce_entries);
		}
		ConfFree(ce->ce_varname);
		ConfFree(ce->ce_vardata);
		MyFree(ce);
		ce = next;
	}
	return;
}

static void free_configfile(ConfigFile *cf)
{
	ConfigFile *next;
	while (cf != NULL) {
		next = cf->cf_next;
		if (cf->cf_entries != NULL) {
			free_configentry(cf->cf_entries);
		}
		ConfFree(cf->cf_filename);
		MyFree(cf);
		cf = next;
	}
	return;
}

static ConfigFile *load_configfile(char *filename, char *confdata)
{
	ConfigFile *curcf, *lastcf;
	ConfigEntry *curce = NULL, *cursect = NULL, **lastce;
	char *ptr, *start;
	int linenum = 1, commentstart = 0, commentlevel = 0;

	lastcf = curcf = (ConfigFile *)MyMallocX(sizeof(ConfigFile));
	ConfDupString(curcf->cf_filename, filename);
	lastce = &(curcf->cf_entries);

	for (ptr = confdata; *ptr != '\0'; ptr++) {
		switch (*ptr) {
			case ';':
				if (curce == NULL) {
					report(0, "%s:%d: ignoring extra semicolon", filename, linenum);
					break;
				}
				*lastce = curce;
				lastce = &(curce->ce_next);
				curce = NULL;
				break;
			case '{':
				if (curce == NULL) {
					report(0, "%s:%d: ignoring blank section-start", filename, linenum);
					break;
				}
				else if (curce->ce_entries != NULL) {
					report(0, "%s:%d: ignoring extra section-start", filename, linenum);
					break;
				}
				curce->ce_sectlinenum = linenum;
				lastce = &(curce->ce_entries);
				cursect = curce;
				curce = NULL;
				break;
			case '}':
				if (curce != NULL) {
					report(1, "%s:%d: ERROR: missing semicolon before section-finish", filename,
						linenum);
					free_configentry(curce);
					free_configfile(curcf);
					return NULL;
				}
				else if (cursect == NULL) {
					report(0, "%s:%d: ignoring extra section-finish", filename, linenum);
					break;
				}
				curce = cursect;
				cursect = cursect->ce_prev;
				lastce = (cursect == NULL) ? &(curcf->cf_entries) : &(cursect->ce_entries);
				while (*lastce != NULL) {
					lastce = &((*lastce)->ce_next);
				}
				break;
			case '#':
				ptr++;
				while (*ptr != '\0' && *ptr != '\n') {
					ptr++;
				}
				if (*ptr == '\0') {
					break;
				}
				ptr--;
				break;
			case '/':
				if (*(ptr + 1) == '/') {
					ptr += 2;
					while (*ptr != '\0' && *ptr != '\n') {
						ptr++;
					}
					if (*ptr == '\0') {
						break;
					}
					ptr--;
				}
				else if (*(ptr + 1) == '*') {
					commentstart = linenum;
					commentlevel = 1;

					for (ptr += 2; *ptr; ptr++) {
						if (*ptr == '/' && *(ptr + 1) == '*') {
							commentlevel++;
							ptr++;
						}
						else if (*ptr == '*' && *(ptr + 1) == '/') {
							commentlevel--;
							ptr++;
						}
						else if (*ptr == '\n') {
							linenum++;
						}
						if (commentlevel == 0) {
							break;
						}
					}
					if (*ptr == '\0') {
						report(1, "%s:%d: ERROR: unexpected EOF: possible unterminated comment "
							"at line %d", filename, linenum, commentstart);
						free_configentry(curce);
						free_configfile(curcf);
						return NULL;
					}
				}
				break;
			case '\"':
				start = ++ptr;
				while (*ptr != '\0') {
					if (*ptr == '\\' && *(ptr + 1) == '\"') {
						char *ptr2 = ptr;
						while ((*ptr2 = *(ptr2 + 1))) {
							ptr2++;
						}
					}
					else if (*ptr == '\"' || *ptr == '\n') {
						break;
					}
					ptr++;
				}
				if (*ptr == '\0' || *ptr == '\n') {
					report(1, "%s:%d: ERROR: unexpected EOF: possible unterminated double-quote",
						filename, linenum);
					free_configentry(curce);
					free_configfile(curcf);
					return NULL;
				}
				if (curce != NULL) {
					if (curce->ce_vardata != NULL) {
						report(0, "%s:%d: ignoring extra data", filename, linenum);
						break;
					}
					else {
						curce->ce_vardata = (char *)MyMalloc((ptr - start) + 1);
						strncpy(curce->ce_vardata, start, ptr - start);
						curce->ce_vardata[ptr - start] = '\0';
					}
				}
				else {
					curce = (ConfigEntry *)MyMallocX(sizeof(ConfigEntry));
					curce->ce_varname = (char *)MyMalloc((ptr - start) + 1);
					strncpy(curce->ce_varname, start, ptr - start);
					curce->ce_varname[ptr - start] = '\0';
					curce->ce_fileptr = curcf;
					curce->ce_varlinenum = linenum;
					curce->ce_prev = cursect;
				}
				break;
			case '\n':
				linenum++;
			case '\t':
			case '\r':
			case '=':
			case ' ':
				break;
			default:
				if (*ptr == '*' && *(ptr + 1) == '/') {
					report(0, "%s:%d: ignoring extra close-comment", filename, linenum);
					ptr++;
					break;
				}
				start = ptr;
				while (*ptr != '\0') {
					if (*ptr == ' ' || *ptr == '=' || *ptr == '\t' || *ptr == '\n' || *ptr == ';') {
						break;
					}
					ptr++;
				}
				if (*ptr == '\0') {
					if (curce != NULL) {
						report(1, "%s:%d: ERROR: unexpected EOF: unterminated variable at line %d",
							filename, linenum, curce->ce_varlinenum);
					}
					else if (cursect != NULL) {
						report(1, "%s:%d: ERROR: unexpected EOF: unterminated section starting "
							"at line %d", filename, linenum, curce->ce_sectlinenum);
					}
					else {
						report(1, "%s:%d: ERROR: unexpected EOF: unknown cause", filename,
							linenum);
					}
					free_configentry(curce);
					free_configfile(curcf);
					return NULL;
				}
				if (curce != NULL) {
					if (curce->ce_vardata != NULL) {
						report(0, "%s:%d: ignoring extra data", filename, linenum);
					}
					else {
						curce->ce_vardata = (char *)MyMalloc((ptr - start) + 1);
						strncpy(curce->ce_vardata, start, ptr - start);
						curce->ce_vardata[ptr - start] = '\0';
					}
				}
				else {
					curce = (ConfigEntry *)MyMallocX(sizeof(ConfigEntry));
					curce->ce_varname = (char *)MyMalloc((ptr - start) + 1);
					strncpy(curce->ce_varname, start, ptr - start);
					curce->ce_varname[ptr - start] = '\0';
					curce->ce_fileptr = curcf;
					curce->ce_varlinenum = linenum;
					curce->ce_prev = cursect;
				}
				if (*ptr == ';' || *ptr == '\n') {
					ptr--;
				}
				break;
		}
	}
	if (curce != NULL) {
		report(1, "%s: ERROR: unexpected EOF: unterminated variable at line %d", filename,
			curce->ce_varlinenum);
		free_configentry(curce);
		free_configfile(curcf);
		return NULL;
	}
	else if (cursect != NULL) {
		report(1, "%s: ERROR: unexpected EOF: unterminated section starting at line %d", filename,
			cursect->ce_sectlinenum);
		free_configfile(curcf);
		return NULL;
	}
	return curcf;
}

static ConfigFile *open_configfile(char *filename)
{
	int fd, len, i = (rehashing == 0) ? 1 : 0;
	struct stat sb;
	ConfigFile *cf;
	char *confdata, *err = (rehashing == 0) ? "ERROR: failed" : "Failed";

	if (BadPtr(filename)) {
		report(0, "Ignoring null config file name.");
		return NULL;
	}

	report(0, "Attempting to open config file %s...", filename);
	if ((fd = open(filename, O_RDONLY)) == -1) {
		report(i, "%s to open config file %s: %s.", err, filename, strerror(errno));
		return NULL;
	}
	if (fstat(fd, &sb) == -1) {
		report(i, "%s to fstat config file %s: %s.", err, filename, strerror(errno));
		close(fd);
		return NULL;
	}
	if (sb.st_size <= 0) {
		report(0, "Ignoring empty config file %s.", filename);
		close(fd);
		return NULL;
	}

	confdata = (char *)MyMalloc(sb.st_size + 1);
	if ((len = read(fd, confdata, sb.st_size)) != sb.st_size) {
		report(i, "%s to read config file %s: %s.", err, filename,
			len == -1 ? strerror(errno) : strerror(EFAULT));
		MyFree(confdata);
		close(fd);
		return NULL;
	}
	confdata[len] = '\0';
	close(fd);
	cf = load_configfile(filename, confdata);
	MyFree(confdata);
	return cf;
}

static ConfigCommand *get_config_command(char *data)
{
	ConfigCommand *cc;
	for (cc = ConfigCommands; cc->name != NULL; cc++) {
		if (!mycmp(cc->name, data)) {
			return cc;
		}
	}
	return NULL;
}

static void parse_configfile(ConfigFile *cf)
{
	ConfigCommandEntry *cce, *next = NULL;
	ConfigCommand *cc;
	ConfigEntry *ce;
	ConfigItem *ci;

	if (cf->cf_entries == NULL) {
		report(0, "%s: ignoring empty file", cf->cf_filename);
		return;
	}
	for (ce = cf->cf_entries; ce != NULL; ce = ce->ce_next) {
		if (BadPtr(ce->ce_varname)) {
			report(0, "%s:%d: ignoring null variable", ce->ce_fileptr->cf_filename,
				ce->ce_varlinenum);
			continue;
		}
		if ((cc = get_config_command(ce->ce_varname)) == NULL) {
			report(0, "%s:%d: ignoring unknown block %s", ce->ce_fileptr->cf_filename,
				ce->ce_varlinenum, ce->ce_varname);
			continue;
		}
		if ((ci = cc->parse_func(ce)) != NULL) {
			if (cc->test_func == NULL || cc->free_func == NULL) {
				continue;
			}
			cce = (ConfigCommandEntry *)MyMallocX(sizeof(ConfigCommandEntry));
			cce->cc = cc;
			cce->ce = ce;
			cce->ci = ci;
			cce->prev = NULL;
			if ((cce->next = conf) != NULL) {
				cce->next->prev = cce;
			}
			if (conf_tail == NULL) {
				conf_tail = cce;
			}
			conf = cce;
		}
	}
	if (conf_fail == 0) {
		report(0, "Validating data in config file %s...", cf->cf_filename);
		for (cce = conf_tail; cce != NULL; cce = cce->prev) {
			if (!cce->cc->test_func(cce->ce, cce->ci)) {
				cce->cc->free_func(cce->ci);
			}
		}
		if (conf_fail > 0) {
			report(0, "ERROR: failed to validate config file %s.", cf->cf_filename);
		}
		else {
			report(0, "Successfully validated config file %s.", cf->cf_filename);
		}
	}
	for (cce = conf; cce != NULL; cce = next) {
		next = cce->next;
		if (cce->prev != NULL) {
			cce->prev->next = cce->next;
		}
		else {
			conf = cce->next;
			conf_tail = NULL;
		}
		if (cce->next != NULL) {
			cce->next->prev = cce->prev;
		}
		ConfFree(cce);
	}
	return;
}

static void conf_init(char *filename)
{
	ConfigFile *cf;
	ConfigItem_include *ci;

	chmod(filename, 0600);
	if ((cf = open_configfile(filename)) != NULL) {
		report(0, "Successfully loaded config file %s.", filename);
		parse_configfile(cf);
		free_configfile(cf);

		for (ci = ci_include; ci != NULL; ci = (ConfigItem_include *)ci->next) {
			if (ci->loaded == 0) {
				ci->loaded = 1;
				conf_init(ci->filename);
			}
		}
	}
	return;
}

int conf_ipaddr(char *ipaddr, int at)
{
	char *s;
	int reached = 0;

	for (s = ipaddr; *s != '\0'; s++) {
		if (at == 1 && reached == 0) {
			if (*s == '@') {
				reached = 1;
			}
			continue;
		}
		if (*s != '.' && !isdigit(*s)) {
			return 0;
		}
	}
	return 1;
}

long conf_size(char *val)
{
	char *numbuf, *p = val;
	long ret = 0;
	int i;

	if (val == NULL) {
		return 0;
	}

	numbuf = (char *)MyMalloc(strlen(val) + 1);
	for (i = 0; *p != '\0'; p++) {
		if (isdigit(*p)) {
			numbuf[i++] = *p;
			continue;
		}
		if (isalpha(*p)) {
			ret = atol(numbuf);
			if (tolower(*p) == 't') {
				ret *= 1099511627776;
			}
			else if (tolower(*p) == 'g') {
				ret *= 1073741824;
			}
			else if (tolower(*p) == 'm') {
				ret *= 1048576;
			}
			else if (tolower(*p) == 'k') {
				ret *= 1024;
			}
			break;
		}
	}
	if (ret == 0) {
		ret = atol(numbuf);
	}
	MyFree(numbuf);
	return ret;
}

char *conf_bantype(long flag)
{
	ConfigOption *co;
	for (co = ban_types; co->name != NULL; co++) {
		if (flag == co->flag) {
			return co->name;
		}
	}
	return "unknown";
}

int conf_doban(aClient *cptr, aClient *sptr, ConfigItem_ban *ban, char *zhost, int zfd)
{
	char *bantype, *msgtype, *inpath, zstring[512];
	int zlen = 0, network = 0, gcos = 0;

	if (ban->type & BAN_USERHOST) {
		if (ban->type & BAN_USERHOST_S) {
			bantype = "AutoKill";
			msgtype = "AutoKilled";
			network = 1;
		}
		else {
			bantype = (ban->flag.temporary == 1) ? "Temporary K:Line" :
				(ban->expire > 0) ? "Timed K:Line" : "Permanent K:Line";
			msgtype = "K:Lined";
		}
	}
	else if (ban->type & BAN_IP) {
		if (ban->type & BAN_IP_S) {
			bantype = "SZ:Line";
			msgtype = "SZ:Lined";
			network = 1;
		}
		else {
			bantype = (ban->flag.temporary == 1) ? "Temporary Z:Line" : "Permanent Z:Line";
			msgtype = "Z:Lined";
		}
	}
	else if (ban->type & BAN_GCOS) {
		bantype = (ban->type & BAN_GCOS_S) ? "SG:Line" : "G:Line";
		msgtype = (ban->type & BAN_GCOS_S) ? "SG:Lined" : "G:Lined";
		network = (ban->type & BAN_GCOS_S) ? 1 : 0;
		gcos = 1;
	}
	else {
		bantype = "Ban";
		msgtype = "Banned";
	}

	ircstp->is_ref++;
	if (zfd >= 0) {
		ircdlog(LOG_BAN, "%s active for *@%s", bantype, zhost);
		sendto_realops_lev(REJ_LEV, "%s active for *@%s (%s)", bantype, zhost, ban->reason);
		zlen = ircsnprintf(zstring, 512, "ERROR :Your host has been %s: %s.\r\n", msgtype, ban->reason);
		write(zfd, zstring, zlen);
		return 0;
	}
	inpath = get_client_name(sptr, TRUE);
	ircdlog(LOG_BAN, "%s active for %s", bantype, inpath);
	sendto_realops_lev(REJ_LEV, "%s active for %s (%s)", bantype, inpath, ban->reason);
	if (gcos == 1) {
		sendto_one(sptr, ":%s NOTICE %s :*** Your GCOS (realname) is not permitted on this %s: %s",
			me.name, sptr->name, network == 1 ? "network" : "server", ban->reason);
	}
	else {
		sendto_one(sptr, ":%s NOTICE %s :*** You are not welcome on this %s: %s.", me.name, sptr->name,
			network == 1 ? "network" : "server", ban->reason);
	}
	sendto_one(sptr, ":%s NOTICE %s :*** For more information, mail %s and include everything shown "
		"here.", me.name, sptr->name, network == 1 ? IRCNETWORK_KLINEADDRESS : KLINEADDRESS);
	if (ban->expire > 0) {
		sendto_one(sptr, ":%s NOTICE %s :*** This %s will expire in %s.", me.name, cptr->name,
			msgtype, secstodhms((ban->setat - timeofday) + ban->expire));
	}
#ifdef HAVE_THROTTLE_ON
	if (THROTTLE_REJECTED_CLIENTS == 1) {
		if (throttle_force(sptr->host)) {
			sendto_one(sptr, ":%s NOTICE %s :*** Your host has now been force-throttled.",
				me.name, sptr->name);
		}
	}
#endif
	sendto_one(sptr, err_str(ERR_YOUREBANNEDCREEP), me.name, sptr->name, msgtype);
	return exit_client(cptr, sptr, &me, msgtype);
}

void check_dynconf()
{
	char *p;
	if (BadPtr(IRCNETWORK)) {
		ConfDupString(IRCNETWORK, "IRC");
		ConfDupString(IRCNETWORK005, "IRC");
		report(0, "servinfo::network::name: missing parameter, using %s", IRCNETWORK);
	}
	if (BadPtr(IRCNETWORK_KLINEADDRESS)) {
		ConfDupString(IRCNETWORK_KLINEADDRESS, "poorly@configured.network");
		report(0, "servinfo::network::kline-address: missing parameter, using %s",
			IRCNETWORK_KLINEADDRESS);
	}
	else {
		for (p = IRCNETWORK_KLINEADDRESS; *p; p++) {
			if (*p == '@' || *p == ':') {
				break;
			}
		}
		if (*p == '\0') {
			ConfDupString(IRCNETWORK_KLINEADDRESS, "poorly@configured.network");
			report(0, "servinfo::network::kline-address: invalid parameter (bogus e-mail "
				"address/url), using %s", IRCNETWORK_KLINEADDRESS);
		}
	}
	if (BadPtr(NETADMIN_MASK)) {
		p = (char *)MyMalloc(strlen(ci_servinfo->name) + 10);
		ircsprintf(p, "netadmin.%s", ci_servinfo->name);
		ConfDupString(NETADMIN_MASK, p);
		MyFree(p);
		report(0, "servinfo::masking::netadmin: missing parameter, using %s", NETADMIN_MASK);
	}
	if (BadPtr(ADMIN_MASK)) {
		p = (char *)MyMalloc(strlen(ci_servinfo->name) + 7);
		ircsprintf(p, "netadmin.%s", ci_servinfo->name);
		ConfDupString(ADMIN_MASK, p);
		MyFree(p);
		report(0, "servinfo::masking::admin: missing parameter, using %s", ADMIN_MASK);
	}
	if (BadPtr(COADMIN_MASK)) {
		p = (char *)MyMalloc(strlen(ci_servinfo->name) + 9);
		ircsprintf(p, "coadmin.%s", ci_servinfo->name);
		ConfDupString(COADMIN_MASK, p);
		MyFree(p);
		report(0, "servinfo::masking::coadmin: missing parameter, using %s", COADMIN_MASK);
	}
	if (BadPtr(OPER_MASK)) {
		p = (char *)MyMalloc(strlen(ci_servinfo->name) + 6);
		ircsprintf(p, "oper.%s", ci_servinfo->name);
		ConfDupString(OPER_MASK, p);
		MyFree(p);
		report(0, "servinfo::masking::oper: missing parameter, using %s", OPER_MASK);
	}
	if (BadPtr(USER_PREFIX)) {
		ConfDupString (USER_PREFIX, "masked");
		report(0, "servinfo::masking::user-prefix: missing parameter, using %s", USER_PREFIX);
	}
	return;
}

void free_dynconf()
{
	ConfFree(AUTOJOINCHANS);
	ConfFree(KLINEADDRESS);
	ConfFree(IRCNETWORK);
	ConfFree(IRCNETWORK_KLINEADDRESS);
	ConfFree(AUP_URL);
	ConfFree(SERVICES_SERVER);
	ConfFree(STATS_SERVER);
	ConfFree(NETADMIN_MASK);
	ConfFree(ADMIN_MASK);
	ConfFree(COADMIN_MASK);
	ConfFree(OPER_MASK);
	ConfFree(USER_PREFIX);
	ConfFree(PROXYMONITOR_URL);
	ConfFree(PROXYMONITOR_HOST);
	return;
}

void init_conf()
{
	if (default_class == NULL) {
		default_class = (ConfigItem_class *)MyMallocX(sizeof(ConfigItem_class));
		DupString(default_class->name, "default");
		default_class->pingtime = DEFAULT_PINGTIME;
		default_class->connectfreq = DEFAULT_CONNECTFREQ;
		default_class->maxclients = DEFAULT_MAXCLIENTS;
		default_class->sendq = MAXSENDQLENGTH;
		default_class->flag.temporary = 0;
		default_class->flag.permanent = 1;
	}
	if (ci_include == NULL) {
		ci_include = (ConfigItem_include *)MyMallocX(sizeof(ConfigItem_include));
		ConfDupString(ci_include->filename, ircd_config_file);
		ci_include->loaded = 1;
	}

	memset(&dConf, '\0', sizeof(Conf));
	conf_init(ircd_config_file);

	if (ci_servinfo == NULL) {
		report(1, "ERROR: missing servinfo{} configuration.");
	}
	if (ci_allow == NULL) {
		report(1, "ERROR: missing allow{} configuration.");
	}
	if (conf_fail == 0) {
		check_dynconf();
		if (conf_fail == 0) {
			*aup_notice = '\0';
			aup_notice_len = 0;
			if (!BadPtr(AUP_URL)) {
				aup_notice_len = ircsnprintf(aup_notice, 256, "NOTICE AUTH :*** Connecting to this "
					"server constitutes agreement with our acceptable use policy, available "
					"at %s. If you do not agree with this policy, you must disconnect "
					"immediately.\r\n", AUP_URL);
			}
		}
	}
	return;
}

void run_listeners(void)
{
	ConfigItem_bind *ci;
	Listener *l = NULL;

	for (ci = ci_bind; ci != NULL; ci = (ConfigItem_bind *)ci->next) {
		if ((l = find_listener(ci->ip, ci->port)) == NULL) {
			add_listener(ci);
		}
		else {
			l->bind = ci;
			l->options = ci->options;
		}
	}
	return;
}

void conf_cleanup(void)
{
	ConfigItem_class *class;
	ConfigItem_link *link;
	ConfigItem_bind *bind;
	ConfigItem *next = NULL;
	int i = 0;

	for (class = ci_class; class != NULL; class = (ConfigItem_class *)next) {
		next = (class->flag.temporary == 1 && class->clients <= 0) ?
			free_class((ConfigItem *)class) : class->next;
	}
	for (link = ci_link; link != NULL; link = (ConfigItem_link *)next) {
		next =  (link->flag.temporary == 1 && link->refcount <= 0) ?
			free_link((ConfigItem *)link) : link->next;
	}
	for (bind = ci_bind; bind != NULL; bind = (ConfigItem_bind *)next) {
		if (bind->flag.permanent == 0 && bind->flag.temporary == 1 && bind->clients <= 0) {
			next = free_bind((ConfigItem *)bind);
			i = 1;
		}
		else {
			next = bind->next;
		}
	}
	if (i == 1) {
		close_listeners();
	}
	return;
}

int AllowClient(aClient *cptr, char *sockhost)
{
	ConfigItem_allow *ci;
	IP_ENTRY *ip_found;
	char *p;
	int ulen = 0, attach = 0;
	static char uhost[USERLEN + HOSTLEN + 2], uhost2[USERLEN + HOSTLEN + 2];

	for (ci = ci_allow; ci != NULL; ci = (ConfigItem_allow *)ci->next) {
		if (ci->class->flag.temporary == 1) {
			continue;
		}

		for (ulen = 0, p = ci->hostname; *p; p++) {
			if (*p == '@') {
				ulen = ircsprintf(uhost, "%s@", cptr->username);
				break;
			}
		}
		if (ulen == 0) {
			*uhost = '\0';
		}
		strncat(uhost, sockhost, (USERLEN + HOSTLEN + 2) - ulen);

		for (ulen = 0, p = ci->ip; *p; p++) {
			if (*p == '@') {
				ulen = ircsprintf(uhost, "%s@", cptr->username);
				break;
			}
		}
		if (ulen == 0) {
			*uhost2 = '\0';
		}
		strncat(uhost2, cptr->host, (USERLEN + HOSTLEN + 2) - ulen);

		if (!match(ci->hostname, uhost) || !match(ci->ip, uhost2)) {
			attach = 1;
			break;
		}
	}
	if (attach == 1 && ci != NULL) {
		ip_found = find_or_add_ip(cptr->ip.s_addr);
		cptr->flags |= FLAGS_HASHED_IP;
		ip_found->count++;

		if (ci->maxperip > 0 && ip_found->count > ci->maxperip) {
			return CLIENTAUTH_TOOMANYIPS;
		}
		if (ci->auth != NULL && (BadPtr(cptr->passwd) || !auth_check(ci->auth, cptr->passwd))) {
			return CLIENTAUTH_INVALIDPW;
		}
		if (!BadPtr(cptr->passwd)) {
			memset(cptr->passwd, '\0', sizeof(cptr->passwd));
		}
		if (ci->class->clients + 1 > ci->class->maxclients) {
			if (!BadPtr(ci->redirect_server) && ci->redirect_port > 0) {
				sendto_one(cptr, rpl_str(RPL_REDIR), me.name, cptr->name,
					ci->redirect_server, ci->redirect_port);
			}
			return CLIENTAUTH_ILINEFULL;
		}
		cptr->class = ci->class;
		cptr->class->clients++;
		return CLIENTAUTH_MATCHED;
	}
	return CLIENTAUTH_NOMATCH;
}

void rehash(int signal)
{
	ConfigItem_class *class;
	ConfigItem_allow *allow;
	ConfigItem_oper *oper;
	ConfigItem_uline *uline;
	ConfigItem_link *link;
	ConfigItem_bind *bind;
	ConfigItem_ban *ban;
	ConfigItem_except *except;
	ConfigItem_vhost *vhost;
	ConfigItem_include *include;
	ConfigItem *next = NULL;
	aClient *acptr;
	int i = 0;

	rehashing = 1;
	if (signal == SIGHUP) {
		sendto_ops("Got signal SIGHUP, reloading server configuration.");
		ircdlog(LOG_DEFAULT, "got signal SIGHUP, reloading server configuration");
#ifdef ULTRIX
		if (form() > 0) {
			exit(0);
		}
		write_pidfile();
#endif
	}

	free_dynconf();
	auth_destroy_struct(die_auth);
	auth_destroy_struct(restart_auth);

	if (ci_servinfo != NULL) {
		free_servinfo((ConfigItem *)ci_servinfo);
	}
	if (ci_admin != NULL) {
		ConfFree(ci_admin->line1);
		ConfFree(ci_admin->line2);
		ConfFree(ci_admin->line3);
		ConfFree(ci_admin);
	}
	for (class = ci_class; class != NULL; class = (ConfigItem_class *)class->next) {
		if (class->flag.permanent == 0) {
			class->flag.temporary = 1;
		}
	}
	for (allow = ci_allow; allow != NULL; allow = (ConfigItem_allow *)next) {
		next = free_allow((ConfigItem *)allow);
	}
	for (oper = ci_oper; oper != NULL; oper = (ConfigItem_oper *)next) {
		next = free_oper((ConfigItem *)oper);
	}
	for (uline = ci_uline; uline != NULL; uline = (ConfigItem_uline *)next) {
		next = del_ConfigItem((ConfigItem *)uline, (ConfigItem **)&ci_uline);
		ConfFree(uline->servername);
		MyFree(uline);
	}
	for (link = ci_link; link != NULL; link = (ConfigItem_link *)link->next) {
		link->flag.temporary = 1;
	}
	for (bind = ci_bind; bind != NULL; bind = (ConfigItem_bind *)bind->next) {
		if (bind->flag.permanent == 0) {
			bind->flag.temporary = 1;
		}
	}
	for (ban = ci_ban; ban != NULL; ban = (ConfigItem_ban *)next) {
		if (!(ban->type & BAN_NICKNAME_S) && !(ban->type & BAN_IP_S) && !(ban->type & BAN_GCOS_S)
		  && ban->flag.remove_on_rehash == 1) {
			next = free_ban((ConfigItem *)ban);
		}
		else {
			next = ban->next;
		}
	}
	for (except = ci_except; except != NULL; except = (ConfigItem_except *)next) {
		next = del_ConfigItem((ConfigItem *)except, (ConfigItem **)&ci_except);
		ConfFree(except->mask);
		MyFree(except);
	}
	for (vhost = ci_vhost; vhost != NULL; vhost = (ConfigItem_vhost *)next) {
		next = free_vhost((ConfigItem *)vhost);
	}
	for (include = ci_include; include != NULL; include = (ConfigItem_include *)next) {
		next = del_ConfigItem((ConfigItem *)include, (ConfigItem **)&ci_include);
		MyFree(include->filename);
	}

	init_conf();
#ifndef STATIC_MODULES
	check_Modules();
#endif
	close_listeners();
	conf_cleanup();
	close_listeners();
	run_listeners();

	build_isupport();
#ifdef HAVE_THROTTLE_ON
	throttle_rehash();
#endif
	load_message_file(&motd);
	load_message_file(&opermotd);
	load_message_file(&rules);

	for (i = 0; i <= highest_fd; i++) {
		if ((acptr = local[i]) && !IsMe(acptr)) {
			if (IsRegistered(acptr)) {
				acptr->pingval = (acptr->class != NULL) ?
					acptr->class->pingtime : default_class->pingtime;
			}
			acptr->sendqlen = (acptr->class != NULL) ? acptr->class->sendq : default_class->sendq;
		}
	}

	me.ip.s_addr = (!BadPtr(ci_servinfo->bindip)) ? inet_addr(ci_servinfo->bindip) : INADDR_ANY;
	rehashing = 0;
	sendto_ops("Rehash complete.");
	do_bancheck = 1;
	return;
}

ConfigItem *parse_servinfo(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp, *ceppp;
	ConfigOption *co;
	char *tmp;

	CheckBlockEntries(ce, "servinfo");
	if (ci_servinfo == NULL) {
		ci_servinfo = (ConfigItem_servinfo *)MyMallocX(sizeof(ConfigItem_servinfo));
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "servinfo");
		if (!mycmp(cep->ce_varname, "name")) {
			CheckParam(cep, "servinfo::name");
			ConfDupString(ci_servinfo->name, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "desc")) {
			CheckParam(cep, "servinfo::desc");
			ConfDupString(ci_servinfo->desc, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "default-bindip")) {
			CheckParam(cep, "servinfo::default-bindip");
			ConfDupString(ci_servinfo->bindip, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "floodcount")) {
			CheckParam(cep, "servinfo::floodcount");
			FLOODCOUNT = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "knockdelay")) {
			CheckParam(cep, "servinfo::knockdelay");
			KNOCKDELAY = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "max-chans-per-user")) {
			CheckParam(cep, "servinfo::max-chans-per-user");
			MAXCHANSPERUSER = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "max-clients")) {
			CheckParam(cep, "servinfo::max-clients");
			MAXCLIENTS = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "linkdepth")) {
			CheckParam(cep, "servinfo::linkdepth");
			LINKDEPTH = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "accept-notice-time")) {
			CheckParam(cep, "servinfo::accept-notice-time");
			ACCEPTNOTICETIME = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "kline-address")) {
			CheckParam(cep, "servinfo::kline-address");
			ConfDupString(KLINEADDRESS, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "autojoin")) {
			CheckParam(cep, "servinfo::autojoin");
			ConfDupString(AUTOJOINCHANS, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "statsfile-save-freq")) {
			CheckParam(cep, "servinfo::statsfile-save-freq");
			STATSFILE_SAVE_FREQ = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "anti-spam-quit-message-time")) {
			CheckParam(cep, "servinfo::anti-spam-quit-message-time");
			ANTI_SPAM_QUIT_MESSAGE_TIME = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "connauth-timeout")) {
			CheckParam(cep, "servinfo::connauth-timeout");
			CONNAUTH_TIMEOUT = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "options")) {
			CheckSection(cep, "servinfo::options");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::options");
				if (!mycmp(cepp->ce_varname, "short-motd")) {
					SHORTMOTD = 1;
				}
				else if (!mycmp(cepp->ce_varname, "hide-ulines")) {
					HIDEULINES = 1;
				}
				else if (!mycmp(cepp->ce_varname, "check-identd")) {
					CHECKIDENTD = 1;
				}
				else if (!mycmp(cepp->ce_varname, "show-headers")) {
					SHOWHEADERS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "no-mixed-case")) {
					NOMIXEDCASE = 1;
				}
				else if (!mycmp(cepp->ce_varname, "fail-oper-notice")) {
					FAILOPERNOTICE = 1;
				}
				else if (!mycmp(cepp->ce_varname, "ignore-first-char")) {
					IGNOREFIRSTCHAR = 1;
				}
				else if (!mycmp(cepp->ce_varname, "logging")) {
					LOGGING = 1;
				}
				else if (!mycmp(cepp->ce_varname, "dynamic-channels")) {
					DYNCHANS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "no-oper-flood")) {
					NOOPERFLOOD = 1;
				}
				else if (!mycmp(cepp->ce_varname, "spy-notices")) {
					SPYNOTICES = 1;
				}
				else if (!mycmp(cepp->ce_varname, "allow-knocks")) {
					ALLOWKNOCKS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "custom-quit-messages")) {
					CUSTOMQUITMESSAGES = 1;
				}
				else if (!mycmp(cepp->ce_varname, "usernames-in-trace")) {
					USERNAMES_IN_TRACE = 1;
				}
				else if (!mycmp(cepp->ce_varname, "show-invisible-lusers")) {
					SHOW_INVISIBLE_LUSERS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "anti-nick-flood")) {
					ANTINICKFLOOD = 1;
				}
				else if (!mycmp(cepp->ce_varname, "anti-away-flood")) {
					ANTIAWAYFLOOD = 1;
				}
				else if (!mycmp(cepp->ce_varname, "write-perm-klines-to-file")) {
					WRITE_PERM_KLINES_TO_FILE = 1;
				}
				else if (!mycmp(cepp->ce_varname, "flatten-links")) {
					FLATTEN_LINKS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "anti-spambot-checks")) {
					ANTI_SPAMBOT_CHECKS = 1;
				}
				else if (!mycmp(cepp->ce_varname, "anti-acebot-checks")) {
					ANTI_ACEBOT_CHECKS = 1;
				}
#ifdef HAVE_THROTTLE_ON
				else if (!mycmp(cepp->ce_varname, "throttle-rejected-clients")) {
					THROTTLE_REJECTED_CLIENTS = 1;
				}
#endif
				else if (!mycmp(cepp->ce_varname, "allow-map")) {
					ALLOWMAP = 1;
				}
				else {
					UnknownVar(cepp, "servinfo::options");
				}
			}
		}
		else if (!mycmp(cep->ce_varname, "modes-on-connect")) {
			CheckSection(cep, "servinfo::modes-on-connect");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::modes-on-connect");
				for (co = servinfo_modesonconnect; co->name != NULL; co++) {
					if (!mycmp(co->name, cepp->ce_varname)) {
						dConf.modesonconnect |= co->flag;
						break;
					}
				}
				if (co->name == NULL) {
					UnknownVar(cepp, "servinfo::modes-on-connect");
				}
			}
		}
#ifndef STATIC_MODULES
		else if (!mycmp(cep->ce_varname, "modules-path")) {
			CheckParam(cep, "servinfo::modules-path");
			load_modules_from_path(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "modules")) {
			CheckSection(cep, "servinfo::modules");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::modules");
				handle_module_call(cepp->ce_varname, NULL, MOD_CALLTYPE_LOAD);
			}
		}
#endif
		else if (!mycmp(cep->ce_varname, "auth")) {
			CheckParam(cep, "servinfo::auth");
			if (!mycmp(cep->ce_vardata, "die")) {
				die_auth = auth_parse_conf_block(cep, "servinfo::auth die");
			}
			else if (!mycmp(cep->ce_vardata, "restart")) {
				restart_auth = auth_parse_conf_block(cep, "servinfo::auth restart");
			}
			else {
				UnknownVar(cep, "servinfo::auth");
			}
		}
#ifdef HAVE_THROTTLE_ON
		else if (!mycmp(cep->ce_varname, "throttle")) {
			CheckSection(cep, "servinfo::throttle");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::throttle");
				if (!mycmp(cepp->ce_varname, "count")) {
					CheckParam(cepp, "servinfo::throttle::count");
					THROTTLE_COUNT = atoi(cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "time")) {
					CheckParam(cepp, "servinfo::throttle::time");
					THROTTLE_TIME = atime(cepp->ce_vardata);
				}
				else {
					UnknownVar(cepp, "servinfo::throttle");
				}
			}
		}
#endif
		else if (!mycmp(cep->ce_varname, "pacetime")) {
			CheckSection(cep, "servinfo::pacetime");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::pacetime");
				if (!mycmp(cepp->ce_varname, "simple")) {
					CheckParam(cepp, "servinfo::pacetime::simple");
					PACETIME_SIMPLE = atime(cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "intense")) {
					CheckParam(cepp, "servinfo::pacetime::intense");
					PACETIME_INTENSE = atime(cepp->ce_vardata);
				}
				else {
					UnknownVar(cepp, "servinfo::pacetime");
				}
			}
		}
		else if (!mycmp(cep->ce_varname, "masking-keys")) {
			CheckSection(cep, "servinfo::masking-keys");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::masking-keys:");
				if (!mycmp(cepp->ce_varname, "key1")) {
					CheckParam(cepp, "servinfo::masking-keys::key1");
					MASK_KEY1 = atol(cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "key2")) {
					CheckParam(cepp, "servinfo::masking-keys::key2");
					MASK_KEY2 = atol(cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "key3")) {
					CheckParam(cepp, "servinfo::masking-keys::key3");
					MASK_KEY3 = atol(cepp->ce_vardata);
				}
				else {
					UnknownVar(cepp, "servinfo::masking-keys");
				}
			}
		}
		else if (!mycmp(cep->ce_varname, "proxy-monitor")) {
			CheckSection(cep, "servinfo::proxy-monitor");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "servinfo::proxy-monitor");
				if (!mycmp(cepp->ce_varname, "host")) {
					CheckParam(cepp, "servinfo::proxy-monitor::host");
					ConfDupString(PROXYMONITOR_HOST, cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "url")) {
					CheckParam(cepp, "servinfo::proxy-monitor::url");
					tmp = (char *)MyMalloc(strlen(cepp->ce_vardata) + 8);
					ircsprintf(tmp, "http://%s", cepp->ce_vardata);
					ConfDupString(PROXYMONITOR_URL, tmp);
					MyFree(tmp);
				}
				else {
					UnknownVar(cepp, "servinfo::proxy-monitor");
				}
			}
		}
		else if (!mycmp(cep->ce_varname, "network")) {
			CheckSection(cep, "servinfo::network");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cep, "servinfo::network");
				if (!mycmp(cepp->ce_varname, "name")) {
					CheckParam(cepp, "servinfo::network::name");
					ConfDupString(IRCNETWORK, cepp->ce_vardata);
					for (tmp = cepp->ce_vardata; *tmp != '\0'; tmp++) {
						if (*tmp == ' ') {
							*tmp = '-';
						}
					}
					ConfDupString(IRCNETWORK005, cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "kline-address")) {
					CheckParam(cepp, "servinfo::network::kline-address");
					ConfDupString(IRCNETWORK_KLINEADDRESS, cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "aup-url")) {
					CheckParam(cepp, "servinfo::network::aup-url");
					ConfDupString(AUP_URL, cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "servers")) {
					CheckSection(cep, "servinfo::network::servers");
					for (ceppp = cepp->ce_entries; ceppp != NULL; ceppp = ceppp->ce_next) {
						CheckVar(ceppp, "servinfo::network::servers");
						if (!mycmp(ceppp->ce_varname, "services")) {
							CheckParam(ceppp, "servinfo::network::servers::services");
							ConfDupString(SERVICES_SERVER, ceppp->ce_vardata);
						}
						else if (!mycmp(ceppp->ce_varname, "stats")) {
							CheckParam(ceppp, "servinfo::network::servers::stats");
							ConfDupString(STATS_SERVER, ceppp->ce_vardata);
						}
						else {
							UnknownVar(ceppp, "servinfo::network::servers");
						}
					}
				}
				else if (!mycmp(cepp->ce_varname, "masking")) {
					CheckSection(cepp, "servinfo::network::masking");
					for (ceppp = cepp->ce_entries; ceppp != NULL; ceppp = ceppp->ce_next) {
						CheckVar(ceppp, "servinfo::network::masking");
						if (!mycmp(ceppp->ce_varname, "netadmin")) {
							CheckParam(ceppp, "servinfo::network::masking::netadmin");
							ConfDupString(NETADMIN_MASK, ceppp->ce_vardata);
						}
						else if (!mycmp(ceppp->ce_varname, "admin")) {
							CheckParam(ceppp, "servinfo::network::masking::admin");
							ConfDupString(ADMIN_MASK, ceppp->ce_vardata);
						}
						else if (!mycmp(ceppp->ce_varname, "coadmin")) {
							CheckParam(ceppp, "servinfo::network::masking::coadmin");
							ConfDupString(COADMIN_MASK, ceppp->ce_vardata);
						}
						else if (!mycmp(ceppp->ce_varname, "oper")) {
							CheckParam(ceppp, "servinfo::network::masking::oper");
							ConfDupString(OPER_MASK, ceppp->ce_vardata);
						}
						else if (!mycmp(ceppp->ce_varname, "user-prefix")) {
							CheckParam(ceppp, "servinfo::network::masking::user-prefix");
							ConfDupString(USER_PREFIX, ceppp->ce_vardata);
						}
						else {
							UnknownVar(ceppp, "servinfo::network::masking");
						}
					}
				}
				else {
					UnknownVar(cepp, "servinfo::network");
				}
			}
		}
		else {
			UnknownVar(cep, "servinfo");
		}
	}
	return (ConfigItem *)ci_servinfo;
}

int test_servinfo(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_servinfo *ci = (ConfigItem_servinfo *)cip;
	char *p;

	if (BadPtr(ci->name)) {
		ConfDupString(ci->name, "chat.server");
		report(0, "%s:%d: servinfo::name: missing parameter, using %s", ce->ce_fileptr->cf_filename,
			ce->ce_varlinenum, ci->name);
	}
	else {
		for (p = ci->name; *p != '\0'; p++) {
			if (*p == '.') {
				break;
			}
		}
		if (*p == '\0') {
			p = (char *)MyMalloc(strlen(ci->name) + 5);
			ircsprintf(p, "irc.%s", ci->name);
			ConfDupString(ci->name, p);
			MyFree(p);
			report(0, "%s:%d: servinfo::name: invalid parameter (bogus server name), using %s",
				ce->ce_fileptr->cf_filename, ce->ce_varlinenum, p);
		}
	}
	if (BadPtr(ci->desc)) {
		ConfDupString(ci->desc, "RageIRCd v2 Server");
		report(0, "%s:%d: servinfo::desc: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->desc);
	}
	if (!BadPtr(ci->bindip) && !conf_ipaddr(ci->bindip, 0)) {
		ConfFree(ci->bindip);
		report(0, "%s:%d: servinfo::default-bindip: invalid parameter (bogus ip address), "
			"using INADDR_ANY", ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
#ifdef HAVE_THROTTLE_ON
	if (THROTTLE_COUNT < 0) {
		THROTTLE_COUNT = 0;
		report(0, "%s:%d: servinfo::throttle::count: invalid parameter (< 0), throttle{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	else if (THROTTLE_TIME < 0) {
		THROTTLE_TIME = 0;
		report(0, "%s:%d: servinfo::throttle::time: invalid parameter (< 0s), throttle{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
#endif
	if (PACETIME_SIMPLE < 0) {
		PACETIME_SIMPLE = 0;
		report(0, "%s:%d: servinfo::pacetime::simple: invalid parameter (< 0s), pacetime{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	if (PACETIME_INTENSE < 0) {
		PACETIME_INTENSE = 0;
		report(0, "%s:%d: servinfo::pacetime::intense: invalid parameter (< 0s, pacetime{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	if (MASK_KEY1 < 10000) {
		report(1, "%s:%d: ERROR: servinfo::masking-keys::key1: invalid parameter (< 10000)",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return 0;
	}
	if (MASK_KEY2 < 10000) {
		report(1, "%s:%d: ERROR: servinfo::masking-keys::key2: invalid parameter (< 10000)",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return 0;
	}
	if (MASK_KEY3 < 10000) {
		report(1, "%s:%d: ERROR: servinfo::masking-keys::key3: invalid parameter (< 10000)",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return 0;
	}
	if (!BadPtr(PROXYMONITOR_URL) && BadPtr(PROXYMONITOR_HOST)) {
		ConfFree(PROXYMONITOR_URL);
		report(0, "%s:%d: servinfo::proxy-monitor::host: missing parameter, proxy-monitor{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	else if (BadPtr(PROXYMONITOR_URL) && !BadPtr(PROXYMONITOR_HOST)) {
		ConfFree(PROXYMONITOR_HOST);
		report(0, "%s:%d: servinfo::proxy-monitor::url: missing parameter, proxy-monitor{} disabled",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
	}
	if (MAXCHANSPERUSER < 1) {
		MAXCHANSPERUSER = DEFAULT_MAXCHANSPERUSER;
		report(0, "%s:%d: servinfo::max-chans-per-user: invalid parameter (< 1), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, MAXCHANSPERUSER);
	}
	if (MAXCLIENTS < 10) {
		MAXCLIENTS = DEFAULT_MAXCLIENTS;
		report(0, "%s:%d: servinfo::max-clients: invalid parameter (< 10), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, MAXCLIENTS);
	}
	if (LINKDEPTH < 1) {
		LINKDEPTH = DEFAULT_LINKDEPTH;
		report(0, "%s:%d: servinfo::linkdepth: invalid parameter (< 1), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, LINKDEPTH);
	}
	if (CONNAUTH_TIMEOUT < 1) {
		CONNAUTH_TIMEOUT = DEFAULT_CONNAUTH_TIMEOUT;
		report(0, "%s:%d: servinfo::connauth-timeout: invalid parameter (< 1s), using %ds",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, CONNAUTH_TIMEOUT);
	}
	if (BadPtr(KLINEADDRESS)) {
		ConfDupString(KLINEADDRESS, "poorly@configured.server");
		report(0, "%s:%d: servinfo::kline-address: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, KLINEADDRESS);
	}
	else {
		for (p = KLINEADDRESS; *p != '\0'; p++) {
			if (*p == '@' || *p == ':') {
				break;
			}
		}
		if (*p == '\0') {
			ConfDupString(KLINEADDRESS, "poorly@configured.server");
			report(0, "%s:%d: servinfo::kline-address: invalid parameter (bogus e-mail address/url), "
				"using %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, KLINEADDRESS);
		}
	}
	return 1;
}

ConfigItem *free_servinfo(ConfigItem *cip)
{
	ConfigItem_servinfo *ci = (ConfigItem_servinfo *)cip;
	ConfFree(ci->desc);
	ConfFree(ci->bindip);
	return NULL;
}

ConfigItem *parse_admin(ConfigEntry *ce)
{
	ConfigEntry *cep;

	CheckBlockEntries(ce, "admin");
	if (ci_admin == NULL) {
		ci_admin = (ConfigItem_admin *)MyMallocX(sizeof(ConfigItem_admin));
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "admin");
		if (!mycmp(cep->ce_varname, "line1")) {
			CheckParam(cep, "admin::line1");
			ConfDupString(ci_admin->line1, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "line2")) {
			CheckParam(cep, "admin::line2");
			ConfDupString(ci_admin->line2, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "line3")) {
			CheckParam(cep, "admin::line3");
			ConfDupString(ci_admin->line3, cep->ce_vardata);
		}
		else {
			UnknownVar(cep, "admin");
		}
	}
	return (ConfigItem *)ci_admin;
}

ConfigItem *parse_class(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp;
	ConfigItem_class *ci;
	ConfigOption *co;
	int new = 1;

	CheckBlockName(ce, "class", "class name");
	CheckBlockEntries(ce, "class");
	if ((ci = find_class(ce->ce_vardata)) == NULL) {
		ci = (ConfigItem_class *)MyMallocX(sizeof(ConfigItem_class));
		ConfDupString(ci->name, ce->ce_vardata);
	}
	else {
		ci->options = 0;
		new = 0;
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "class");
		if (!mycmp(cep->ce_varname, "pingtime")) {
			CheckParam(cep, "class::pingtime");
			ci->pingtime = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "connectfreq")) {
			CheckParam(cep, "class::connectfreq");
			ci->connectfreq = atime(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "max-clients")) {
			CheckParam(cep, "class::max-clients");
			ci->maxclients = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "sendq")) {
			CheckParam(cep, "class::sendq");
			ci->sendq = conf_size(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "options")) {
			CheckSection(cep, "class::options");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "class::options");
				for (co = class_options; co->name != NULL; co++) {
					if (!mycmp(cepp->ce_varname, co->name)) {
						ci->options |= co->flag;
						break;
					}
				}
				if (co->name == NULL) {
					UnknownVar(cepp, "class::options");
				}
			}
		}
		else {
			UnknownVar(cep, "class");
		}
	}
	if (new == 1) {
		add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_class);
	}
	ci->flag.temporary = 0;
	return (ConfigItem *)ci;
}

int test_class(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_class *ci = (ConfigItem_class *)cip;

	if (BadPtr(ci->name)) {
		report(0, "%s:%d: class: missing name, removing block", ce->ce_fileptr->cf_filename,
			ce->ce_varlinenum);
		return 0;
	}
	if (ci->pingtime < 30) {
		ci->pingtime = DEFAULT_PINGTIME;
		report(0, "%s:%d: class %s::pingtime: invalid parameter (< 30s), using %ds",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name, ci->pingtime);
	}
	if (ci->maxclients < 1) {
		ci->maxclients = DEFAULT_MAXCLIENTS;
		report(0, "%s:%d: class %s::max-clients: invalid parameter (< 1), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name, ci->maxclients);
	}
	if (ci->sendq < 1) {
		ci->sendq = MAXSENDQLENGTH;
		report(0, "%s:%d: class %s::sendq: invalid parameter (< 1), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name, ci->sendq);
	}
	return 1;
}

ConfigItem *free_class(ConfigItem *cip)
{
	ConfigItem_class *ci = (ConfigItem_class *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_class);

	ConfFree(ci->name);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_allow(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp;
	ConfigItem_allow *ci;

	CheckBlockEntries(ce, "allow");
	ci = (ConfigItem_allow *)MyMallocX(sizeof(ConfigItem_allow));
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "allow");
		if (!mycmp(cep->ce_varname, "hostname")) {
			CheckParam(cep, "allow::hostname");
			ConfDupString(ci->hostname, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "ip")) {
			CheckParam(cep, "allow::ip");
			ConfDupString(ci->ip, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "maxperip")) {
			CheckParam(cep, "allow::maxperip");
			ci->maxperip = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "class")) {
			CheckParam(cep, "allow::class");
			if ((ci->class = find_class(cep->ce_vardata)) == NULL) {
				UnknownClass(cep, "allow::class");
				ci->class = default_class;
			}
		}
		else if (!mycmp(cep->ce_varname, "auth")) {
			ci->auth = auth_parse_conf_block(cep, "allow::auth");
		}
		else if (!mycmp(cep->ce_varname, "redirect")) {
			CheckSection(cep, "allow::redirect");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "allow::redirect");
				if (!mycmp(cepp->ce_varname, "server")) {
					CheckParam(cepp, "allow::redirect::server");
					ConfDupString(ci->redirect_server, cepp->ce_vardata);
				}
				else if (!mycmp(cepp->ce_varname, "port")) {
					CheckParam(cepp, "allow::redirect::port");
					ci->redirect_port = atoi(cepp->ce_vardata);
				}
				else {
					UnknownVar(cepp, "allow::redirect");
				}
			}
		}
		else {
			UnknownVar(cep, "allow");
		}
	}
	add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_allow);
	return (ConfigItem *)ci;
}

int test_allow(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_allow *ci = (ConfigItem_allow *)cip;
	int remove_redirect = 0;

	if (BadPtr(ci->hostname)) {
		ConfDupString(ci->hostname, "*");
		report(0, "%s:%d: allow::hostname: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->hostname);
	}
	if (BadPtr(ci->ip)) {
		ConfDupString(ci->ip, "*");
		report(0, "%s:%d: allow::ip: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->ip);
	}
	else if (!conf_ipaddr(ci->ip, 1)) {
		ConfDupString(ci->ip, "*");
		report(0, "%s:%d: allow::ip: invalid parameter (bogus ip address), using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->ip);
	}
	if (ci->class == NULL) {
		ci->class = default_class;
		report(0, "%s:%d: allow::class: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->class->name);
	}
	if (!BadPtr(ci->redirect_server) && (ci->redirect_port < 1 || ci->redirect_port > 65535)) {
		report(0, "%s:%d: allow::redirect::port: invalid parameter (bogus port), removing section",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		remove_redirect = 1;
	}
	else if (BadPtr(ci->redirect_server) && ci->redirect_port > 0 && ci->redirect_port <= 65535) {
		report(0, "%s:%d: allow::redirect::server: missing parameter, removing section",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		remove_redirect = 1;
	}
	if (remove_redirect == 1) {
		ci->redirect_port = 0;
		ConfFree(ci->redirect_server);
	}
	return 1;
}

ConfigItem *free_allow(ConfigItem *cip)
{
	ConfigItem_allow *ci = (ConfigItem_allow *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_allow);

	ConfFree(ci->hostname);
	ConfFree(ci->ip);
	ConfFree(ci->redirect_server);
	auth_destroy_struct(ci->auth);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_oper(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp;
	ConfigItem_oper *ci;
	ConfigOption *co;
	int new = 1;

	CheckBlockName(ce, "oper", "oper name");
	CheckBlockEntries(ce, "oper");
	if ((ci = find_oper(ce->ce_vardata)) == NULL) {
		ci = (ConfigItem_oper *)MyMallocX(sizeof(ConfigItem_oper));
		ConfDupString(ci->name, ce->ce_vardata);
	}
	else {
		ci->options = 0;
		new = 0;
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "oper");
		if (!mycmp(cep->ce_varname, "source")) {
			source_parse_conf_block(cep, "oper::source", (Source **)&ci->source);
		}
		else if (!mycmp(cep->ce_varname, "class")) {
			CheckParam(cep, "oper::class");
			if ((ci->class = find_class(cep->ce_vardata)) == NULL) {
				UnknownClass(cep, "oper::class");
				ci->class = default_class;
			}
		}
		else if (!mycmp(cep->ce_varname, "auth")) {
			ci->auth = auth_parse_conf_block(cep, "oper::auth");
		}
		else if (!mycmp(cep->ce_varname, "options")) {
			CheckSection(cep, "oper::options");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "oper::options");
				for (co = oper_options; co->name != NULL; co++) {
					if (!mycmp(cepp->ce_varname, co->name)) {
						ci->options |= co->flag;
						break;
					}
				}
				if (co->name == NULL) {
					UnknownVar(cepp, "oper::options");
				}
			}
		}
		else if (!mycmp(cep->ce_varname, "autojoin")) {
			CheckParam(cep, "oper::autojoin");
			ConfDupString(ci->autojoinchans, cep->ce_vardata);
		}
		else {
			UnknownVar(cep, "oper");
		}
	}
	if (new == 1) {
		add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_oper);
	}
	return (ConfigItem *)ci;
}

int test_oper(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_oper *ci = (ConfigItem_oper *)cip;

	if (BadPtr(ci->name)) {
		report(0, "%s:%d: oper: missing name, removing block", ce->ce_fileptr->cf_filename,
			ce->ce_varlinenum);
		return 0;
	}
	if (ci->auth == NULL) {
		report(0, "%s:%d: oper %s::auth: missing section, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name);
		return 0;
	}
	if (ci->source == NULL) {
		report(0, "%s:%d: oper %s::source: missing section, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name);
		return 0;
	}
	if (ci->class == NULL) {
		ci->class = default_class;
		report(0, "%s:%d: oper %s::class: missing section, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->name, ci->class->name);
	}
	return 1;
}

ConfigItem *free_oper(ConfigItem *cip)
{
	ConfigItem_oper *ci = (ConfigItem_oper *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_oper);

	ConfFree(ci->name);
	auth_destroy_struct(ci->auth);
	source_destroy_list((Source **)&ci->source);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_uline(ConfigEntry *ce)
{
	ConfigEntry *cep;
	ConfigItem_uline *ci;

	CheckBlockEntries(ce, "uline");
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "uline");
		if ((ci = find_uline(cep->ce_varname)) == NULL) {
			ci = (ConfigItem_uline *)MyMallocX(sizeof(ConfigItem_uline));
			ConfDupString(ci->servername, cep->ce_varname);
			add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_uline);
		}
	}
	return NULL;
}

ConfigItem *parse_link(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp;
	ConfigItem_link *ci;
	ConfigOption *co;
	int new = 1;

	CheckBlockName(ce, "link", "server name");
	CheckBlockEntries(ce, "link");
	if ((ci = find_link(ce->ce_vardata, "*")) == NULL) {
		ci = (ConfigItem_link *)MyMallocX(sizeof(ConfigItem_link));
		ConfDupString(ci->servername, ce->ce_vardata);
	}
	else {
		ci->options = 0;
		new = 0;
	}
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "link");
		if (!mycmp(cep->ce_varname, "fromip")) {
			CheckParam(cep, "link::fromip");
			ConfDupString(ci->fromip, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "bindip")) {
			CheckParam(cep, "link::bindip");
			ConfDupString(ci->bindip, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "port")) {
			CheckParam(cep, "link::port");
			ci->port = atoi(cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "class")) {
			CheckParam(cep, "link::class");
			if ((ci->class = find_class(cep->ce_vardata)) == NULL) {
				UnknownClass(cep, "link::class");
				ci->class = default_class;
			}
		}
		else if (!mycmp(cep->ce_varname, "auth")) {
			ci->auth = auth_parse_conf_block(cep, "link::auth");
		}
		else if (!mycmp(cep->ce_varname, "options")) {
			CheckSection(cep, "link::options");
			for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
				CheckVar(cepp, "link::options");
				for (co = link_options; co->name != NULL; co++) {
					if (!mycmp(cepp->ce_varname, co->name)) {
						ci->options |= co->flag;
						break;
					}
				}
				if (co->name == NULL) {
					UnknownVar(cepp, "link::options");
				}
			}
		}
		else {
			UnknownVar(cep, "link");
		}
	}
	if (new == 1) {
		add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_link);
	}
	ci->flag.temporary = 0;
	return (ConfigItem *)ci;
}

int test_link(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_link *ci = (ConfigItem_link *)cip;

	if (BadPtr(ci->servername)) {
		report(0, "%s:%d: link: missing servername, removing block", ce->ce_fileptr->cf_filename,
			ce->ce_varlinenum);
		return 0;
	}
	if (ci->auth == NULL) {
		report(0, "%s:%d: link %s::auth: missing section, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->servername);
		return 0;
	}
	if (BadPtr(ci->fromip)) {
		ConfDupString(ci->fromip, "*");
		report(0, "%s:%d: link %s::fromip: missing parameter, using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->servername, ci->fromip);
	}
	else if (!conf_ipaddr(ci->fromip, 0)) {
		ConfDupString(ci->fromip, "*");
		report(0, "%s:%d: link %s::fromip: invalid parameter (bogus ip address), using %s",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->servername, ci->fromip);
	}
	if (!BadPtr(ci->bindip) && !conf_ipaddr(ci->bindip, 0)) {
		ConfFree(ci->bindip);
		report(0, "%s:%d: link %s::bindip: invalid parameter (bogus ip address), removing variable",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->servername);
	}
	if (ci->port < 1 || ci->port > 65535) {
		ci->port = DEFAULT_PORTNUM;
		report(0, "%s:%d: link %s::port: invalid parameter (bogus port), using %d",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->servername, ci->port);
	}
	if (ci->class == NULL) {
		ci->class = default_class;
		report(0, "%s:%d: link %s::class: missing parameter, using %s", ce->ce_fileptr->cf_filename,
			ce->ce_varlinenum, ci->servername, ci->class->name);
	}
	return 1;
}

ConfigItem *free_link(ConfigItem *cip)
{
	ConfigItem_link *ci = (ConfigItem_link *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_link);

	ConfFree(ci->servername);
	ConfFree(ci->fromip);
	ConfFree(ci->bindip);
	auth_destroy_struct(ci->auth);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_bind(ConfigEntry *ce)
{
	ConfigEntry *cep, *cepp;
	ConfigItem_bind *ci;
	ConfigOption *co;
	int port, new;
	char ipaddr[HOSTIPLEN + 1];

	CheckBlockName(ce, "bind", "ip address");
	if (!conf_ipaddr(ce->ce_vardata, 0)) {
		report(0, "%s:%d: bind: invalid parameter (bogus ip address), ignoring block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return NULL;
	}
	strncpyzt(ipaddr, ce->ce_vardata, HOSTIPLEN + 1);
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		port = atoi(cep->ce_varname);
		if (port < 1 || port > 65535) {
			report(0, "%s:%d: bind %s: invalid parameter (bogus port), ignoring variable",
				cep->ce_fileptr->cf_filename, cep->ce_varlinenum, ipaddr);
			continue;
		}
		if ((ci = find_bind(ipaddr, port)) == NULL) {
			ci = (ConfigItem_bind *)MyMallocX(sizeof(ConfigItem_bind));
			ConfDupString(ci->ip, ipaddr);
			ci->port = port;
			new = 1;
		}
		else {
			ci->options = 0;
			new = 0;
		}
		for (cepp = cep->ce_entries; cepp != NULL; cepp = cepp->ce_next) {
			for (co = bind_options; co->name != NULL; co++) {
				if (!mycmp(cepp->ce_varname, co->name)) {
					ci->options |= co->flag;
					break;
				}
			}
			if (co->name == NULL) {
				UnknownVar(cepp, "bind::port");
			}
		}
		if (new == 1) {
			add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_bind);
		}
		ci->flag.temporary = 0;
	}
	return NULL;
}

ConfigItem *free_bind(ConfigItem *cip)
{
	ConfigItem_bind *ci = (ConfigItem_bind *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_bind);

	ConfFree(ci->ip);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_ban(ConfigEntry *ce)
{
	ConfigEntry *cep;
	ConfigItem_ban *ci;
	ConfigOption *co;

	CheckBlockName(ce, "ban", "ban type");
	CheckBlockEntries(ce, "ban");
	for (co = ban_types; co->name != NULL; co++) {
		if (!mycmp(co->name, ce->ce_vardata)) {
			break;
		}
	}
	if (co->name == NULL) {
		report(0, "%s:%d: ban: unknown variable %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
			ce->ce_vardata);
		return NULL;
	}
	ci = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ci->type = co->flag;
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "ban");
		if (!mycmp(cep->ce_varname, "mask")) {
			CheckParam(cep, "ban::mask");
			ConfDupString(ci->mask, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "reason")) {
			CheckParam(cep, "ban::reason");
			ConfDupString(ci->reason, cep->ce_vardata);
		}
		else {
			UnknownVar(cep, "ban");
		}
	}
	add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_ban);
	ci->flag.removable = 0;
	ci->flag.remove_on_rehash = 1;
	return (ConfigItem *)ci;
}

int test_ban(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_ban *ci = (ConfigItem_ban *)cip;

	if (BadPtr(ci->mask)) {
		report(0, "%s:%d: ban %s::mask: missing parameter, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, conf_bantype(ci->type));
		return 0;
	}
	if (BadPtr(ci->reason)) {
		report(0, "%s:%d: ban %s::reason: missing parameter, using default",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, conf_bantype(ci->type));
		ConfDupString(ci->reason, "no reason");
	}
	return 1;
}

ConfigItem *free_ban(ConfigItem *cip)
{
	ConfigItem_ban *ci = (ConfigItem_ban *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_ban);

	ConfFree(ci->mask);
	ConfFree(ci->reason);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_except(ConfigEntry *ce)
{
	ConfigEntry *cep;
	ConfigItem_except *ci;

	CheckBlockEntries(ce, "except");
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "except");
		if ((ci = find_except(cep->ce_varname)) == NULL) {
			ci = (ConfigItem_except *)MyMallocX(sizeof(ConfigItem_except));
			ConfDupString(ci->mask, cep->ce_varname);
			add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_except);
		}
	}
	return NULL;
}

ConfigItem *parse_vhost(ConfigEntry *ce)
{
	ConfigEntry *cep;
	ConfigItem_vhost *ci;

	CheckBlockName(ce, "vhost", "vhost mask");
	CheckBlockEntries(ce, "vhost");
	ci = (ConfigItem_vhost *)MyMallocX(sizeof(ConfigItem_vhost));
	ConfDupString(ci->mask, ce->ce_vardata);
	for (cep = ce->ce_entries; cep != NULL; cep = cep->ce_next) {
		CheckVar(cep, "vhost");
		if (!mycmp(cep->ce_varname, "login")) {
			CheckParam(cep, "vhost::login");
			ConfDupString(ci->login, cep->ce_vardata);
		}
		else if (!mycmp(cep->ce_varname, "auth")) {
			ci->auth = auth_parse_conf_block(cep, "vhost::auth");
		}
		else if (!mycmp(cep->ce_varname, "source")) {
			source_parse_conf_block(cep, "vhost::source", (Source **)&ci->source);
		}
		else {
			UnknownVar(cep, "vhost");
		}
	}
	add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_vhost);
	return (ConfigItem *)ci;
}

int test_vhost(ConfigEntry *ce, ConfigItem *cip)
{
	ConfigItem_vhost *ci = (ConfigItem_vhost *)cip;

	if (BadPtr(ci->mask)) {
		report(0, "%s:%d: vhost: missing vhost mask, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum);
		return 0;
	}
	if (BadPtr(ci->login)) {
		report(0, "%s:%d: vhost %s::login: missing parameter, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->mask);
		return 0;
	}
	if (ci->auth == NULL) {
		report(0, "%s:%d: vhost %s::auth: missing section, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->mask);
		return 0;
	}
	if (ci->source == NULL) {
		report(0, "%s:%d: vhost %s::source: missing section, removing block",
			ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ci->mask);
		return 0;
	}
	return 1;
}

ConfigItem *free_vhost(ConfigItem *cip)
{
	ConfigItem_vhost *ci = (ConfigItem_vhost *)cip;
	ConfigItem *rci = del_ConfigItem(cip, (ConfigItem **)&ci_vhost);

	ConfFree(ci->mask);
	ConfFree(ci->login);
	auth_destroy_struct(ci->auth);
	source_destroy_list((Source **)&ci->source);
	ConfFree(ci);
	return rci;
}

ConfigItem *parse_include(ConfigEntry *ce)
{
	ConfigItem_include *ci_ptr, *ci;

	CheckBlockName(ce, "include", "file name");
	for (ci_ptr = ci_include; ci_ptr != NULL; ci_ptr = (ConfigItem_include *)ci_ptr->next) {
		if (!mycmp(ci_ptr->filename, ce->ce_vardata)) {
			report(0, "%s:%d: include: config file %s already %s", ce->ce_fileptr->cf_filename,
				ce->ce_varlinenum, ce->ce_vardata,
				ci_ptr->loaded == 0 ? "makred for inclusion" : "loaded");
			return NULL;
		}
	}
	ci = (ConfigItem_include *)MyMallocX(sizeof(ConfigItem_include));
	ConfDupString(ci->filename, ce->ce_vardata);
	add_ConfigItem((ConfigItem *)ci, (ConfigItem **)&ci_include);
	return (ConfigItem *)ci;
}

ConfigItem_class *find_class(char *name)
{
	ConfigItem_class *ci;
	for (ci = ci_class; ci != NULL; ci = (ConfigItem_class *)ci->next) {
		if (!mycmp(ci->name, name)) {
			return ci;
		}
	}
	return NULL;
}

ConfigItem_oper *find_oper(char *name)
{
	ConfigItem_oper *ci;
	for (ci = ci_oper; ci != NULL; ci = (ConfigItem_oper *)ci->next) {
		if (!mycmp(ci->name, name)) {
			return ci;
		}
	}
	return NULL;
}

ConfigItem_uline *find_uline(char *servername)
{
	ConfigItem_uline *ci;
	for (ci = ci_uline; ci != NULL; ci = (ConfigItem_uline *)ci->next) {
		if (!match(ci->servername, servername)) {
			return ci;
		}
	}
	return NULL;
}

ConfigItem_link *find_link(char *servername, char *fromip)
{
	ConfigItem_link *ci;
	for (ci = ci_link; ci != NULL; ci = (ConfigItem_link *)ci->next) {
		if (!match(ci->servername, servername) || !match(servername, ci->servername)) {
			if (!match(ci->fromip, fromip) || !match(fromip, ci->fromip)) {
				return ci;
			}
		}
	}
	return NULL;
}

ConfigItem_bind *find_bind(char *ip, int port)
{
	ConfigItem_bind *ci;
	for (ci = ci_bind; ci != NULL; ci = (ConfigItem_bind *)ci->next) {
		if (ci->port == port && !match(ci->ip, ip)) {
			return ci;
		}
	}
	return NULL;
}

ConfigItem_ban *find_ban(long type, char *mask)
{
	ConfigItem_ban *ci;
	for (ci = ci_ban; ci != NULL; ci = (ConfigItem_ban *)ci->next) {
		if (ci->type & type && !match(ci->mask, mask)) {
			if (find_except(mask) == NULL) {
				return ci;
			}
		}
	}
	return NULL;
}

ConfigItem_except *find_except(char *mask)
{
	ConfigItem_except *ci;
	for (ci = ci_except; ci != NULL; ci = (ConfigItem_except *)ci->next) {
		if (!match(ci->mask, mask)) {
			return ci;
		}
	}
	return NULL;
}
