/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/send.c: sending routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: send.c,v 1.45 2002/12/15 21:26:21 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "ssl.h"
#include "zlink.h"
#include <stdio.h>
#include <time.h>
#include "memory.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendbuf[2048];
static char remotebuf[2048];
static int send_message(aClient *, char *, int);

#ifdef HAVE_OPENSSL_ON
static char rc4buf[16384];
#endif

static int sentalong[MAXCONNECTIONS];
static int sent_serial;

void init_send()
{
	memset(sentalong, '\0', sizeof(int) * MAXCONNECTIONS);
	sent_serial = 0;
	return;
}

#define HIGHEST_SERIAL INT_MAX
#define INC_SERIAL	if (sent_serial == HIGHEST_SERIAL) { \
				memset(sentalong, '\0', sizeof(sentalong)); \
				sent_serial = 0; \
			} \
			sent_serial++;

static int dead_link(aClient *to, char *notice, int sockerr)
{
	int errtmp = errno;

	to->sockerr = sockerr;
	to->flags = FLAGS_DEADSOCKET;

	DBufClear(&to->recvQ);
	DBufClear(&to->sendQ);

	if (IsServer(to) && !(to->flags & FLAGS_CLOSING)) {
		char fbuf[512], *path = get_client_name(to, HIDEME), *errtxt = strerror(errtmp);

		ircsprintf(fbuf, "%s", notice);
		sendto_gnotice(fbuf, path, errtxt);
		ircsprintf(fbuf, ":%s GNOTICE :%s", me.name, notice);
		sendto_serv_butone(to, fbuf, path, errtxt);
	}
	return -1;
}

static int send_message(aClient *to, char *msg, int len)
{
	static int SQinK;
	int server;

	if (to->from != NULL) {
		to = to->from;
	}
	if (IsServer(to) || (to->serv != NULL && (to->serv->sflags & SFLAGS_RC4_NEG))) {
		server = 1;
		if (len > 510) {
			msg[511] = '\n';
			msg[512] = '\0';
			len = 512;
		}
		else {
			msg[len] = '\n';
			msg[++len] = '\0';
		}
	}
	else {
		server = 0;
		if (len > 509) {
			msg[510] = '\r';
			msg[511] = '\n';
			msg[512] = '\0';
			len = 512;
		}
		else {
			msg[len] = '\r';
			msg[++len] = '\n';
			msg[++len] = '\0';
		}
	}
	if (IsMe(to)) {
		sendto_ops("Trying to send to myself! [%s]", msg);
		return 0;
	}
	if (IsDead(to)) {
		return 0;
	}
	if (DBufLength(&to->sendQ) > to->sendqlen) {
		if (IsServer(to)) {
			sendto_ops("Max SendQ limit exceeded for %s: %d > %d", get_client_name(to, HIDEME),
				DBufLength(&to->sendQ), to->class ? to->class->sendq : MAXSENDQLENGTH);
		}
		to->flags |= FLAGS_SENDQEX;
		return dead_link(to, "Max SendQ exceeded for %s, closing link", 0);
	}

	to->sendM++;
	me.sendM++;
	if (to->listener != NULL) {
		to->listener->sendM++;
	}

	if (server == 1) {
		if (to->serv->sflags & SFLAGS_ZIP_OUT) {
			int ldata = to->serv->sflags & SFLAGS_BURST;
			msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);
			if (len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
				return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
			}
			if (len == 0) {
				return 0;
			}
		}
#ifdef HAVE_OPENSSL_ON
		if (to->serv->sflags & SFLAGS_RC4_OUT) {
			rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
			msg = rc4buf;
		}
#endif
	}
	if (dbuf_put(&to->sendQ, msg, len, server) < 0) {
		return dead_link(to, "Buffer allocation error for %s, closing link", IRCERR_BUFALLOC);
	}
#ifdef ALWAYS_SEND_DURING_SPLIT
	if (currently_processing_netsplit && !(to->flags & FLAGS_BLOCKED)) {
		send_queued(to);
		return 0;
	}
#endif

	SQinK = (DBufLength(&to->sendQ) >> 10);
	if (IsServer(to)) {
		if (SQinK > to->lastsq) {
			send_queued(to);
		}
	}
	else {
		if (SQinK > to->lastsq + 4) {
			send_queued(to);
		}
	}
	return 0;
}

int send_queued(aClient *to)
{
	char *msg;
	int len, rlen, more_data = 0;
	int server = IsServer(to) ? 1 : 0;

	if (IsDead(to)) {
		return -1;
	}
	if (server == 1) {
		if ((to->serv->sflags & SFLAGS_ZIP_OUT) && zip_is_data_out(to->serv->zip_out)) {
			if (DBufLength(&to->sendQ)) {
				more_data = 1;
			}
			else {
				int ldata = to->serv->sflags & SFLAGS_BURST;
				msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
				if (len == -1) {
					sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
					return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
				}
#ifdef HAVE_OPENSSL_ON
				if (to->serv->sflags & SFLAGS_RC4_OUT) {
					rc4_process_stream(to->serv->rc4_out, msg, len);
				}
#endif
				if (!dbuf_put(&to->sendQ, msg, len, server)) {
					return dead_link(to, "Buffer allocation error for %s", IRCERR_BUFALLOC);
				}
			}
		}
	}
	while (DBufLength(&to->sendQ) > 0) {
		msg = dbuf_map(&to->sendQ, (size_t *)&len);

		if ((rlen = deliver_it(to, msg, len)) < 0) {
			return dead_link(to, "Write error to %s, closing link (%s)", errno);
		}
		dbuf_delete(&to->sendQ, rlen);
		to->lastsq = (DBufLength(&to->sendQ) >> 10);
		if (rlen < len) {
			break;
		}

		if (more_data && !DBufLength(&to->sendQ)) {
			int ldata = to->serv->sflags & SFLAGS_BURST;

			more_data = 0;
			msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
			if (len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
				return dead_link(to, "Zip outpu terror for %s", IRCERR_ZIP);
			}
#ifdef HAVE_OPENSSL_ON
			if (to->serv->sflags & SFLAGS_RC4_OUT) {
				rc4_process_stream(to->serv->rc4_out, msg, len);
			}
#endif
			if (!dbuf_put(&to->sendQ, msg, len, server)) {
				return dead_link(to, "Buffer allocation error for %s", IRCERR_BUFALLOC);
			}
		}
	}
	if (server == 1 && (to->serv->sflags & SFLAGS_SOBSENT) && HasCapab(to, CAPAB_BURST) &&
	  DBufLength(&to->sendQ) < 20480) {
		if (!(to->serv->sflags & SFLAGS_BURST)) {
			to->serv->sflags &= ~SFLAGS_SOBSENT;
			sendto_one(to, "BURST %d", DBufLength(&to->sendQ));

#ifdef HTM_LOCK_ON_NETBURST
			if (!(to->serv->sflags & SFLAGS_EOBRECV)) {
				HTMLOCK = 0;
			}
#endif
		}
	}
	return IsDead(to) ? -1 : 0;
}

void sendto_one(aClient *to, char *pattern, ...)
{
	va_list vl;
	int len;

	va_start(vl, pattern);
	len = ircvsprintf(sendbuf, pattern, vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to)) {
		sendto_ops("Trying to send to myself! [%s]", sendbuf);
		return;
	}
	send_message(to, sendbuf, len);
	va_end(vl);
	return;
}

void vsendto_one(aClient *to, char *pattern, va_list vl)
{
	int len = ircvsprintf(sendbuf, pattern, vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to) && to->fd >= 0) {
		sendto_ops("Trying to send to myself! [%s]", sendbuf);
		return;
	}
	send_message(to, sendbuf, len);
	return;
}

static inline int prefix_buffer(int remote, aClient *from, char *prefix, char *buffer, char *pattern, va_list vl)
{
	char *p;
	int msglen, sidx = 1;

	*buffer = ':';
	if (remote == 0 && IsPerson(from)) {
		int flag = 0;
		for (p = from->name; *p; p++) {
			buffer[sidx++] = *p;
		}
		if (*from->username != '\0') {
			buffer[sidx++] = '!';
			for (p = from->username; *p; p++) {
				buffer[sidx++] = *p;
			}
		}
		if (*from->host != '\0' && !MyConnect(from)) {
			buffer[sidx++] = '@';
			for (p = MaskedHost(from); *p; p++) {
				buffer[sidx++] = *p;
			}
			flag = 1;
		}
		if (flag == 0 && MyConnect(from) && from->host != '\0') {
			buffer[sidx++] = '@';
			for (p = MaskedHost(from); *p; p++) {
				buffer[sidx++] = *p;
			}
		}
	}
	else {
		for (p = prefix; *p; p++) {
			buffer[sidx++] = *p;
		}
	}
	msglen = ircvsprintf(&buffer[sidx], pattern + 3, vl);
	msglen += sidx;
	return msglen;
}

static inline int check_fake_direction(aClient *from, aClient *to)
{
	if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
		if (IsServer(from)) {
			sendto_ops("Message to %s[%s] dropped from %s (Fake direction)",
				to->name, to->from->name, from->name);
			return -1;
		}
		sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
			MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name,
			to->from->name);
		sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
			me.name, to->name, to->username, MaskedHost(to), to->from->name);
		to->flags |= FLAGS_KILLED;
		exit_client(NULL, to, &me, "Ghosted client");

		if (IsPerson(from)) {
			sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
				to->username, MaskedHost(to), to->from);
		}
		return -1;
	}
	return 0;
}

void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	int didlocal = 0, didremote = 0;
	va_list vl;
	char *pfix;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	INC_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		acptr = cm->cptr;
		if (acptr->from == one || HasMode(acptr, UMODE_DEAF)) {
			continue;
		}
		if (MyClient(acptr)) {
			if (didlocal == 0) {
				didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			send_message(acptr, sendbuf, didlocal);
			sentalong[acptr->from->fd] = sent_serial;
		}
		else {
			if (didremote == 0) {
				didremote = prefix_buffer(1, from, pfix, remotebuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			if (sentalong[acptr->from->fd] != sent_serial) {
				send_message(acptr, remotebuf, didremote);
				sentalong[acptr->from->fd] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}

void sendto_serv_butone(aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);

	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from)) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_snick_servs_butone(aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from) || !HasCapab(cptr, CAPAB_SNICK)) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_nickip_servs_butone(int nickip, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from) || (nickip && !HasCapab(cptr, CAPAB_NICKIP))
		  || (!nickip && HasCapab(cptr, CAPAB_NICKIP))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_ssjoin_servs_butone(int ssjv2, aChannel *chptr, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	if (chptr && *chptr->chname != '#') {
		return;
	}

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from) || (ssjv2 == 1 && !HasCapab(cptr, CAPAB_SSJv2)) ||
		  (ssjv2 == 0 && !HasCapab(cptr, CAPAB_SSJOIN))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_noquit_servs_butone(int noquit, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || one == cptr || (noquit && !HasCapab(cptr, CAPAB_NOQUIT)) ||
		  (!noquit && HasCapab(cptr, CAPAB_NOQUIT))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_common_channels(aClient *from, char *pattern, ...)
{
	chanMember *chans, *users;
	aClient *cptr;
	va_list vl;
	char *pfix;
	int msglen = 0;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	INC_SERIAL

	if (from->fd >= 0) {
		sentalong[from->fd] = sent_serial;
	}
	if (from->user !=  NULL) {
		for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
			for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}

				sentalong[cptr->fd] = sent_serial;
				if (msglen == 0) {
					msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
				}
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	if (MyConnect(from)) {
		if (msglen == 0) {
			msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
		}
		send_message(from, sendbuf, msglen);
	}
	va_end(vl);
	return;
}

void send_quit_to_common_channels(aClient *from, char *reason)
{
	chanMember *chans, *users;
	aClient *cptr;
	int msglen;

	INC_SERIAL

	msglen = ircsprintf(sendbuf, ":%s!%s@%s QUIT :%s", from->name, from->username, MaskedHost(from), reason);
	if (from->fd >= 0) {
		sentalong[from->fd] = sent_serial;
	}
	for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
		if (!can_send(from, chans->chptr, reason)) {
			for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}

				sentalong[cptr->fd] = sent_serial;
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	return;
}

void send_part_to_common_channels(aClient *from, char *reason)
{
	chanMember *chans, *users;
	aClient *cptr;
	int msglen = 0;

	for (chans = from->user->channel; chans != NULL; chans = chans->nextchan) {
		if (can_send(from, chans->chptr, reason)) {
			msglen = ircsprintf(sendbuf, ":%s!%s@%s PART %s", from->name, from->username,
				MaskedHost(from), chans->chptr->chname);
			INC_SERIAL
			if (from->fd >= 0) {
				sentalong[from->fd] = sent_serial;
			}
			for (users = chans->chptr->members; users != NULL; users = users->nextuser) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}
				sentalong[cptr->fd] = sent_serial;

				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	return;
}

#ifdef FLUD
void sendto_channel_butlocal(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		acptr = cm->cptr;
		if (acptr->from == one || HasMode(acptr, UMODE_DEAF)) {
			continue;
		}
		if (!MyFludConnect(acptr)) {
			if (sentalong[acptr->from->fd] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[acptr->from->fd] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}
#endif

void sendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	va_list vl;
	int didlocal = 0;
	char *pfix;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		acptr = cm->cptr;
		if (HasMode(acptr, UMODE_DEAF)) {
			continue;
		}
		if (MyConnect(acptr)) {
			if (didlocal == 0) {
				didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			send_message(acptr, sendbuf, didlocal);
		}
	}
	va_end(vl);
	return;
}

static int match_it(aClient *one, char *mask, int what)
{
	if (what == MATCH_HOST) {
		return (match(mask, one->host) == 0);
	}
	else {
		return (match(mask, one->user->server) == 0);
	}
}

void sendto_match_servs(aChannel *chptr, aClient *from, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	if (chptr != NULL && *chptr->chname == '&') {
		return;
	}

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || cptr == from) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_match_butone(aClient *one, aClient *from, char *mask, int what, char *pattern, ...)
{
	int i;
	aClient *cptr, *acptr;
	int cansendlocal, cansendglobal;
	va_list vl;

	va_start(vl, pattern);
	if (MyConnect(from)) {
		cansendlocal = OPHasFlag(from, OFLAG_LNOTICE) ? 1 : 0;
		cansendglobal = OPHasFlag(from, OFLAG_GNOTICE) ? 1 : 0;
	}
	else {
		cansendlocal = cansendglobal = 1;
	}
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || cptr == one) {
			continue;
		}
		if (IsServer(cptr)) {
			if (cansendglobal == 0) {
				continue;
			}
			for (acptr = client; acptr != NULL; acptr = acptr->next) {
				if (IsRegisteredUser(acptr) && match_it(acptr, mask, what) && acptr->from == cptr) {
					break;
				}
			}
			if (acptr == NULL) {
				continue;
			}
		}
		else if (!cansendlocal || !(IsRegisteredUser(cptr) && match_it(cptr, mask, what))) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_all_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsMe(cptr) || one == cptr) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_ops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || IsMe(cptr)) {
			continue;
		}
		switch (lev) {
			case CCONN_LEV:
				if (!HasMode(cptr, UMODE_LCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case GCCONN_LEV:
				if (!HasMode(cptr, UMODE_GCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case REJ_LEV:
				if (!HasMode(cptr, UMODE_LCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SKILL_LEV:
				if (!HasMode(cptr, UMODE_KNOTICE)) {
					continue;
				}
				break;
			case SPY_LEV:
				if (!HasMode(cptr, UMODE_SPYING) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DCCSEND_LEV:
				if (!HasMode(cptr, UMODE_DCCNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case FLOOD_LEV:
				if (!HasMode(cptr, UMODE_FNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SPAM_LEV:
				if (!HasMode(cptr, UMODE_SPAMBOT) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DEBUG_LEV:
				if (!HasMode(cptr, UMODE_DNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			default:
				if (!HasMode(cptr, UMODE_SNOTICE)) {
					continue;
				}
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_ops(char *pattern, ...)
{
	va_list vl;

	va_start(vl, pattern);
	vsendto_realops(pattern, vl);
	va_end(vl);
	return;
}

void sendto_ops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cptr = client; cptr; cptr = cptr->next) {
		i = cptr->from->fd;
		if (!OPHasFlag(cptr, OFLAG_WALLOPS) || sentalong[i] == sent_serial || (one && cptr->from == one)) {
			continue;
		}
		sentalong[i] = sent_serial;
		vsendto_prefix_one(cptr->from, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_wallops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || !(IsRegistered(cptr) && (OPHasFlag(cptr, OFLAG_WALLOPS)
		  || IsServer(cptr))) || cptr == one) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void send_globops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	va_list vl;
	fdlist *l;

	va_start(vl, pattern);
	l = &oper_fdlist;
	for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
		if (!(cptr = local[fd]) || !HasMode(cptr, UMODE_OPER) || !OPHasFlag(cptr, OFLAG_GLOBOPS)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** Global -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void send_chatops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	va_list vl;
	fdlist *l;

	va_start(vl, pattern);
	l = &oper_fdlist;
	for (fd = l->entry[i = 0]; i <= l->last_entry; fd = l->entry[++i]) {
		if (!(cptr = local[fd]) || !HasMode(cptr, UMODE_OPER) || !HasMode(cptr, UMODE_CHATOPS)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** ChatOps -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	char *idx, *par;
	int flag = 0, sidx = 0;
	va_list vl, vl2;

	va_start(vl, pattern);
	vl2 = vl;
	par = va_arg(vl, char *);

	if (from != NULL) {
		if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl2);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				va_end(vl);
				return;
			}
			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
				MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name,
				to->name, me.name, to->name, to->username, MaskedHost(to), to->from->name);
			to->flags |= FLAGS_KILLED;
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
					to->username, MaskedHost(to), to->from);
			}
			va_end(vl);
			return;
		}
		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			for (idx = from->name; *idx; idx++) {
				sender[sidx++] = *idx;
			}
			if (*from->username != '\0') {
				sender[sidx++] = '!';
				for (idx = from->username; *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}
			if (*from->host != '\0' && !MyConnect(from)) {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx; idx++) {
					sender[sidx++] = *idx;
				}
				flag = 1;
			}
			if (flag == 0 && MyConnect(from) && *from->host != '\0') {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}
			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;
	for (idx = par; *idx; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx; idx++) {
		temp[sidx++] = *idx;
	}
	temp[sidx] = '\0';
	vsendto_one(to, temp, vl);
	va_end(vl);
	return;
}

void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	int flag = 0, sidx = 0;
	va_list vl2 = vl;
	char *idx, *par = va_arg(vl2, char *);

	if (from != NULL) {
		if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				return;
			}
			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->username,
				MaskedHost(to), from->name, from->username, MaskedHost(from), to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
				to->username, MaskedHost(to), to->from->name);
			to->flags |= FLAGS_KILLED;
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
					to->username, MaskedHost(to), to->from);
			}
			return;
		}
		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			for (idx = from->name; *idx; idx++) {
				sender[sidx++] = *idx;
			}
			if (*from->username != '\0') {
				sender[sidx++] = '!';
				for (idx = from->username; *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}
			if (*from->host != '\0' && !MyConnect(from)) {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx; idx++) {
					sender[sidx++] = *idx;
				}
				flag = 1;
			}
			if (flag == 0 && MyConnect(from) && *from->host != '\0') {
				sender[sidx++] = '@';
				for (idx = MaskedHost(from); *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}
			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;
	for (idx = par; *idx; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx; idx++) {
		temp[sidx++] = *idx;
	}
	temp[sidx] = '\0';
	vsendto_one(to, temp, vl2);
	return;
}

void sendto_fdlist(fdlist *listp, char *pattern, ...)
{
	int len, j, fd;
	va_list vl;

	va_start(vl, pattern);
	len = ircvsprintf(sendbuf, pattern, vl);
	for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j]) {
		send_message(local[fd], sendbuf, len);
	}
	va_end(vl);
	return;
}

void vsendto_fdlist(fdlist *listp, char *pattern, va_list vl)
{
	int len, j, fd;
	len = ircvsprintf(sendbuf, pattern, vl);

	for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j]) {
		send_message(local[fd], sendbuf, len);
	}
	return;
}

void sendto_realops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	va_start(vl, pattern);
	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !HasMode(cptr, UMODE_OPER)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void vsendto_realops(char *pattern, va_list vl)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	fdlist *listp;

	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !HasMode(cptr, UMODE_OPER)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	return;
}

void sendto_realops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int i, fd, len = 0;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	listp = &oper_fdlist;
	va_start(vl, pattern);

	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd])) {
			continue;
		}
		switch (lev) {
			case CCONN_LEV:
				if (!HasMode(cptr, UMODE_LCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case GCCONN_LEV:
				if (!HasMode(cptr, UMODE_GCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case REJ_LEV:
				if (!HasMode(cptr, UMODE_LCLICONN) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SKILL_LEV:
				if (!HasMode(cptr, UMODE_KNOTICE)) {
					continue;
				}
				break;
			case SPY_LEV:
				if (!HasMode(cptr, UMODE_SPYING) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DCCSEND_LEV:
				if (!HasMode(cptr, UMODE_DCCNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case FLOOD_LEV:
				if (!HasMode(cptr, UMODE_FNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case SPAM_LEV:
				if (!HasMode(cptr, UMODE_SPAMBOT) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
			case DEBUG_LEV:
				if (!HasMode(cptr, UMODE_DNOTICE) || !HasMode(cptr, UMODE_OPER)) {
					continue;
				}
				break;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void ts_warn(char *pattern, ...)
{
	static ts_val last = 0;
	static int warnings = 0;
	ts_val now;
	va_list vl;

	va_start(vl, pattern);

	now = time(NULL);
	if (now - last < 5) {
		if (++warnings > 5) {
			return;
		}
	}
	else {
		last = now;
		warnings = 0;
	}
	vsendto_realops(pattern, vl);
	va_end(vl);
	return;
}

void sendto_locops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd, len = 0;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	va_start(vl, pattern);
	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !OPHasFlag(cptr, OFLAG_LOCOPS)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_gnotice(char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || IsMe(cptr) || !HasMode(cptr, UMODE_RSTAFF)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Routing -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_channelflag_butone(int flag, aClient *one, aClient *from, aChannel *chptr,
  char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		acptr = cm->cptr;
		if (acptr->from == one || !(cm->flags & flag) || HasMode(acptr, UMODE_DEAF)) {
			continue;
		}
		if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[acptr->from->fd] = sent_serial;
		}
		else {
			if (sentalong[acptr->from->fd] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[acptr->from->fd] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}

void flush_connections(int fd)
{
	int i;
	aClient *cptr;

	if (fd == me.fd) {
		for (i = highest_fd; i >= 0; i--) {
			if (!(cptr = local[i]) || cptr->flags & FLAGS_BLOCKED || DBufLength(&cptr->sendQ) <= 0) {
				continue;
			}
			send_queued(cptr);
		}
	}
	else if (fd >= 0 && (cptr = local[fd]) && !(cptr->flags & FLAGS_BLOCKED) && DBufLength(&cptr->sendQ) > 0) {
		send_queued(cptr);
	}
	return;
}

void dump_connections(int fd)
{
	int i;
	aClient *cptr;

	if (fd == me.fd) {
		for (i = highest_fd; i >= 0; i--) {
			if (!(cptr = local[i]) || DBufLength(&cptr->sendQ) <= 0) {
				continue;
			}
			send_queued(cptr);
		}
	}
	else if (fd >= 0 && (cptr = local[fd]) && DBufLength(&cptr->sendQ) > 0) {
		send_queued(cptr);
	}
	return;
}
