/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ssl.c: client-to-server SSL functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ssl.c,v 1.11 2002/10/28 21:36:14 rageircd Exp $
 */

#include "setup.h"

#ifdef HAVE_OPENSSL_ON

#include "config.h"
#include "common.h"
#include "struct.h"
#include "ssl.h"
#include "h.h"
#include "memory.h"
#include <string.h>
#include <unistd.h>

SSL_CTX *ircd_ctx;

extern void set_sock_opts(int, aClient *, Listener *);

static int ssl_pem_passwd_cb(char *, int, int, void *);
static int init_ircd_ctx();
static char *ssl_get_init_error_message(int);
static void ssl_handle_fatal_error(int, int, aClient *);

static int ssl_pem_passwd_cb(char *buf, int bufsize, int rwflag, void *password)
{
	static int been_here = 0;
	static char bbuf[1024];
	char *pass;

	if (been_here == 0) {
		if ((pass = getpass(" --- Enter password for SSL private key file: ")) != NULL) {
			strncpyzt(buf, pass, bufsize);
			strncpyzt(bbuf, pass, 1024);
			been_here = 1;
			return strlen(buf);
		}
	}
	else {
		strncpyzt(buf, bbuf, bufsize);
		return strlen(buf);
	}
	report(1, "No password for SSL private key file entered.");
	return 0;
}

static int init_ircd_ctx()
{
	if ((ircd_ctx = SSL_CTX_new(SSLv23_server_method())) == NULL) {
		return SSL_ERR_CTX_NEW;
	}
	SSL_CTX_set_default_passwd_cb(ircd_ctx, ssl_pem_passwd_cb);

	if (SSL_CTX_use_certificate_file(ircd_ctx, SSL_CERT_FILE, SSL_FILETYPE_PEM) <= 0) {
		SSL_CTX_free(ircd_ctx);
		return SSL_ERR_CTX_USE_CERT;
	}
	if (SSL_CTX_use_PrivateKey_file(ircd_ctx, SSL_KEY_FILE, SSL_FILETYPE_PEM) <= 0) {
		SSL_CTX_free(ircd_ctx);
		return SSL_ERR_CTX_USE_KEY;
	}
	if (SSL_CTX_check_private_key(ircd_ctx) == 0) {
		SSL_CTX_free(ircd_ctx);
		return SSL_ERR_CTX_CHECK_KEY;
	}
	return 1;
}

static char *ssl_get_init_error_message(int error)
{
	switch (error) {
		case SSL_ERR_CTX_NEW:
			return "Failed to create new SSL ctx\n";
		case SSL_ERR_CTX_USE_CERT:
			return "Failed to open SSL certificate file\n";
		case SSL_ERR_CTX_USE_KEY:
			return "Failed to open SSL private key file\n";
		case SSL_ERR_CTX_CHECK_KEY:
			return "Authentication failure (server certificate does not match private key file)\n";
		default:
			return "Unknown error occured in ssl_get_init_error_message()\n";
	}
	return NULL;
}

char *init_ssl()
{
	int i;

	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
	if ((i = init_ircd_ctx()) != 1) {
		return ssl_get_init_error_message(i);
	}
	report(0, "Loaded SSL certificate and private key file.");
	return NULL;
}

static void ssl_handle_fatal_error(int ssl_error, int what, aClient *cptr)
{
	int errno_tmp = errno;
	char *error_str = strerror(errno_tmp);
	char *ssl_error_str, *ssl_function;
	char *client_path = get_client_name(cptr, FALSE);

	switch (ssl_error) {
		case SSL_ERROR_NONE:
			ssl_error_str = "No error";
			break;
		case SSL_ERROR_SSL:
			ssl_error_str = "Internal SSL error or protocol error";
			break;
		case SSL_ERROR_WANT_CONNECT:
			ssl_error_str = "SSL functions requested connect()";
			break;
		case SSL_ERROR_WANT_READ:
			ssl_error_str = "SSL functions requested read()";
			break;
		case SSL_ERROR_WANT_WRITE:
			ssl_error_str = "SSL functions requested write()";
			break;
		case SSL_ERROR_WANT_X509_LOOKUP:
			ssl_error_str = "SSL requested an X509 lookup which didn't arrive";
			break;
		case SSL_ERROR_SYSCALL:
			ssl_error_str = "Underlying syscall error";
			break;
		case SSL_ERROR_ZERO_RETURN:
			ssl_error_str = "Underlying socket operation returned zero";
			break;
		default:
			ssl_error_str = "Unknown SSL error (?)";
	}
	switch (what) {
		case SAFE_SSL_READ:
			ssl_function = "SSL_read()";
			break;
		case SAFE_SSL_WRITE:
			ssl_function = "SSL_write()";
			break;
		case SAFE_SSL_ACCEPT:
			ssl_function = "SSL_accept()";
			break;
		default:
			ssl_function = "Unknown SSL function (?)";
	}

	sendto_realops_lev(DEBUG_LEV, "%s to %s aborted with%s error (%s). [%s]",
		ssl_function, client_path, (errno > 0) ? "" : "no ", error_str, ssl_error_str);
	ircdlog(LOG_ERROR, "%s to %s aborted with%s error (%s). [%s]",
		ssl_function, client_path, (errno > 0) ? "" : "no ", error_str, ssl_error_str);

	errno = (errno_tmp != 0) ? errno_tmp : EIO;
	cptr->sockerr = IRCERR_SSL;
	cptr->flags |= FLAGS_DEADSOCKET;
	return;
}

int safe_SSL_accept(aClient *cptr, int fd)
{
	int ssl_error;
	if ((ssl_error = SSL_accept(cptr->ssl)) <= 0) {
		switch ((ssl_error = SSL_get_error(cptr->ssl, ssl_error))) {
			case SSL_ERROR_SYSCALL:
				if (errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR) {
					case SSL_ERROR_WANT_READ:
					case SSL_ERROR_WANT_WRITE:
						return 1;
				}
			default:
				ssl_handle_fatal_error(ssl_error, SAFE_SSL_ACCEPT, cptr);
		}
		return -1;
	}
	return 1;
}

int safe_SSL_read(aClient *cptr, void *buf, size_t size)
{
	int len, ssl_error;
	if ((len = SSL_read(cptr->ssl, buf, size)) <= 0) {
		switch ((ssl_error = SSL_get_error(cptr->ssl, len))) {
			case SSL_ERROR_SYSCALL:
				if (errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR) {
					case SSL_ERROR_WANT_READ:
						errno = EWOULDBLOCK;
						return 0;
				}
			case SSL_ERROR_SSL:
				if (errno == EAGAIN) {
					return 0;
				}
			default:
				ssl_handle_fatal_error(ssl_error, SAFE_SSL_READ, cptr);
				return -1;
		}
	}
	return len;
}

int safe_SSL_write(aClient *cptr, const void *buf, size_t size)
{
	int len, ssl_error;
	if ((len = SSL_write(cptr->ssl, buf, size)) <= 0) {
		switch ((ssl_error = SSL_get_error(cptr->ssl, len))) {
			case SSL_ERROR_SYSCALL:
				if (errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR) {
					case SSL_ERROR_WANT_WRITE:
						errno = EWOULDBLOCK;
						return 0;
				}
			case SSL_ERROR_SSL:
				if (errno == EAGAIN) {
					return 0;
				}
			default:
				ssl_handle_fatal_error(ssl_error, SAFE_SSL_WRITE, cptr);
				return -1;
		}
	}
	return len;
}

int SSL_smart_shutdown(SSL *ssl)
{
	int i, retval = 0;
	for (i = 0; i < 4; i++) {
		if ((retval = SSL_shutdown(ssl)) != 0) {
			break;
		}
	}
	return retval;
}

int ssl_do_handshake(aClient *cptr)
{
	if ((cptr->ssl = SSL_new(ircd_ctx)) == NULL) {
		sendto_realops_lev(DEBUG_LEV, "Failed to create new SSL object for %s on fd %d",
			cptr->sockhost, cptr->fd);
		return 0;
	}
	cptr->flags |= FLAGS_SECURE;
	SSL_set_fd(cptr->ssl, cptr->fd);
	if (!safe_SSL_accept(cptr, cptr->fd)) {
		SSL_set_shutdown(cptr->ssl, SSL_RECEIVED_SHUTDOWN);
		SSL_smart_shutdown(cptr->ssl);
		SSL_free(cptr->ssl);
		return 0;
	}
	return 1;
}

char *ssl_get_cipher_info(SSL *ssl)
{
	SSL_CIPHER *c = SSL_get_current_cipher(ssl);
	static char buf[400], *ver = NULL;
	int bits;

	switch (ssl->session->ssl_version) {
		case SSL2_VERSION:
			ver = "SSLv2";
			break;
		case SSL3_VERSION:
			ver = "SSLv3";
			break;
		case TLS1_VERSION:
			ver = "TLSv1";
			break;
		default:
			ver = "UNKNOWN";
	}
	SSL_CIPHER_get_bits(c, &bits);
	ircsprintf(buf, "%s-%s-%dbits", ver, SSL_get_cipher(ssl), bits);
        return buf;
}

#endif
